#include <itf/option.h>
#include <itf/itf.h>
#include <sim/sim.h>
#include "prop-bp.h"

#define DB(x)


int command_connect(tokc,tokv)
int tokc;
char *tokv[];
{
  int net_n;
#ifdef DBM_NETWORK
  int dbm = 1;
#else
  int dbm = 0;
#endif DBM_NETWORK
  int sl,su,dl,du;
  int s_complex = 0, d_complex = 0;
  int *slayer, *dlayer;
  struct RANGE *sunit_range;
  struct RANGE *dunit_range;
  int partial = 0;
  int restrict = 0;
  float partial_p = 0.0;
  int block = 0;
  int block_size, block_start = 0;
  static struct OPTION options[] =
    {
      {OPTPRES|OPTINTARG,"partial",	0},
      {OPTPRES|OPTINTARG,"block",	0},
      {OPTPRES|OPTINTARG,"restrict",	0},
      {OPTPRES,          "slayer",	0},
      {OPTPRES,          "sunit",	0},
      {OPTPRES,          "dlayer",	0},
      {OPTPRES,          "dunit",	0},
      {OPTLAST,NULL,0}
    };

  IUsage(" [%ppartial <p>] [%pblock <block-size> [<block-start>]]\
 [%pslayer <range>] [%psunit <lrange>:<urange>]\
 [%pdlayer <range>]  [%pdunit <lrange>:<urange>]\
 [%prestrict <int>]");
  if (GiveHelp(tokc)) {
    ISynopsis("connect two groups of units.");
    IHelp(IHelpArgs,
"Connect two groups of units of the current backpropagation network.",
"OPTIONS",
"%ppartial <p>",
"     use partial (rather than complete) connectivity, where <p> is the",
"     proportion of posible connections to create.",
"%pblock <block-size> [<block-start>]",
"     use block (rather than complete) connectivity, where <block-size> is",
"     the number of destination units, and (for layer-to-unit connections)",
"     <block-start> is the number of the first destination unit in the block.",
"%prestrict <1 or -1>",
"     restrict weights to be non-negative (1) or non-positive (-1).",
"%pslayer <range>",
"     source layers in <range>",
"%psunit <lrange>:<urange>",
"     source units in <urange> in layers in <lrange>",
"%pdlayer <range>",
"     destination layers in <range>",
"%pdunit <lrange>:<urange>",
"     destination units in <urange> in layers in <lrange>",
NULL);
    return;
  }
  if (!NetExists(net)) {
    IError("current net is null");
    return;
  }
  net_n = net->my_n;
  AllocDisplayRanges(net,&dlayer,&slayer,&dunit_range,&sunit_range);
  /***
   * process options
   */
  tokv++;
  while (IsOption(tokv[0])) {
    if (OptionMatch(tokv[0],"partial",options,&partial)) {
      if (IExpectNumber(tokv[1])) {
	partial_p = atof(*++tokv); tokc--;
      }
     } else if (OptionMatch(tokv[0],"restrict",options,NULL)) {
       if (!IExpectInteger(tokv[1])) return;
       restrict = atoi(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"block",options,&block)) {
      if (IExpectInteger(tokv[1])) {
	block_size = atoi(*++tokv); tokc--;
	if (IIsInteger(tokv[1])) {
	  block_start = atoi(*++tokv); tokc--;
	}
      }
    } else if (OptionMatch(tokv[0],"slayer",options,NULL)       ||
	       OptionMatch(tokv[0],"dlayer",options,NULL)       ||
	       OptionMatch(tokv[0],"sunit", options,&s_complex) ||
	       OptionMatch(tokv[0],"dunit", options,&d_complex)) {
      if (!SetDisplayRanges(net,&tokv,dlayer,slayer,dunit_range,sunit_range))
	return;
    } else {
      IError("option not understood: \"%s\"",tokv[0]);
      return;
    }
    tokv++;
    tokc--;
  }
  /*
   * For each source layer...
   */
  for (sl=0; sl<net->n_layers; sl++)
    if (slayer[sl])
      /*
       * For each destination layer...
       */
      for (dl=0; dl<net->n_layers; dl++)
	if (dlayer[dl])
	  if ((dbm && sl==dl) || s_complex || partial) 
	    /*
	     * Unit to unit connections
	     */
	    for (su=sunit_range[sl].lo; su<=sunit_range[sl].hi; su++)
	      for (du=dunit_range[dl].lo; du<=dunit_range[dl].hi; du++) {
		if (   (!dbm || sl!=dl || su > du) 	/* no self conn */
		    && (!partial || with_prob(partial_p))) {
		  DB(printf("connecting %d:%d to %d:%d\n",sl,su,dl,du););
		  /*
		   * used to be
		   * Connect(net_n,sl,su,net_n,dl,du,LINK);
		   */
		  ConnectUnitBlock(net_n,sl,net_n,dl,du,su,1,LINK);
		}
	      }
	  else 
	    if (d_complex) 
	      /*
	       * Layer to unit connections
	       */
	      for (du=dunit_range[dl].lo; du<=dunit_range[dl].hi; du++)
		if (block) {
		  DB(printf("connecting %d to %d:%d block %d %d\n",sl,dl,du,block_size,block_start););
		  ConnectUnitBlock(net_n,sl,net_n,dl,du,block_start,block_size,LINK);
		} else {
		  DB(printf("connecting %d to %d:%d\n",sl,dl,du););
		  ConnectUnitComplete(net_n,sl,net_n,dl,du,LINK);
		}
	    else
	      /*
	       * Layer to layer connections
	       */
	      if (block) {
		DB(printf("connecting %d to %d block %d\n",sl,dl,block_size););
		ConnectBlock(net_n,sl,net_n,dl,block_size,LINK);
	      } else {
		DB(printf("connecting %d to %d\n",sl,dl););
		ConnectComplete(net_n,sl,net_n,dl,LINK);
	      }
  if (restrict)
    RestrictWeights(net,slayer,sunit_range,dlayer,dunit_range,restrict);
}
