#include <itf/option.h>
#include <itf/itf.h>
#include <sim/sim.h>
#include <lib/example.h>
#include "example-bp.h"
#include "prop-bp.h"

#define DB(x) x

#define PLOT_ENERGY 	1
#define PLOT_ENTROPY	2
#define PLOT_FENERGY	4

int ave_energy_p = FALSE;	/* whether energy calcs. returns averages */

static int nplots_g;
static int npoints_g;
static real **plot_arr;

extern real **fp;	/* prop-dbm.c */

#define energy(s,d,l) \
  (UNIT_OUTPUT(s)*UNIT_OUTPUT(d)*LINK_WEIGHT(l))
#define entropy(p) \
  (((p)==0.0||(p)==1.0) ? 0.0 : ((p)*log(p)+(1.0-(p))*log(1.0-(p))))


/******************************************************************************
 * Utilities
 */

int
  Connected(net,sl,dl)
NETWORK *net;
int sl,dl;
{
  DoLayer(dl) {
    ForEachUnit(UNIT) {
      ForEachLinkSet(LINK,UNIT) {
	if (OtherLayerN == sl) return 1;
      } EndForEach;
    } EndForEach;
  } EndDo;
  return 0;
}

real
  Energy(net,slayer,dlayer)
NETWORK *net;
int slayer,dlayer;
{
  int n=0;
  real e = 0.0;
  if (!dlayer) {
    ForEachLinkCompiled(LINK,UNIT) {
      e -= energy(s_unit,d_unit,link);
      n++;
    } EndForEach;
  } else {
    DoLayer(dlayer) {
      ForEachUnit(UNIT) {
	ForEachLinkSet(LINK,UNIT) {
	  if (!slayer || OtherLayerN==slayer || (slayer==dlayer && IsOtherBiasLayer))
	    ApplyToLinks(if (!link->lesioned) {e -= energy(other,unit,link); n++;});
	} EndForEach;
      } EndForEach;
    } EndDo;
  }
  if (ave_energy_p) e /= ((real)n);
  return(e);
}

real
  Entropy(net,dlayer)
NETWORK *net;
int dlayer;
{
  int n=0;
  real y,e = 0.0;
  if (!dlayer) {
    ForEachUnitCompiled(UNIT) {
      y  = InvScaleOutput(UNIT_OUTPUT(unit));
      e += entropy(y);
      n++;
    } EndForEach;
  } else {
    DoLayer(dlayer) {
      ForEachUnit(UNIT) {
	if (!unit->lesioned) {
	  y  = InvScaleOutput(UNIT_OUTPUT(unit));
	  e += entropy(y);
	  n++;
	}
      } EndForEach;
    } EndDo;
  }
  if (ave_energy_p) e /= ((real)n);
  return(e);
}

real
  FreeEnergy(net,slayer,dlayer)
NETWORK *net;
int slayer,dlayer;
{
  return (Energy(net,slayer,dlayer) +
	  Entropy(net,dlayer)*((real)(net->p->temp+net->p->final_temp)));
}

real
  SumFreeEnergy(net,slayer,dlayer)
NETWORK *net;
int *slayer,*dlayer;
{
  int sl,dl;
  real sum=0.0;
  for_each (dl,net->n_layers) 
    if (dlayer[dl])
      for_each (sl,net->n_layers) 
	if (slayer[sl])
	  sum += FreeEnergy(net,sl,dl);
  return sum;
}


/******************************************************************************
 * Energy 
 */

void
  CreateEnergyPlotArr(net,slayer,dlayer,plot)
NETWORK *net;
int *slayer,*dlayer,plot;
{
  int sl,dl;
  nplots_g = 0;
  for (dl=net->n_layers-1; dl>0; dl--)
    if (dlayer[dl]) {
      if (plot&PLOT_ENTROPY) nplots_g++;
      if (plot&PLOT_ENERGY) {
	if (slayer[0]) nplots_g++;
	for (sl=net->n_layers-1; sl>0; sl--)
	  if (slayer[sl] && Connected(net,sl,dl)) nplots_g++;
      }
    } /* dlayer */
  if (plot&PLOT_FENERGY) nplots_g++;
  if (plot_arr) free(plot_arr);
  plot_arr = ((real**)calloc_2d(nplots_g,npoints_g,sizeof(real)));
}

void
  ComputeEnergy(net,slayer,dlayer,plot,n)
NETWORK *net;
int *slayer,*dlayer,plot,n;
{
  int sl,dl,p=0;
  for (dl=net->n_layers-1; dl>0; dl--)
    if (dlayer[dl]) {
      if (plot&PLOT_ENTROPY) plot_arr[p++][n] = Entropy(net,dl);
      if (plot&PLOT_ENERGY) {
	if (slayer[0]) plot_arr[p++][n] = Energy(net,0,dl);
	for (sl=net->n_layers-1; sl>0; sl--) 
	  if (slayer[sl] && Connected(net,sl,dl))
	    plot_arr[p++][n] = Energy(net,sl,dl);
      } /* energy */
    } /* dlayer */
  if (plot&PLOT_FENERGY) plot_arr[p++][n] = FreeEnergy(net,0,0);
}

void
  PrintEnergyPlotTitle(net,ex1_n,ex2_n)
NETWORK *net;
int ex1_n,ex2_n;
{
  fprintf(dout,"title %s ",net->name);
  PrintExampleName(current_el->index[ex1_n]);
  fprintf(dout," to ");
  PrintExampleName(current_el->index[ex2_n]);
  fprintf(dout,"\n");
}

void
  PrintEnergyPlotInit(net,slayer,dlayer,plot)
NETWORK *net;
int *slayer,*dlayer,plot;
{
  int p,n,sl,dl;
  int emax,egap = -((int)HUGE);
  int *emin = ITempCalloc(int,nplots_g);
  for (p=0; p<nplots_g; p++) emin[p] = ((int)HUGE);
  /*
   * Calculate plotting ranges
   */
  for (dl=net->n_layers-1,p=0; dl>0; dl--)
    if (dlayer[dl]) {
      if (plot&PLOT_ENTROPY) {
	for (n=0,emax = -(int)HUGE; n<npoints_g; n++) {
	  emin[p] = min(emin[p],floor(plot_arr[p][n]));
	  emax    = max(emax,ceil(plot_arr[p][n]));
	}
	egap = max(egap,emax-emin[p]);  p++;
      }
      if (plot&PLOT_ENERGY) {
	if (slayer[0]) {
	  for (n=0,emax = -(int)HUGE; n<npoints_g; n++) {
	    emin[p] = min(emin[p],floor(plot_arr[p][n]));
	    emax    = max(emax,ceil(plot_arr[p][n]));
	  }
	  egap = max(egap,emax-emin[p]);  p++;
	}
	for (sl=net->n_layers-1; sl>0; sl--) 
	  if (slayer[sl] && Connected(net,sl,dl)) {
	    for (n=0,emax = -(int)HUGE; n<npoints_g; n++) {
	      emin[p] = min(emin[p],floor(plot_arr[p][n]));
	      emax    = max(emax,ceil(plot_arr[p][n]));
	    }
	    egap = max(egap,emax-emin[p]);  p++;
	  }
      } /* energy */
    } /* dlayer */
    if (plot&PLOT_FENERGY) {
      for (n=0,emax = -(int)HUGE; n<npoints_g; n++) {
	emin[p] = min(emin[p],floor(plot_arr[p][n]));
	emax    = max(emax,ceil(plot_arr[p][n]));
      }
    }
  /*
   * Print out ranges
   */
  fprintf(dout,"graph ");
  for (dl=net->n_layers-1,p=0; dl>0; dl--)
    if (dlayer[dl]) {
      if (plot&PLOT_ENTROPY)
	fprintf(dout,"%d %d %d entropy %d ",dl,emin[p],egap+emin[p++],dl);
      if (plot&PLOT_ENERGY) {
	if (slayer[0]) fprintf(dout,"b%d %d %d energy bias%d ",dl,emin[p],egap+emin[p++],dl);
	for (sl=net->n_layers-1; sl>0; sl--)
	  if (slayer[sl] && Connected(net,sl,dl)) 
	    fprintf(dout,"%d%d %d %d energy %d-%d ",sl,dl,emin[p],egap+emin[p++],sl,dl);
      }
    } /* dlayer */
  if (plot&PLOT_FENERGY) fprintf(dout,"fe %d %d free energy ",emin[p],emax);
  fprintf(dout,"\n");
}

void
  PrintEnergyPlot(net,slayer,dlayer,plot)
NETWORK *net;
int *slayer,*dlayer,plot;
{
  int sl,dl,n,p;
  for (n=0,p=0; n<npoints_g; n++,p=0) {
    fprintf(dout,"data ");
    for (dl=net->n_layers-1; dl>0; dl--)
      if (dlayer[dl]) {
	if (plot&PLOT_ENTROPY) fprintf(dout,"%d %f ",dl,plot_arr[p++][n]);
	if (plot&PLOT_ENERGY) {
	  if (slayer[0]) {
	    fprintf(dout,"b%d %f ",dl,plot_arr[p++][n]);
	  }
	  for (sl=net->n_layers-1; sl>0; sl--) 
	    if (slayer[sl] && Connected(net,sl,dl)) 
	      fprintf(dout,"%d%d %f ",sl,dl,plot_arr[p++][n]);
	} /* energy */
      } /* dlayer */
    if (plot&PLOT_FENERGY) fprintf(dout,"fe %f ",plot_arr[p++][n]);
    fprintf(dout,"\n");
  } /* npoints */
}


/******************************************************************************
 * plot_energy()
 */

int
  command_plot_energy(tokc,tokv)
int tokc;
char *tokv[];
{
  int i;
  int clamped_layer_n = 0,  ave = 0;
  int  ex1_n = -1,      ex2_n = -1;
  char *ex1_str = NULL, *ex2_str = NULL;
  real *out1,*out2;
  real step, lo = 0.0, hi = 1.0;
  int init=TRUE;
  int plot_arg, plot = PLOT_FENERGY;
  int *slayer, *dlayer;
  struct RANGE *sunit_range;
  struct RANGE *dunit_range;
  static struct OPTION options[] =
    {
      {OPTPRES|OPTREALARG,"lo",		0},
      {OPTPRES|OPTREALARG,"hi",		0},
      {OPTPRES|OPTREALARG,"npoints",	0},
      {OPTPRES|OPTINTARG,"clamp",	0},
      {OPTPRES,"slayer",		0},
      {OPTPRES,"dlayer",		0},
      {OPTSIGN,"energy",		0},
      {OPTSIGN,"entropy",		0},
      {OPTSIGN,"fenergy",		0},
      {OPTSIGN,"init",			0},
      {OPTSIGN,"average",		0},
      {OPTLAST,NULL,0}
    };
  IUsage("<example1> <example2> \
 [%pclamp <int>]  [%senergy] [%sentropy] [%sfenergy] \
 [%pslayer <range>] [%pdlayer <range>] \
 [%pnpoints <int>] [%plo <real>] [%phi <real>] \
 [%sinit] [%saverage]");
  if (GiveHelp(tokc)) {
    ISynopsis("Plot enengy on line between two examples.");
    IHelp(IHelpArgs,
"Plot enengy on line between fixedpoints of two examples.",
"Output format for stripgraph.",
"OPTIONS",
"<example1>      : first (left) example (<int> or <comment>)",
"<example2>      : second (right) example (<int> or <comment>)",
"%pclamp <int>    : number of layer to clamp (defaults to output layer).",
"%senergy       : plot energy of sets of connections (default %n).",
"%sentropy      : plot entropy of groups of units (default %n).",
"%sfenergy      : plot free energy of subset of network (default %y).",
"%pslayer <range> : source layer(s) to compute energy over.",
"%pdlayer <range> : destination layer(s) to compute energy over.",
"%pnpoints <int>  : number of points to plot",
"%plo <real>      : left starting position (normalized to distance",
"                   between examples with left example = 0 (default 0.0)",
"%phi <real>      : right starting position (normalized to distance",
"                   between examples with right example = 1 (default 1.0)",
"%sinit         : print initialization (default %y).",
"%saverage      : print energy/entropy averages (default %n)",
	  NULL);
    return;
  }
  if (!NetExists(net)) {
    IError("current net is null");
    return;
  }
  if (current_el == NULL) {
    IError("current example set is NULL");
    return;
  }
  npoints_g = 100;
  AllocDisplayRanges(net,&dlayer,&slayer,&dunit_range,&sunit_range);
  /***
   * process options
   */
  tokv++;	/* ignore command name */
  if (IIsInteger(*tokv)) ex1_n = atoi(*tokv++);
  else			 ex1_str = *tokv++; 
  if (IIsInteger(*tokv)) ex2_n = atoi(*tokv++);
  else			 ex2_str = *tokv++; 
  tokc -= 2;
  while (IsOption(tokv[0])) {
    if        (OptionMatch(tokv[0],"energy",options,&plot_arg)) {
      if (plot_arg) plot |=  PLOT_ENERGY;
      else	    plot &= ~PLOT_ENERGY;
    } else if (OptionMatch(tokv[0],"entropy",options,&plot_arg)) {
      if (plot_arg) plot |=  PLOT_ENTROPY;
      else	    plot &= ~PLOT_ENTROPY;
    } else if (OptionMatch(tokv[0],"fenergy",options,&plot_arg)) {
      if (plot_arg) plot |=  PLOT_FENERGY;
      else	    plot &= ~PLOT_FENERGY;
    } else if (OptionMatch(tokv[0],"clamp",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      clamped_layer_n = atoi(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"slayer",options,NULL) ||
	       OptionMatch(tokv[0],"dlayer",options,NULL)) {
      if (!SetDisplayRanges(net,&tokv,dlayer,slayer,dunit_range,sunit_range))
	return;
    } else if (OptionMatch(tokv[0],"lo",options,NULL)) {
      if (!IExpectNumber(tokv[1])) return;
      lo = atof(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"hi",options,NULL)) {
      if (!IExpectNumber(tokv[1])) return;
      hi = atof(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"npoints",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      npoints_g = atoi(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"init",options,&init)) {
    } else if (OptionMatch(tokv[0],"average",options,&ave)) {
    } else {
      IError("option not understood: \"%s\"",tokv[0]);
      return;
    }
    tokv++;
    tokc--;
  }
  if ((ex1_n==-1 && ex1_str==NULL) || (ex2_n==-1 && ex2_str==NULL)) {
    IError("Must specify two examples");  return;
  }
  step = (hi-lo)/((real)npoints_g);			/* from lo to hi */
  npoints_g++;
  ave_energy_p = ave;
  if (!clamped_layer_n) clamped_layer_n = net->p->output_layer_n;
  net->layer[net->p->input_layer_n].p->clamped = TRUE;
  net->layer[clamped_layer_n].p->clamped       = TRUE;
  /* 
   * Save ex2 states
   */
  if (ex2_n == -1) ex2_n = ExampleNumber(current_el,ex2_str);
  out2 = ITempCalloc(real,net->layer[clamped_layer_n].n_units);
  if (!LoadExample(net,ex2_n,ex2_str)) return;
  DoLayer(clamped_layer_n) {
    ForEachUnit(UNIT) {
      out2[unit_n] = UNIT_TARGET(unit);
    } EndForEach;
  } EndDo;
  /* 
   * Save ex1 states
   */
  if (ex1_n == -1) ex1_n = ExampleNumber(current_el,ex1_str);
  out1 = ITempCalloc(real,net->layer[clamped_layer_n].n_units);
  if (!LoadExample(net,ex1_n,ex1_str)) return;
  DoLayer(clamped_layer_n) {
    ForEachUnit(UNIT) {
      out1[unit_n] = UNIT_TARGET(unit);
    } EndForEach;
  } EndDo;
  /*
   * For each data point...
   */
  if (init) CreateEnergyPlotArr(net,slayer,dlayer,plot);
  for (i=0; i<npoints_g; i++) {
    /*
     * Set clamped unit states
     */
    DoLayer(clamped_layer_n) {
      ForEachUnit(UNIT) {
	UNIT_TARGET(unit) = out1[unit_n] + (lo+i*step)*(out2[unit_n]-out1[unit_n]);
      } EndForEach;
    } EndDo;
    /*
     * Run network and compute error
     */
    SettleToFixedpoint(net,NEGATIVE,FALSE,FALSE,FALSE,FALSE);
    ComputeEnergy(net,slayer,dlayer,plot,i);	/* compute data */
  } /* npoints */
  if (init) {
    PrintEnergyPlotTitle(net,ex1_n,ex2_n);
    PrintEnergyPlotInit(net,slayer,dlayer,plot);
  }
  PrintEnergyPlot(net,slayer,dlayer,plot);	/* actually write data */
}


/******************************************************************************
 * plot_3d_energy()
 */

int
  command_plot_3d_energy(tokc,tokv)
int tokc;
char *tokv[];
{
  int i,j,k,sl=0,dl=0;
  int  ex1_n = -1,      ex2_n = -1;
  char *ex1_str = NULL, *ex2_str = NULL;
  real *out1,*out2,*states;
  int n_units;
  static struct OPTION options[] =
    {
      {OPTPRES|OPTREALARG,"npoints",	0},
      {OPTPRES|OPTINTARG,"slayer",	0},
      {OPTPRES|OPTINTARG,"dlayer",	0},
      {OPTLAST,NULL,0}
    };
  IUsage("<example1> <example2> [%pslayer <int>] [%pdlayer <int>] \
 [%pnpoints <int>]");
  if (GiveHelp(tokc)) {
    ISynopsis("Plot enengy on plane through two examples.");
    IHelp(IHelpArgs,
"Plot enengy on plan through fixedpoints of two examples.",
"Output format for tg.",
"OPTIONS",
"<example1>      : first (left) example (<int> or <comment>)",
"<example2>      : second (right) example (<int> or <comment>)",
"%pslayer <int>  : source layer to compute energy over.",
"%pdlayer <int>  : destination layer to compute energy over.",
"%pnpoints <int> : number of points to plot",
	  NULL);
    return;
  }
  if (!NetExists(net)) {IError("current net is null"); return; }
  if (current_el == NULL) {IError("current example set is NULL"); return; }
  npoints_g = 100;
  /***
   * process options
   */
  tokv++;	/* ignore command name */
  if (IIsInteger(*tokv)) ex1_n = atoi(*tokv++);
  else			 ex1_str = *tokv++; 
  if (IIsInteger(*tokv)) ex2_n = atoi(*tokv++);
  else			 ex2_str = *tokv++; 
  tokc -= 2;
  while (IsOption(tokv[0])) {
    if (OptionMatch(tokv[0],"slayer",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      sl = atoi(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"dlayer",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      dl = atoi(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"npoints",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      npoints_g = atoi(*++tokv); tokc--;
    } else {
      IError("option not understood: \"%s\"",tokv[0]);
      return;
    }
    tokv++;
    tokc--;
  }
  if ((ex1_n==-1 && ex1_str==NULL) || (ex2_n==-1 && ex2_str==NULL)) {
    IError("Must specify two examples");  return;
  }
  npoints_g++;
  n_units = net->p->n_units;
  states = ITempCalloc(real,n_units);
  /* 
   * Save ex1 states
   */
  if (ex1_n == -1) ex1_n = ExampleNumber(current_el,ex1_str);
  out1 = ITempCalloc(real,n_units);
  if (!LoadExample(net,ex1_n,ex1_str)) return;
  PerformNet(net);
  SaveUnitStates(net,out1);
  /* 
   * Save ex2 states
   */
  if (ex2_n == -1) ex2_n = ExampleNumber(current_el,ex2_str);
  out2 = ITempCalloc(real,n_units);
  if (!LoadExample(net,ex2_n,ex2_str)) return;
  PerformNet(net);
  SaveUnitStates(net,out2);
  /*
   * Write out header
   */
  fprintf(dout,"DATA: NORMAL_GRID\n");
  PrintExampleName(current_el->index[ex1_n]);
  fprintf(dout,"\n");
  PrintExampleName(current_el->index[ex2_n]);
  fprintf(dout,"\nEnergy\n0 0 0\n%d %d\n",npoints_g,npoints_g);
  /*
   * For each data point...
   */
  for_each (i,npoints_g)
    for_each (j,npoints_g) {
      /*
       * Set unit states
       */
      for_each (k,n_units) {
	states[k] = (i*out1[k] + j*out2[k])/((float) (npoints_g-1));
	if      (states[k] > UnitUpperBound) states[k] = UnitUpperBound;
	else if (states[k] < UnitLowerBound) states[k] = UnitLowerBound;
      }
      RestoreUnitStates(net,states);
      fprintf(dout,"%d %d %f\n",i,j,FreeEnergy(net,sl,dl));
      fflush(dout);
    }
  fprintf(dout,"DATA: END\n");
}
