#include <itf/option.h>
#include <itf/itf.h>
#include <sim/sim.h>
#include <lib/example.h>
#include "example-bp.h"
#include "prop-bp.h"

#define DB(x) x

#define SIM_HAM		1
#define SIM_COS		2
#define SIM_COR		4
#define SIM_DST		8

static int  *ord	 	= NULL;
static char reorder_label[20];
static real display_min, display_max;
static int  display_scale	= FALSE;
static int  display_single	= FALSE;
#define display_label_width	(display_single? 8 : 4)

extern real	layer_cos_angle();	/* perform-bp.c */
#ifdef DBM_NETWORK
extern real	Energy();		/* energy-bp.c */
extern int 	ave_energy_p;		/* energy-bp.c */
#endif DBM_NETWORK



/*****************************************************************************
 * Utilities
 */

char **calloc_2d ( N1, N2, size )
unsigned N1, N2, size;
{
  register int i;
  char	**ptr = (char**) calloc ( N1, sizeof(char*) );
  if( ptr==NULL ) { 
    return( NULL );
  }
  for( i=0; i< N1 ; i++ ) 
    if(NULL==(ptr[i]= calloc ( N2, size ))) 
      return( NULL );
  return ( ptr );
}

real
  cos_angle(v1,v2,n)
real *v1,*v2;
int n;
{
  int i;
  real l12 = 0.0, l1 = 0.0, l2 = 0.0;
  for_each(i,n) {
    l1  += v1[i]*v1[i];    l2  += v2[i]*v2[i];    l12 += v1[i]*v2[i];
  }
  if (l1==0.0 || l2==0.0) return 0.0;
  return (l12/sqrt(l1*l2));
}

int
  hamming_distance(v1,v2,n)
real *v1,*v2;
int n;
{
  int i,dist = 0;
  for_each(i,n) if (v1[i] != v2[i]) dist++;
  return dist;
}

float
  correlation(v1,v2,lv)
float *v1,*v2;
int lv;
{
  int i;
  float d1,d2,flv=((float) lv);
  float mean1=0.0, mean2=0.0, var1=0.0, var2=0.0, cov=0.0;
  for_each(i,lv) {
    mean1 += v1[i];   mean2 += v2[i];
  }
  mean1 /= flv;  mean2 /= flv;   
  for_each(i,lv) {
    d1 = (v1[i]-mean1);
    d2 = (v2[i]-mean2);
    var1 += d1*d1;
    var2 += d2*d2;
    cov  += d1*d2;
  }
  if (var1==0.0 || var2 == 0.0) return 0.0;
  return ((cov/flv)/(sqrt(var1/flv)*sqrt(var2/flv)));
}

real
  euclidean_distance(v1,v2,n)
real *v1,*v2;
int n;
{
  int i;
  real dist = 0.0;
  for_each(i,n) dist += sqr(v1[i]-v2[i]);
  return sqrt(dist);
}

void
  swap(v,i,j)
real *v;
int i,j;
{
  real tmp = v[i];  v[i] = v[j];  v[j] = tmp;
}

real
  gauss(x,sd)
real x,sd;
{
  return(exp(-(x*x)/(2.0*sd*sd)));
}

real
  length(v,n)
real *v;
int n;
{
  int i;
  real len = 0.0;
  for_each(i,n) len += v[i]*v[i];
  return sqrt(len);
}

void
  normalize(v,n)
real *v;
int n;
{
  int i;
  real len = length(v,n);
  if (len != 0.0)
    for_each(i,n) v[i] /= len;
}

float
  matrix_correlation(m1,m2,lm)
float **m1,**m2;
int lm;
{
  int i,j;
  float d1,d2,sz=((float) lm*lm);
  float mean1=0.0, mean2=0.0, var1=0.0, var2=0.0, cov=0.0;
  for_each(j,lm)
    for_each(i,lm) {
      mean1 += m1[j][i];   mean2 += m2[j][i];
    }
  mean1 /= sz;  mean2 /= sz;   
  for_each(j,lm)
    for_each(i,lm) {
      d1 = (m1[j][i]-mean1);
      d2 = (m2[j][i]-mean2);
      var1 += d1*d1;
      var2 += d2*d2;
      cov  += d1*d2;
    }
  if (var1==0.0 || var2 == 0.0) return 0.0;
  return ((cov/sz)/(sqrt(var1/sz)*sqrt(var2/sz)));
}

float
  matrix_correlation_nodiag(m1,m2,lm)
float **m1,**m2;
int lm;
{
  int i,j;
  float d1,d2,sz=((float) lm*(lm-1));
  float mean1=0.0, mean2=0.0, var1=0.0, var2=0.0, cov=0.0;
  for_each(j,lm)
    for_each(i,lm) 
      if (i != j) {
	mean1 += m1[j][i];   mean2 += m2[j][i];
      }
  mean1 /= sz;  mean2 /= sz;   
  for_each(j,lm)
    for_each(i,lm) 
      if (i != j) {
	d1 = (m1[j][i]-mean1);
	d2 = (m2[j][i]-mean2);
	var1 += d1*d1;
	var2 += d2*d2;
	cov  += d1*d2;
      }
  if (var1==0.0 || var2 == 0.0) return 0.0;
  return ((cov/sz)/(sqrt(var1/sz)*sqrt(var2/sz)));
}


float
  matrix_ave_cos_angle(m1,m2,ni,nj)
float **m1,**m2;
int ni,nj;
{
  int i;
  float ave = 0.0;
  for_each (i,ni) ave += cos_angle(m1[i],m2[i],nj);
  return ave/ni;
}


char
  *MetricName(metric)
int metric;
{
  switch(metric) {
  case SIM_HAM: return("Hamming");
  case SIM_COS: return("Proximity");
  case SIM_COR: return("Correlation");
  case SIM_DST: return("Distance");
  default:      return "";
  }
}

void
  SetUnitInputsFromLayers(net,dl,slayer,t)
NETWORK *net;
int dl,*slayer,t;
{
  DoLayer(dl) {
    ForEachUnit(UNIT) {
      unit->x = 0.0;
      ForEachLinkSet(LINK,UNIT) {
	if (slayer[OtherLayerN])
	  ApplyToEachLink
	    (
	     unit->x += LINK_WEIGHT(link) * UNIT_OUTPUT_T(other,t-1);
	     );
      } EndForEach;
    } EndForEach;
  } EndDo;
}

real
  similarity(v1,v2,n,metric)
real *v1,*v2;
int n,metric;
{
  switch(metric) {
  case SIM_HAM: return(-hamming_distance(v1,v2,n));
  case SIM_COS: return(cos_angle(v1,v2,n));
  case SIM_COR: return(correlation(v1,v2,n));
  case SIM_DST: return(-euclidean_distance(v1,v2,n));
  default:      return 0.0;
  }
}

void
  FillSimilarityMatrix(mat,patterns,n_pats,n_elts,metric)
real **mat,**patterns;
int n_pats,n_elts,metric;
{
  int i,j;
  real val, hi = -HUGE, lo = HUGE;
  for_each (i,n_pats)
    for (j=0; j<=i; j++) {
      mat[i][j] = mat[j][i] = val = similarity(patterns[i],patterns[j],n_elts,metric);
      lo = min(lo,val);  hi = max(hi,val);
    }
  if (display_scale) 
    for_each (i,n_pats)
      for (j=0; j<=i; j++) 
	mat[i][j] = mat[j][i] = (mat[i][j]-lo)/(hi-lo);	/* scales to [0 1] */
}

real
  **SimilarityMatrix(patterns,n_pats,n_elts,metric)
real **patterns;
int n_pats,n_elts,metric;
{
  real **mat = ((real**)calloc_2d(n_pats,n_pats,sizeof(real)));
  FillSimilarityMatrix(mat,patterns,n_pats,n_elts,metric);
  return mat;
}


#define diag(x) (1.0-sqrt(x))

real
  diagonality(M,n)
real **M;
int n;
{
  int i,j;
  real m, sumM = 0.0, sum = 0.0, fn = ((float) n);
  for_each(j,n) 
    for (i=0; i<j; i++) {
      sumM += (m=M[ord[i]][ord[j]]);
      sum  += m*diag(fabs(i-j)/fn);
    }
  return sum/sumM;
}

real
  diagonality_gap(a,b,M,n)
int a,b;
real **M;
int n;
{
  int i,j;
  real gap = 0.0, fn = ((float) n);
  for_each(j,n) 
    if (j != a && j != b)
      gap += (M[ord[j]][ord[a]]-M[ord[j]][ord[b]])*
	     (diag(fabs(j-b)/fn) - diag(fabs(j-a)/fn));
  return gap;
}


/*****************************************************************************
 * ReorderSimilarityMatrix()
 */
	       
void
  ReorderSimilarityMatrix(sim,n_pats)
real **sim;
int n_pats;
{
  int i,j,done;
  real init_t  = net->p->alpha;
  real final_t = net->p->epsilon;
  real t_decay = net->p->weight_decay;
  real t       = init_t;
  real dg;
  do {
    i = random()%n_pats;
    j = random()%n_pats;
    if (i != j) {
      dg = diagonality_gap(i,j,sim,n_pats);
      if (with_prob(sigmoid(dg/t))) swap(ord,i,j);
      t *= t_decay;
    }
  } while (t > final_t);
  do {
    done = TRUE;
    for_each(j,n_pats)
      for (i=0; i<j; i++) {
	dg = diagonality_gap(i,j,sim,n_pats);
	if (dg > 0.0) {
	  swap(ord,i,j);
	  done = FALSE;
	}
      }
  } while (!done);
  DB(
     for_each(i,n_pats) printf(" %2d",ord[i]);
     printf("\n");
     );
}

#define dissim(x)	(1.0-(x))

real
  stress_1D(pos,sim,n_pats)
real *pos,**sim;
int n_pats;
{
  int i,j;
  real d,diff, stress=0.0, sum=0.0;
  for_each(i,n_pats)
    for (j=0; j<i; j++) {
      d       = dissim(sim[i][j]);
      diff    = d - fabs(pos[i]-pos[j]);
      stress += diff*diff;
      sum    += d*d;
    }
  return sqrt(stress/sum);
}

void
  center_1D(pos,n)
real *pos;
int n;
{
  int i;
  real sum;
  for_each(i,n) sum += pos[i];
  sum /= ((float) n);
  for_each(i,n) pos[i] -= sum;
}

void
  SSA_1D(sim,n_pats)
real **sim;
int n_pats;
{
  int i,j,min_i;
  real s,old_s;
  real noise = net->p->input_noise;
  real n_tol = net->p->alpha;
  real decay = net->p->weight_decay;
  real s_tol = net->p->tolerance;
  real e     = net->p->epsilon;
  real *pos = ((real*)calloc(n_pats,sizeof(real)));
  real *dpos = ((real*)calloc(n_pats,sizeof(real)));
  for_each(j,n_pats) pos[j] = Noise(-0.5,0.5);
  center_1D(pos,n_pats);
  s = stress_1D(pos,sim,n_pats);
  do {
    for_each(j,n_pats) {
      dpos[j] = 0;
      for_each(i,n_pats) 
	if (i != j) 
	  dpos[j] += (1.0 - fabs(pos[j]-pos[i])/dissim(sim[j][i]))*
	             (pos[j]-pos[i]);
    }
    for_each(j,n_pats) pos[j] += e*dpos[j]/((real)n_pats);
    if (noise > n_tol) for_each(j,n_pats) pos[j] += Noise(-noise,noise);
    center_1D(pos,n_pats);
    noise *= decay;
    old_s = s;
    s = stress_1D(pos,sim,n_pats);
  } while (fabs(s-old_s) > s_tol || noise > n_tol);
  for_each(j,n_pats) {
    min_i = 0;
    old_s = HUGE;
    for_each(i,n_pats)
      if (pos[i] < old_s) {
	min_i = i;
	old_s = pos[i];
      }
    DB(printf(" %6.3f",pos[min_i]););
    ord[j] = min_i;
    pos[min_i] = HUGE;
  }
  DB(
     printf("\n");
     for_each(i,n_pats) printf(" %2d",ord[i]);
     printf("\n");
     printf("stress = %f\n",s);
     );
}


int command_write_ord(tokc,tokv)
int tokc;
char *tokv[];
{
  int i;
  static struct OPTION options[] =
    {
      {OPTPRES|OPTSTRARG,	"label",	0},
      {OPTLAST,NULL,0}
    };
  IUsage("[%plabel <str>]");
  if (GiveHelp(tokc)) {
    ISynopsis("write out example ordering");
    IHelp(IHelpArgs,	  
	  NULL);
    return;
  }
  tokv++;	/* ignore command name */
  while (IsOption(tokv[0])) {
    if        (OptionMatch(tokv[0],"label",options,NULL)) {
      strcpy(reorder_label,*++tokv); tokc--;
    } else {
      IError("option not understood: \"%s\"",tokv[0]);
      return;
    }
    tokv++;
    tokc--;
  }
  fprintf(dout,"read-ord");
  if (reorder_label[0]) fprintf(dout," +label %s",reorder_label);
  for_each(i,current_el->n_examples) fprintf(dout," %d",ord[i]);
  fprintf(dout,"\n");
}

int command_read_ord(tokc,tokv)
int tokc;
char *tokv[];
{
  int i;
  static struct OPTION options[] =
    {
      {OPTPRES|OPTSTRARG,	"label",	0},
      {OPTLAST,NULL,0}
    };
  IUsage("[%plabel <str>]");
  if (GiveHelp(tokc)) {
    ISynopsis("read in example ordering");
    IHelp(IHelpArgs,	  
	  NULL);
    return;
  }
  tokv++;	/* ignore command name */
  while (IsOption(tokv[0])) {
    if        (OptionMatch(tokv[0],"label",options,NULL)) {
      strcpy(reorder_label,*++tokv); tokc--;
    } else {
      IError("option not understood: \"%s\"",tokv[0]);
      return;
    }
    tokv++;
    tokc--;
  }
  if (ord==NULL) ord = ((int*)calloc(current_el->n_examples,sizeof(int)));
  for_each(i,current_el->n_examples) ord[i] = atoi(*tokv++);
  if (reorder_label[0]) fprintf(dout,"%s reordering\n",reorder_label);
}

void
  NormalizePatterns(pats,n_pats,n_elts)
real **pats;
int n_pats,n_elts;
{
  int i,j;
  real s, vmax = -((real)HUGE), vmin = ((real)HUGE);
  if (display_scale) {
    for_each (i,n_pats)
      for_each (j,n_elts) {
	vmin = min(vmin,pats[i][j]);
	vmax = max(vmax,pats[i][j]);
      }
    if (vmin*vmax<0) {		/* opp sign: maintain 0.0=>0.0 */
      vmax = max(-vmin,vmax);
      vmin = -vmax;
    }
    printf("Scaling values between %f and %f\n",vmin,vmax);
    s = (display_max-display_min)/(vmax-vmin);
    for_each (i,n_pats)
      for_each (j,n_elts)
	pats[i][j] = s*(pats[i][j]-vmin) + display_min;
  } else {
    for_each (i,n_pats)
      for_each (j,n_elts) {
	vmin       = max(display_min,pats[i][j]);
	pats[i][j] = min(display_max,vmin);
      }
  }
}

void
  DisplayPatterns(pats,n_pats,n_elts,pat_min,elt_min)
real **pats;
int n_pats,n_elts,pat_min,elt_min;
{
  int i,j,lw = display_label_width;
  char *com;
  real v;
  NormalizePatterns(pats,n_pats,n_elts);
  fprintf(dout,"width %d\n",n_elts+lw);
  /* COLUMN NUMBERS */
  fprintf(dout,"window 1 %d Column numbers 1\n",n_elts+lw); 
  fprintf(dout,"s 0 %d ",lw+max(0,9-elt_min));
  for_each (i,n_elts) 
    if (i+1+elt_min>=10) {
      if (!display_single) fprintf(dout," ");
      fprintf(dout,"%d",((int)((i+1+elt_min)/10)));
    }
  fprintf(dout,"\nwindow 1 %d Column numbers 2\n",n_elts+lw); 
  fprintf(dout,"\ns 0 %d ",lw);
  for_each (i,n_elts) {
    if (!display_single) fprintf(dout," ");
    fprintf(dout,"%d",(i+1+elt_min)%10);
  }
  fprintf(dout,"\n");
  /* ROW LABEL AND VALUES */
  for_each (j,n_pats) {
    fprintf(dout,"window 1 %d Row %d\n",n_elts+lw,j); 
    fprintf(dout,"s %d 0 %-5s%2d\n",
	    0,   /* j+2, */
	    current_el->index[ord[j]+pat_min]->comment,
	    j+1+pat_min);
    for_each (i,n_elts)
      fprintf(dout,"b %d %d %.3f\n",0 /*j+2*/,i+lw,pats[ord[j]][ord[i]]);
  }
}

void
  PrintPatterns(pats,n_pats,n_elts,pat_min,elt_min)
real **pats;
int n_pats,n_elts,pat_min,elt_min;
{
  int i,j;
  NormalizePatterns(pats,n_pats,n_elts);
  fprintf(dout,"\t");
  for_each (i,n_elts) fprintf(dout," %6d",i+1+elt_min);
  fprintf(dout,"\n");
  for_each (j,n_pats) {
    PrintExampleName(current_el->index[ord[j]+pat_min]);
    fprintf(dout,"%d\t",j+1+pat_min);
    for_each (i,n_elts) 
      fprintf(dout," %6.4f",pats[ord[j]][ord[i]]);
    fprintf(dout,"\n");
  }
}
 


/******************************************************************************
 * command_states
 */

int
  command_states(tokc,tokv)
int tokc;
char *tokv[];
{
  int i,j,ex_n,t,dl,du,seed=0;
  int tmin=0,tmax,time=0,exmin,exmax,eltmin=0,print=FALSE;
  int debug = FALSE, inputs = FALSE, targets = FALSE, comp=FALSE;
  int n_pats,n_elts,metric = FALSE, display=FALSE, reorder = FALSE;
  int ssa = FALSE,smlayer=FALSE, n_inputs, n_outputs, n_ord;
  int use_dmin,use_dmax;
  real value,*vals,outprox,semprox;
  real **pats,**sim,**inpats,**outpats,**insim,**outsim,**sempats,**semsim;
  int example = -1, example_n = -1;
  char *com,*example_str = NULL;
  int *slayer,sl_specified = FALSE, *dlayer;
#ifdef DBM_NETWORK
  int old_max_iterations;
  real ave_energy, ave_total_energy;
  int old_ave_energy_p = ave_energy_p;
#endif DBM_NETWORK
  struct RANGE *sunit_range;
  struct RANGE *dunit_range;
  EXAMPLE_CODE ex_status = EXAMPLE_OK;
  static struct OPTION options[] =
    {
      {OPTPRES|OPTINTARG,"t",		0},
      {OPTPRES|OPTINTARG,"tmin",	0},
      {OPTPRES|OPTINTARG,"tmax",  	0},
      {OPTSIGN,		"ex",		0},
      {OPTPRES|OPTINTARG,"exmin",	0},
      {OPTPRES|OPTINTARG,"exmax",  	0},
      {OPTPRES|OPTINTARG,"seed",  	0},
      {OPTPRES|OPTINTARG,"smlayer",  	0},
      {OPTSIGN|OPTREALARG,"dmin",  	0},
      {OPTSIGN|OPTREALARG,"dmax",  	0},
      {OPTSIGN,		"scale",	0},
      {OPTSIGN,		"inputs",	0},
      {OPTSIGN,		"targets",	0},
      {OPTSIGN,		"display",	0},
      {OPTSIGN,		"print",	0},
      {OPTSIGN,		"reorder",	0},
      {OPTSIGN,		"single",	0},
      {OPTPRES,         "ssa",		0},
      {OPTPRES,         "slayer",	0},
      {OPTPRES,         "sunit",	0},
      {OPTPRES,        	"dlayer",	0},
      {OPTPRES,        	"dunit",	0},
      {OPTPRES,        	"comp",		0},
      {OPTPRES,        	"ham",		0},
      {OPTPRES,        	"cos",		0},
      {OPTPRES,        	"cor",		0},
      {OPTPRES,        	"dst",		0},
      {OPTPRES,        	"debug",	0},
      {OPTLAST,NULL,0}
    };
  IUsage(" [%sex <int>] [%pexmin <int>] [%pexmax <int>] [%ssingle]\
 [%pslayer <range>] [%pdlayer <range>] [%pcomp] [%sdmin <real>]\
 [%pdunit <unit range>] [%pt [<int>]] [%ptmin <int>] [%ptmax <int>]\
 [%pham] [%pcos] [%pcor] [%pdst] [%sreorder] [%pseed <n>] [%pdebug]\
 [%sscale] [%sdisplay] [%sprint] [%psmlayer <int>]");
  if (GiveHelp(tokc)) {
    ISynopsis("print unit states of some network layers.");
    IHelp
      (IHelpArgs,
"Print the states of units in some layers of the network.",
"OPTIONS",
"%sex <n>       : show states for example n (%n supresses).",
"%sexmin <n>    : minimum example number.",
"%sexmax <n>    : maximum example number.",
"%smlayer <n>   : semantic layer.",
"%sinputs       : show inputs to units instead of states (default %n).",
"%stargets      : show targets of units instead of states (default %n).",
"%sscale        : scale to [0 1] (%n supresses)",
"%ssingle       : display in single blot format.",
"%sdisplay      : display states or similarity (in bboard format)",
"%sprint        : print states or similarity",
"%pt <n>          : show states for iteration n (all if not specified).",
"%ptmin <n>       : only show states for iteration n and after.",
"%ptmax <n>       : only show states for iteration n and before.",
"%pslayer <range> : source layers (only makes sense with %pinputs).",
"%pdlayer <range> : destination layers.",
"%pdunit <range>  : destination units.",
"%pdmin <real>    : minimum value to display (%n supresses)",
"%pclayer         : compare similarity matrix with that of input/output.",
"%pham	          : compute similarity matrix using hamming distance.",
"%pcos	          : compute similarity matrix using angle cosing.",
"%pcor	          : compute similarity matrix using correlation.",
"%pdst	          : compute similarity matrix using eqcuidean distance.",
"%preorder        : do (not) reorder similarity matrix.",
"%pssa            : use SSA to do reordering.",
"%pseed           : random number seed.",
       NULL);
    return;
  }
  if (!NetExists(net)) {
    IError("current net is null");
    return;
  }
  display_single = FALSE;
  exmin = 0; 
  exmax = current_el->n_examples-1;
  tmin = tmax = net->p->max_t;
  AllocDisplayRanges(net,&dlayer,&slayer,&dunit_range,&sunit_range);
  tokv++;	/* ignore command name */
  while (IsOption(tokv[0])) {
    if (OptionMatch(tokv[0],"slayer",options,&sl_specified) ||
	OptionMatch(tokv[0],"dlayer",options,NULL) ||
	OptionMatch(tokv[0],"sunit", options,NULL) ||
	OptionMatch(tokv[0],"dunit", options,NULL)) {
      if (!SetDisplayRanges(net,&tokv,dlayer,slayer,dunit_range,sunit_range))
	return;
      if (sl_specified) inputs = TRUE;
    } else if (OptionMatch(tokv[0],"t",options,&time)) {
      if (IIsInteger(tokv[1])) {
	tmin = tmax = atoi(*++tokv);
      } else tmin = 0;
    } else if (OptionMatch(tokv[0],"tmin",options,&time)) {
      if (!IExpectInteger(tokv[1])) return;
      tmin = atoi(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"tmax",options,&time)) {
      if (!IExpectInteger(tokv[1])) return;
      tmax = atoi(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"smlayer",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      smlayer = atoi(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"ex",options,&example)) {
      if (example) {
	example_n = (IIsInteger(tokv[1])?atoi(*++tokv):ExampleNumber(current_el,example_str= *++tokv));
	exmin = exmax = example_n;
	n_pats = 1;
	tokc--;
      } else {
	exmin = exmax = -1;
      }
    } else if (OptionMatch(tokv[0],"exmin",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      exmin = atoi(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"exmax",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      exmax = atoi(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"dmin",options,&use_dmin)) {
      if (use_dmin) {
	if (!IExpectNumber(tokv[1])) return;
	display_min = atof(*++tokv); tokc--;
      } else display_min = -((real)HUGE);
    } else if (OptionMatch(tokv[0],"dmax",options,&use_dmax)) {
      if (use_dmax) {
	if (!IExpectNumber(tokv[1])) return;
	display_max = atof(*++tokv); tokc--;
      } else display_max = ((real)HUGE);
    } else if (OptionMatch(tokv[0],"comp",options,&comp)) {
    } else if (OptionMatch(tokv[0],"inputs",options,&inputs)) {
    } else if (OptionMatch(tokv[0],"targets",options,&targets)) {
    } else if (OptionMatch(tokv[0],"display",options,&display)) {
    } else if (OptionMatch(tokv[0],"print",options,&print)) {
    } else if (OptionMatch(tokv[0],"scale",options,&display_scale)) {
    } else if (OptionMatch(tokv[0],"single",options,&display_single)) {
    } else if (OptionMatch(tokv[0],"ssa",options,&ssa)) {
    } else if (OptionMatch(tokv[0],"reorder",options,&reorder)) {
      reorder_label[0] = NULL;
      if (!reorder) for_each(i,current_el->n_examples) ord[i] = i;
    } else if (OptionMatch(tokv[0],"ham",options,NULL)) {
      metric = SIM_HAM;
    } else if (OptionMatch(tokv[0],"cos",options,NULL)) {
      metric = SIM_COS;
    } else if (OptionMatch(tokv[0],"cor",options,NULL)) {
      metric = SIM_COR;
    } else if (OptionMatch(tokv[0],"dst",options,NULL)) {
      metric = SIM_DST;
    } else if (OptionMatch(tokv[0],"debug",options,&debug)) {
    } else if (OptionMatch(tokv[0],"seed",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      srandom(atoi(*++tokv)); tokc--;
    } else {
      IError("option not understood: \"%s\"",tokv[0]);
      return;
    }
    tokv++;
    tokc--;
  }
  if (debug) {
    PrintDisplayRanges(net,dlayer,slayer,dunit_range,sunit_range);
    return;
  }
  n_pats = exmax - exmin + 1;
  n_inputs  = net->layer[np->input_layer_n].n_units;
  n_outputs = net->layer[np->output_layer_n].n_units;
  for_each (i,net->n_layers) if (dlayer[i]) n_elts = net->layer[i].n_units;
  pats = ((real**)calloc_2d(n_pats,n_elts,sizeof(real)));
  n_ord = max(n_pats,n_elts);
  if (ord==NULL || !reorder) {
    ord = ((int*)calloc(n_ord,sizeof(int)));
    for_each(i,n_ord) ord[i] = i;
  }
#ifdef DBM_NETWORK
  old_max_iterations = np->max_iterations;
#endif DBM_NETWORK
  EnsureELIndex(current_el);
  ResetExamples(net);
  if (comp) {
    inpats  = ((real**)calloc_2d(n_pats,n_inputs,sizeof(real)));
    outpats = ((real**)calloc_2d(n_pats,n_outputs,sizeof(real)));
    for_each (j,n_pats) {
      for_each (i,n_inputs) 
	inpats[j][i]  = ScaleOutput(current_el->index[j+exmin]->value[i]);
      for_each (i,n_outputs) 
	outpats[j][i]  = ScaleOutput(current_el->index[j+exmin]->next->value[i]);
    }
    insim  = SimilarityMatrix(inpats,n_pats,n_inputs,metric);
    outsim = SimilarityMatrix(outpats,n_pats,n_outputs,metric);
  }
  if (smlayer) sempats = ((real**)calloc_2d(n_pats,net->layer[smlayer].n_units,sizeof(real)));
  if (metric)  sim     = ((real**)calloc_2d(n_pats,n_pats,sizeof(real)));
  /*
   * For each iteration...
   */
  for (t=tmin; t<=tmax; t++) {
    /*
     * Run examples and save states
     */
#ifdef DBM_NETWORK
    if (time) np->max_iterations = t;
    if (comp) {ave_energy=0.0; ave_total_energy=0.0; ave_energy_p=TRUE;}
#endif DBM_NETWORK
    for (ex_n=exmin; ex_n<=exmax; ex_n++) {
      if (ex_n != -1) {
	if (!LoadExample(net,ex_n,NULL)) return;
	if (!targets && tmax>0) PerformNet(net);
      } 
      if (smlayer && t==tmin)
	DoLayer(smlayer) {
	  ForEachUnit(UNIT) {
	    sempats[ex_n-exmin][unit_n] = UNIT_TARGET_T(unit,t);
	  } EndForEach;
	} EndDo;
      for (dl=0; dl<net->n_layers; dl++)
	if (dlayer[dl]) {
#ifdef DBM_NETWORK
	  if (comp) {
	    ave_energy       += Energy(net,FALSE,dl);
	    ave_total_energy += Energy(net,FALSE,FALSE);
	  }
#endif DBM_NETWORK
	  if (inputs) SetUnitInputsFromLayers(net,dl,slayer,t);
	  DoLayer(dl) {
	    for (du=dunit_range[dl].lo; du<=dunit_range[dl].hi; du++)
	      DoUnit(UNIT,du) {
		value = (inputs ? unit->x : (targets ? UNIT_TARGET_T(unit,t) : UNIT_OUTPUT_T(unit,t)));
		pats[ex_n-exmin][du] = value;
	      } EndDo;
	  } EndDo;
	} /* dlayer */
    } /* example */
    /*
     * Compute similarities
     */
    if (metric) {
      FillSimilarityMatrix(sim,pats,n_pats,n_elts,metric);
      if (reorder) {
	if (ssa) SSA_1D(sim,n_pats);
	else     ReorderSimilarityMatrix(sim,n_pats);
      }
      if (comp) {
	if (time) fprintf((display?stdout:dout),"%3d",t);
	fprintf((display?stdout:dout),"     %f(%f)", 
		matrix_correlation(sim,insim,n_pats),
		matrix_correlation_nodiag(sim,insim,n_pats));
	if (smlayer) {
	  if (t==tmin) semsim = SimilarityMatrix(sempats,n_pats,net->layer[smlayer].n_units,metric);
	  fprintf((display?stdout:dout),"     %f", 
		  matrix_correlation(sim,semsim,n_pats),
		  matrix_correlation_nodiag(sim,semsim,n_pats));
	  if (dlayer[smlayer])
	    fprintf((display?stdout:dout)," %f",
		    matrix_ave_cos_angle(pats,sempats,n_pats,net->layer[smlayer].n_units));
	} /* smlayer */
	fprintf((display?stdout:dout),"     %f(%f)",
		matrix_correlation(sim,outsim,n_pats),
		matrix_correlation_nodiag(sim,outsim,n_pats));
	if (dlayer[np->output_layer_n])
	  fprintf((display?stdout:dout)," %f",
		  matrix_ave_cos_angle(pats,outpats,n_pats,net->layer[np->output_layer_n].n_units));
#ifdef DBM_NETWORK
	fprintf((display?stdout:dout),"     %f %f",
		ave_energy/((real)(exmax-exmin+1)),
		ave_total_energy/((real)(exmax-exmin+1)));
#endif DBM_NETWORK
	fprintf((display?stdout:dout),"\n");
	fflush((display?stdout:dout));
      } /* comp */
    } /* metric */
    /*
     * Display/print states/similarities
     */
    if (display) {
      if (!metric) {
	ord = ((int*)calloc(n_ord,sizeof(int)));
	for_each(i,n_ord) ord[i] = i;
      }
      /* print header */
      if (t==tmin) fprintf(dout,"clear\n");
      fprintf(dout,"name %s\nstdscr\ns -2 0 Network \"%s\"",net->name,net->name);
      if (sl_specified) {
	for_each (i,net->n_layers) if (slayer[i]) fprintf(dout," %d",i);
	fprintf(dout," ==>");
      } else fprintf(dout," Layer");
      for_each (i,net->n_layers) if (dlayer[i]) fprintf(dout," %d",i);
      if      (targets) fprintf(dout," Targets");
      else if (inputs)  fprintf(dout," Inputs");
      else              fprintf(dout," States");
      if (tmin==tmax && tmax != net->p->max_t) fprintf(dout,"[%d]",tmax);
      if (metric) {
	if (reorder_label[0]) fprintf(dout," %s",reorder_label);
	fprintf(dout," %s Similarity",MetricName(metric));
      }
      fprintf(dout,"\nnoboxmarks\n");
      if (metric) DisplayPatterns(sim,n_pats,n_pats,exmin,exmin);
      else        DisplayPatterns(pats,n_pats,n_elts,exmin,eltmin);
      fprintf(dout,"newrow\ns -1 0 b\nend\n");
    } else if (print) {
      if (metric) PrintPatterns(sim,n_pats,n_pats,exmin,exmin);
      else        PrintPatterns(pats,n_pats,n_elts,exmin,eltmin);
    }				/* display */
  } /* iterations */
#ifdef DBM_NETWORK
  if (time) np->max_iterations = old_max_iterations;
  ave_energy_p = old_ave_energy_p;
#endif DBM_NETWORK
}


/******************************************************************************
 * command_matrix
 */
int
  command_matrix(tokc,tokv)
int tokc;
char *tokv[];
{
  int i,j,m,count,n_pats,n_mats=1;
  int print=FALSE, display=FALSE, diagonal=TRUE, logscale=FALSE;
  int use_dmin,use_dmax,diag_max=0,off_max=0;
  real **pats;
  char name[10];
  FILE *fp;
  static struct OPTION options[] =
    {
      {OPTSIGN|OPTREALARG,"dmin",  	0},
      {OPTSIGN|OPTREALARG,"dmax",  	0},
      {OPTPRES|OPTINTARG,"nmatrices",  	0},
      {OPTPRES,		"log",		0},
      {OPTSIGN,		"display",	0},
      {OPTSIGN,		"diagonal",	0},
      {OPTSIGN,		"print",	0},
      {OPTSIGN,		"single",	0},
      {OPTSIGN,		"scale",	0},
      {OPTLAST,NULL,0}
    };
  IUsage("<file> [%pnmatrices <int>] [%ssingle] [%sdmin <real>]\
 [%sdmax <real>] [%sdisplay] [%sdiagonal] [%sprint]\
 [%sscale] [%plog]");
  if (GiveHelp(tokc)) {
    ISynopsis("read in and print/display matrix(s).");
    IHelp
      (IHelpArgs,
"Print/display matrix(s)" ,
"OPTIONS",
"file          : file to read matrix(s) from",
"%pnmatrices <int>: number of matrices to add (default 1)",
"%ssingle       : display in single blot format.",
"%sdiagonal     : normalize correct (diagonal) values (default %y).",
"%sscale        : scale between 0 and 1.",
"%sdisplay      : display states or similarity (in bboard format)",
"%sprint        : print states or similarity",
"%plog            : use logarithmic scaling of values.",
"%pdmin <real>    : minimum value to display (%n supresses)",
"%pdmax <real>    : maximum value to display (%n supresses)",
       NULL);
    return;
  }
  EnsureELIndex(current_el);
  display_scale  = FALSE;
  display_single = FALSE;
  tokv++;	/* ignore command name */
  if ((fp=fopen(*tokv,"r"))==NULL) {
    IError("cannot open %s",*tokv); return;}
  tokv++;  tokc--;
  while (IsOption(tokv[0])) {
    if (OptionMatch(tokv[0],"dmin",options,&use_dmin)) {
      if (use_dmin) {
	if (!IExpectNumber(tokv[1])) return;
	display_min = atof(*++tokv); tokc--;
      } else display_min = -((real)HUGE);
    } else if (OptionMatch(tokv[0],"dmax",options,&use_dmax)) {
      if (use_dmax) {
	if (!IExpectNumber(tokv[1])) return;
	display_max = atof(*++tokv); tokc--;
      } else display_max = ((real)HUGE);
    } else if (OptionMatch(tokv[0],"nmatrices",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      n_mats = atoi(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"display",options,&display)) {
      display_scale  = TRUE;
      display_min    = 0;
      display_max    = 1;
    } else if (OptionMatch(tokv[0],"print",options,&print)) {
    } else if (OptionMatch(tokv[0],"log",options,&logscale)) {
    } else if (OptionMatch(tokv[0],"single",options,&display_single)) {
    } else if (OptionMatch(tokv[0],"scale",options,&display_scale)) {
    } else if (OptionMatch(tokv[0],"diagonal",options,&diagonal)) {
    } else {
      IError("option not understood: \"%s\"",tokv[0]);
      return;
    }
    tokv++;
    tokc--;
  }
  n_pats = current_el->n_examples;
  pats = ((real**)calloc_2d(n_pats,n_pats,sizeof(real)));
  if (ord==NULL) {
    ord = ((int*)calloc(n_pats,sizeof(int)));
    for_each (i,n_pats) ord[i] = i;
  }
  for_each (i,n_pats) for_each (j,n_pats) pats[i][j] = 0;
  for_each (m,n_mats)
    for_each (i,n_pats) {
      if (fscanf(fp,"%s",name)!=1) {IError("error reading matrix header %d",i); return;}
      for_each (j,n_pats) {
	if (fscanf(fp,"%d",&count)!=1) {IError("error reading matrix entry [%d][%d]",i,j); return;}
	pats[i][j] += count;
      }
    }
  fclose(fp);
  if (diagonal) {
    for_each (i,n_pats)
      for_each (j,n_pats) 
	if (i==j) diag_max = max(pats[i][j],diag_max);
	else      off_max  = max(pats[i][j],off_max);
    printf("Maximum diagonal=%d   off-diagonal=%d\n",diag_max,off_max);
    for_each (i,n_pats) pats[i][i] *= ((float)off_max)/((float)diag_max);
  }
  if (logscale)
    for_each (i,n_pats) 
      for_each (j,n_pats) 
	if (pats[i][j]>0)
	  pats[i][j] = log(pats[i][j]) ;
  if (display) {
    fprintf(dout,"clear\n");
    fprintf(dout,"name %s\nstdscr\ns -2 0 Network \"%s\" Confusion Matrix\n",net->name,net->name);
    fprintf(dout,"noboxmarks\n");
    DisplayPatterns(pats,n_pats,n_pats,0,0);
    fprintf(dout,"newrow\ns -1 0 b\nend\n");
  } else if (print) PrintPatterns(pats,n_pats,n_pats,0,0);
}
