#include <itf/option.h>
#include <itf/itf.h>
#include <sim/sim.h>
#include <lib/example.h>
#include "example-bp.h"
#include "prop-bp.h"

#define DB(x) x

#ifdef DBM_NETWORK
extern real FreeEnergy();	/* prop-dbm.c */
extern real Energy();		/* prop-dbm.c */
extern real Entropy();		/* prop-dbm.c */
#endif DBM_NETWORK

extern real output_one_minus_cos_angle();	/* perform-bp.c */
extern dist, gap ;				/* perform-bp.c */

int therapy	  = FALSE;
int therapy_proto = FALSE;
int therapy_sim   = FALSE;
int therapy_details = TRUE;
int therapy_split = FALSE;
int therapy_split_size = NULL;
int therapy_class = FALSE;

#define THERAPY_INPUT_AVERAGE	1
#define THERAPY_INPUT_EXTREME	2
#define THERAPY_MIXED		3
static real *therapy_prox = NULL;
static real *weights = NULL;

#define NOT_TRAINING(e)	(current_el->index[e]->training)
#define SAVE(e)		(current_el->index[e]->testing)
#define CLASS(e)	(current_el->index[e]->validation)
#define IS_ERROR(e)	(current_el->index[e]->ignore)

static int proto_word_n[5][8] = {
  {6,1,3,7,4,2,5,0},	/* indoor objects  */
  {1,7,6,3,2,5,0,4},	/* animals         */
  {7,6,3,5,0,1,2,4},	/* body parts      */
  {3,2,0,5,4,1,7,6},	/* foods           */
  {6,4,7,2,5,0,1,3}	/* outdoor objects */
};

static int input_ave_sim_word_n[40] = {
  33,24,32,15,18,25,23,12,9,31,30,16,38,14,26,20,40,1,36,8,28,22,35,39,17,3,11,34,6,19,13,37,7,4,27,5,10,2,29,21
  };
static int input_ext_sim_word_n[40] = {
  12,33,20,22,23,31,15,16,38,9,6,1,18,24,35,13,26,8,10,2,25,32,36,30,39,17,27,11,3,37,14,7,28,5,40,21,29,19,34,4
  };
static int mixed_sim_word_n[40] = {
  20,22,10,34,11,29,6,21,14,16,24,9,36,15,3,12,25,33,13,8,5,4,19,26,39,28,2,35,1,32,40,38,17,37,18,31,23,7,30,27
  };


/******************************************************************************
 * Therapy
 */

void
  SaveTherapyWeights(net)
NETWORK *net;
{
  if (weights==NULL) weights = ((real*)calloc(net->n_links,sizeof(real)));
  net->p->delta_bar_delta = FALSE;
  ForEachLinkCompiled(LINK,UNIT) {
    weights[link_n] = LINK_WEIGHT(link);
  } EndForEach;
}

void
  RestoreTherapyWeights(net)
NETWORK *net;
{
  fprintf(dout,"# Restoring Weights\n");
  ForEachLinkCompiled(LINK,UNIT) {
    LINK_WEIGHT(link) = weights[link_n];
    link->dEdw = 0.0;
  } EndForEach;
}

real
  TherapyWeightChange(net)
NETWORK *net;
{
  int n=0,sl,dl;
  real sum=0.0;
  ForEachLinkCompiled(LINK,UNIT) {
    sum += link->dEdw = fabs(LINK_WEIGHT(link)-weights[link_n]);
    n++;
  } EndForEach;
  fprintf(dout,"# Weight Change: total %f",sum/((real) n));
  for_each(dl,net->n_layers)
    DoLayer(dl) {
      for_each(sl,net->n_layers) {
	if (sl==0 && dl==0) continue;
	sum = 0.0;  n=0;
	ForEachUnit(UNIT) {
	  ForEachLinkSet(LINK,UNIT) {
	    if (OtherLayerN == sl) 
	      ApplyToLinks (if (!link->lesioned) {sum += link->dEdw; n++; link->dEdw = 0.0;});
	  } EndForEach;
	} EndForEach;
	if (n!=0) fprintf(dout," %d>%d %f",sl,dl,sum/((real) n));
      } /* sl */
    } EndDo;
  fprintf(dout,"\n");
}


void
  PrintTherapy()
{
  int e,c, n_err=0, n=0, n_classes=1;
  real ave_prox=0.0;
  for_each(e,current_el->n_examples)
    n_classes = max(n_classes,1+NOT_TRAINING(e));
  for_each(c,n_classes) {
    n=0; n_err=0; ave_prox=0.0; 
    for_each(e,current_el->n_examples)
      if (c==NOT_TRAINING(e)) {
	n++;
	ave_prox += therapy_prox[e];
	if (IS_ERROR(e)) n_err++;
      }
    fprintf(dout,"%s%d: %.3f %.3f ",
	    (c?"U":"T"),(c?c:therapy_class),(ave_prox/((float)n)),(n_err/((float)n)));
    if (therapy_details)
      for_each(e,current_el->n_examples) 
	if (c==NOT_TRAINING(e)) {
	  if (IS_ERROR(e)) fprintf(dout,"*");
	  PrintExampleName(current_el->index[e]);
	  fprintf(dout,"(%.2g) ",therapy_prox[e]);
	}
    fprintf(dout,"\n");
    fflush(dout);
  } /* n_classes */
}


void
  SetExamplePerformance(net)
NETWORK *net;
{
  int e,fail_gap=FALSE;
  int n_ex = current_el->n_examples;
  int ex_n = net->p->example_n;
  real ex_n_prox = output_one_minus_cos_angle(net,current_el->index[ex_n]->next->value);
  if (therapy_prox==NULL) therapy_prox = ((real*)calloc(n_ex,sizeof(real)));
  therapy_prox[ex_n] = ex_n_prox;
  for_each(e,n_ex)
    if (   e != ex_n
	&& gap > output_one_minus_cos_angle(net,current_el->index[e]->next->value)-ex_n_prox) {
      fail_gap = TRUE;
      break;
    }
  IS_ERROR(ex_n) = (ex_n_prox > dist || fail_gap);
}


void
  PrintExamplePerformance(net)
NETWORK *net;
{
  SetExamplePerformance(net);
  if (net->p->example_n == current_el->n_examples-1) PrintTherapy();
}

void
  PrintBinaryPerformance(net)
NETWORK *net;
{
  int e;
  fprintf(dout,"# Binary Performance:\n");
  for_each (e,current_el->n_examples) {
    if (!LoadExample(net,e,NULL)) return;
    PerformNet(net);
    ApplyToEachOutputUnit(UNIT_OUTPUT(unit) = UNIT_BINARY_OUTPUT(unit););
    PrintExamplePerformance(net);
  }
}

void
  ChoosePrototypeTherapy(net)
NETWORK *net;
/* What this does is divide errors and correct responses in each category
 * in half into trained and untrained sets, where (for both errors and correct
 * responses) selection as trained is biased as strongly as possible towards
 * the most prototypical words in the category.  Thus the trained and untrained
 * sets are balanced for correct performance.
 */
{
  int e,train;
  int cat,n_cats=5, word,n_words=8;
  int n_errs,train_n_errs,train_n_corr, train_extra_err=with_prob(0.5);
  fprintf(dout,"# Choosing Prototype Therapy\n");
  for_each(cat,n_cats) {
    n_errs = 0;
    for_each(word,n_words) 
      n_errs += IS_ERROR(cat*n_words+word);
    train_n_errs = n_errs/2;
    if (n_errs%2 && toggle(train_extra_err)) train_n_errs++;
    train_n_corr = n_words/2 - train_n_errs;
    for_each(word,n_words) {
      e = cat*n_words + proto_word_n[cat][word];
      NOT_TRAINING(e) = !(   ( IS_ERROR(e) && train_n_errs-- > 0)
		      || (!IS_ERROR(e) && train_n_corr-- > 0));
    } /* words */
  }
}

void
  ChooseSimilarityTherapy(net)
NETWORK *net;
/* This divides errors and correct response in half into trained and 
 * untrained sets where selection as trained is biased towards words whose
 * average input or mixed similarity with other words is highest.
 */
{
  int e,n,n_errs=0,train_n_errs,train_n_corr;
  int n_words = current_el->n_examples;
  fprintf(dout,"# Choosing %s Similarity Therapy\n",
	  (therapy_sim==THERAPY_INPUT_AVERAGE ? "Average Input" :
	   (therapy_sim==THERAPY_INPUT_EXTREME ? "Extreme Input" :
	    "Mixed")));	/* therapy_sim==THERAPY_MIXED */
  for_each (e,n_words)
    n_errs += IS_ERROR(e);
  train_n_errs = n_errs/2;
  if (n_errs%2 && with_prob(0.5)) train_n_errs++;
  train_n_corr = n_words/2 - train_n_errs;
  for_each (e,n_words) {
    n = (therapy_sim==THERAPY_INPUT_AVERAGE ? input_ave_sim_word_n[e]-1 :
	 (therapy_sim==THERAPY_INPUT_EXTREME ? input_ext_sim_word_n[e]-1 :
	  mixed_sim_word_n[e]-1));		/* therapy_sim==THERAPY_MIXED */

    NOT_TRAINING(e) = !(   ( IS_ERROR(n) && train_n_errs-- > 0)
		    || (!IS_ERROR(n) && train_n_corr-- > 0));
  }
}

void
  ChooseTherapy(net,all)
NETWORK *net;
int all;
{
  int e, therapy_correct=TRUE, therapy_incorrect=FALSE;
  for_each (e,current_el->n_examples) {
    if (all) {
      NOT_TRAINING(e) = FALSE;
    } else {
      if (!LoadExample(net,e,NULL)) return;
      PerformNet(net);
      SetExamplePerformance(net);
      if (IS_ERROR(e)) { 		/* 0=corr  1=err */
	NOT_TRAINING(e) = therapy_correct;
	toggle(therapy_correct);
      } else {
	NOT_TRAINING(e) = therapy_incorrect;
	toggle(therapy_incorrect);
      }
    }
  }
  if (!all) {
    if      (therapy_proto) ChoosePrototypeTherapy(net);
    else if (therapy_sim)   ChooseSimilarityTherapy(net);
  }
  fprintf(dout,"# Trained:   ");
  if (all) {
    fprintf(dout,"all\n");
  } else {
    for_each (e,current_el->n_examples) 
      if (!NOT_TRAINING(e)) {
	PrintExampleName(current_el->index[e]);
	fprintf(dout," ");
      }
    fprintf(dout,"\n");
    fprintf(dout,"# Untrained: ");
    for_each (e,current_el->n_examples)
      if (NOT_TRAINING(e)) {
	PrintExampleName(current_el->index[e]);
	fprintf(dout," ");
      }
    fprintf(dout,"\n");
  }
}

void
  SwapTherapy(net)
NETWORK *net;
{
  int e;
  fprintf(dout,"# Swapping Therapy\n");
  for_each (e,current_el->n_examples) 
    NOT_TRAINING(e) = !NOT_TRAINING(e);
}

void
  SaveTherapy(net)
NETWORK *net;
{
  int e;
  for_each (e,current_el->n_examples) SAVE(e) = NOT_TRAINING(e);
}

void
  RestoreTherapy(net)
NETWORK *net;
{
  int e;
  for_each (e,current_el->n_examples) NOT_TRAINING(e) = SAVE(e);
}

void
  DoSplitTherapy(net)
NETWORK *net;
{
  int e,i,j,n,cat,word;
  int n_words=8, n_cat_train=n_words/2;
  for (i=1; i<n_cat_train; i++)
    for (j=i+1; j<=n_cat_train; j++) {
      /* RESTORE WEIGHTS */
      RestoreTherapyWeights(net);
      /* SET TRAINING */
      for_each (e,current_el->n_examples) {
	cat  = e / n_words;
	word = e % n_words;
	if (word==0) n = 0;
	if (NOT_TRAINING(e)==0) {
	  n++;
	  if (n==i || n==j) NOT_TRAINING(e) = 2;
	}
      }
      /* DO TRAINING */
      net->p->epoch_n = 0;
      fprintf(dout,"# Split Therapy: %d %d\n",i,j);
      EpochTrain(net);
      for_each (e,current_el->n_examples) 
	if (NOT_TRAINING(e)==2) NOT_TRAINING(e) = 0;
    }
}

void
  DoTherapy(net)
NETWORK *net;
{
  net->p->epoch_n = 0;
  if (!net->p->stop_at_epoch) net->p->stop_at_epoch = 50;
  fprintf(dout,"# Therapy: %d epochs\n", net->p->stop_at_epoch);
  if (therapy_split) DoSplitTherapy(net);
  else               EpochTrain(net);
}

void 
  PrintClassDiff(n_cl, ex_init_prox, ex_init_err)
int n_cl;
real *ex_init_prox, *ex_init_err;
{
  int e, n, cl;
  real init_prox, prox, init_err, n_err;

  for (cl=0; cl<=n_cl; cl++) {
    init_prox = prox = init_err = n_err = n = 0;
    for_each (e, current_el->n_examples)
      if (NOT_TRAINING(e)==cl) {
	init_prox += ex_init_prox[e]; init_err += ex_init_err[e];
	prox += therapy_prox[e]; n_err += IS_ERROR(e); n++;
      }
    fprintf(dout,"DIFF T%d => %s%d: %.3f %.3f\n",
	    therapy_class, (cl?"U":"T"),(cl?cl:therapy_class),
	    (init_prox - prox)/((float)n),
	    (init_err  - n_err)/((float)n));
  }
}

void
  DoClassTherapy(net)
NETWORK *net;
{
  int e, cl, n;
  int n_ex = current_el->n_examples ;
  int n_cl = n_ex/therapy_split_size ;
  int train_cor, train_err;
  real prox, n_err;
  real *ex_init_prox = ((real*)calloc(n_ex,sizeof(real)));
  real *ex_init_err  = ((real*)calloc(n_ex,sizeof(real)));

  therapy_split = FALSE;	/* so we can still use DoTherapy */

  SaveTherapyWeights(net);
  for_each (e, n_ex) {
    if (!LoadExample(net, e, NULL)) return;
    PerformNet(net);
    SetExamplePerformance(net);
    ex_init_prox[e] = therapy_prox[e];
    ex_init_err[e]  = IS_ERROR(e);
    CLASS(e) = 1 + e/therapy_split_size;
  }
  /* FOR EACH CLASS.... */
  for (cl=1; cl<=n_cl; cl++) {
    therapy_class = cl;
    fprintf(dout,"# Therapy Class %d: ", cl);
    /* SET HALF OF CLASS TO BE TRAINED, REST NOT */
    train_cor = TRUE;  train_err = FALSE;
    for_each (e, n_ex) {
      NOT_TRAINING(e) = CLASS(e);
      if (CLASS(e) == cl) {
	if (IS_ERROR(e)) {
	  if (train_err) NOT_TRAINING(e) = FALSE; toggle(train_err);
	} else {
	  if (train_cor) NOT_TRAINING(e) = FALSE; toggle(train_cor);
	}
	if (!NOT_TRAINING(e)) fprintf(dout," %d",e);
      }
    }
    fprintf(dout,"\n");
    DoTherapy(net);
    PrintClassDiff(n_cl, ex_init_prox, ex_init_err);
    TherapyWeightChange(net);
    RestoreTherapyWeights(net);
    /* SWAP THERAPY WITHIN CLASS */
    fprintf(dout,"# Therapy Class %d: ", cl);
    for_each (e, n_ex) 
      if (CLASS(e) == cl) {
	NOT_TRAINING(e) = (NOT_TRAINING(e) ? 0 : cl);
	if (!NOT_TRAINING(e)) fprintf(dout," %d",e);
      }
    fprintf(dout,"\n");
    DoTherapy(net);
    PrintClassDiff(n_cl, ex_init_prox, ex_init_err);
    TherapyWeightChange(net);
    RestoreTherapyWeights(net);
  }
  for_each(e, n_ex) NOT_TRAINING(e) = FALSE;
  fprintf(dout,"# Trained: all\n");
  therapy_class = FALSE;
  DoTherapy(net);
  PrintClassDiff(0, ex_init_prox, ex_init_err);
  TherapyWeightChange(net);
  RestoreTherapyWeights(net);
  free(ex_init_err);
  free(ex_init_prox);
}

void
  PerformTherapy(net)
NETWORK *net;
{
  if (therapy_split_size)
    DoClassTherapy(net);
  else {
    SaveTherapyWeights(net);
    if (therapy==-1) SaveTherapy(net);
    else             ChooseTherapy(net,FALSE);
    PrintBinaryPerformance(net);
    DoTherapy(net);
    TherapyWeightChange(net);
    RestoreTherapyWeights(net);
    SwapTherapy(net);
    DoTherapy(net);
    TherapyWeightChange(net);
    RestoreTherapyWeights(net);
    SwapTherapy(net);
    ChooseTherapy(net,TRUE);	/* train on all */
    /* next 3 lines replace DoTherapy(net) to avoid DoSplitTherapy(net) */
    net->p->epoch_n = 0;
    fprintf(dout,"# Therapy: %d epochs\n", net->p->stop_at_epoch);
    EpochTrain(net);
    TherapyWeightChange(net);
    RestoreTherapyWeights(net);
    if (therapy==-1) RestoreTherapy(net);
  }
}


/*****************************************************************************
 * command_set_therapy()
 */

int
  command_set_therapy(tokc,tokv)
int tokc;
char *tokv[];
{ 
  int e;
  char *ex_str=NULL;
  IUsage("");
  if (GiveHelp(tokc)) {
    ISynopsis("set examples to be trained on during therapy.");
    IHelp(IHelpArgs,
	  NULL);
    return;
  }
  if (current_el == NULL) {IError("current example set is NULL");  return;}
  EnsureELIndex(current_el);
  fprintf(dout,"# Therapy: ");
  if (tokc==1) {
    for_each (e,current_el->n_examples) 
      NOT_TRAINING(e) = FALSE;
    fprintf(dout,"all");
  } else {
    for_each (e,current_el->n_examples) 
      NOT_TRAINING(e) = TRUE;
    while (--tokc > 0) {
      e = (IIsInteger(tokv[1]) ? atoi(*++tokv) : ExampleNumber(current_el,ex_str= *++tokv));
      if (e<0 || e >= current_el->n_examples) {
	fprintf(dout,"(No example %d %s)",e,ex_str);
	continue;
      }
      /* training=FALSE (0) actually means it WILL be trained on */
      NOT_TRAINING(e) = FALSE;
      PrintExampleName(current_el->index[e]);
      fprintf(dout," ");
    }
  }
  fprintf(dout,"\n");
  net->p->therapy = TRUE;
}
