/*
 * $Id: blot.c,v 1.1 90/12/07 11:50:11 drew Exp $
 * 
 * $Log:	blot.c,v $
# Revision 1.1  90/12/07  11:50:11  drew
# Initial revision
# 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <itf.h>

static int blot_range_chars = 48;
static int blot_size = 2;	/* 2 characters for a blot */
static int number_size = 3;
static int number_of_blots = 24;	/* number of different sized blots */
static int blot_invert = 0;	/* swap +/- colors ? */
static int blot_first_char = 33;/* starting character of blots - can't use
				   32 as starting char because Xterm can't
				   handle char 32 in bold as first non-blank
				   character on ine */
static int blot_format = B_TEXT;
static char area1[200];		/* used to store mode strings */
static char *bold_on;
static char *attr_off;
static char *sn();

static char blot_buffer[200];
static char blot_buffer2[20];

int autoinit_blot()
{
  char bp[1024];
  char *area=area1;
  char *env;
  int no_blots = 0;
  extern char *tgetstr();
  extern char *getenv();
  if (getenv("TERM")==NULL) {
    warn(NULL,"\"TERM\" environment variable not set - no blots\n");
    attr_off = "";
    bold_on = "";
    no_blots = 1;
  } else {
    tgetent(bp,getenv("TERM"));
    bold_on = tgetstr("md",&area);
    attr_off = tgetstr("me",&area);
    if (bold_on==NULL || attr_off==NULL) {
      warn(NULL,"Terminal (type \"%s\") cannot do bold characters - no blots",getenv("TERM"));
      attr_off = "";
      bold_on = "";
      no_blots = 1;
    }
    if (strchr(bold_on,'$')) *strchr(bold_on,'$') = 0; /* kill the delay */
    if (strchr(attr_off,'$')) *strchr(attr_off,'$') = 0; /* kill the delay */
    IBindInt("invert-blots",&blot_invert,P_OBJECT,0);
    IBindInt("number-of-blots",&number_of_blots,P_OBJECT,0);
    IBindInt("blot-size",&blot_size,P_OBJECT,0);
    if (!no_blots && (env = getenv("BLOT")) && (atoi(env)!=0)) {
      blot_size = atoi(env);
      blot_invert = 0;
      if (blot_size < 0) {
	blot_invert = 1;
	blot_size = - blot_size;
      }
      number_of_blots = blot_range_chars/blot_size;
      blot_format = B_BOLD;
    } else {
      if (bold_on==NULL || *bold_on==0 || atoi(default_blot)==0) {
	blot_format = B_NUMBER;
	number_size = 5;
	blot_size = 2;
      } else {
	/* if (getenv("BLOT")==NULL)
	  warn(NULL,"Environment BLOT not set - will try to use blots anyway");
	  */
	blot_format = B_BOLD;
	number_size = 5;
	blot_size = atoi(default_blot);
	blot_invert = 0;
	if (blot_size < 0) {
	  blot_invert = 1;
	  blot_size = - blot_size;
	}
	number_of_blots = blot_range_chars/blot_size;
      }
    }
  }
}

void ISetDefaultStreamBlot(stream)
struct STREAM *stream;
{
  stream->blot.range_chars = blot_range_chars;
  stream->blot.size = blot_size;
  stream->blot.number_size = number_size;
  stream->blot.number = number_of_blots;
  stream->blot.invert = blot_invert;
  stream->blot.first_char = blot_first_char;
  stream->blot.format = blot_format;
}

static int BlotSize()
{
  if (current_outstream)
    return current_outstream->blot.size;
  else
    return blot_size;
}

static int NumberSize()
{
  if (current_outstream)
    return current_outstream->blot.number_size;
  else
    return number_size;
}

/*
static int BlotRangeChars()
{
  if (current_outstream)
    return current_outstream->blot.range_chars;
  else
    return blot_range_chars;
}
*/

static int NumberOfBlots()
{
  if (current_outstream)
    return current_outstream->blot.number;
  else
    return number_of_blots;
}

static int BlotInvert()
{
  if (current_outstream)
    return current_outstream->blot.invert;
  else
    return blot_invert;
}

static int BlotFirstChar()
{
  if (current_outstream)
    return current_outstream->blot.first_char;
  else
    return blot_first_char;
}

static int BlotUseNumbers()
{
  if (current_outstream)
    return current_outstream->blot.format == B_NUMBER;
  else
    return blot_format == B_NUMBER;
}

static int BlotUseText()
{
  if (current_outstream)
    return current_outstream->blot.format == B_TEXT;
  else
    return blot_format == B_TEXT;
}

static int BlotUseBold()
{
  if (current_outstream)
    return current_outstream->blot.format == B_BOLD;
  else
    return blot_format == B_BOLD;
}

int BlotLen()
{
  if (BlotUseNumbers())
    return NumberSize();
  else
    return BlotSize();
}

char *BlotStart()
{
  if (BlotUseBold())
    return bold_on;
  else if (BlotUseText())
    return "\\b{";
  else
    return "";
}

char *BlotEnd()
{
  if (BlotUseBold())
    return attr_off;
  else if (BlotUseText())
    return "}";
  else
    return "";
}

char *RawBlot(x)
real x;
{
  int i,c,sign;
  if (BlotUseNumbers()) {
    if (NumberSize() > 3) {
      sprintf(blot_buffer," %*.2f",NumberSize()-1,x);
      return blot_buffer;
    } else {
      return sn(x);
    }
  } else if (BlotUseText()) {
    sprintf(blot_buffer," %.2f",x);
    return blot_buffer;
  } else {
    i = NumberOfBlots()*x;
    if (i>=0)
      sign=0;
    else {
      sign=1;
      i = -i;
    }
    i++;
    if (i>NumberOfBlots()) i=NumberOfBlots();
    c = BlotFirstChar() + (NumberOfBlots()-i)*BlotSize();
    /* be careful - character 127 doesn't work well, so use zero from the lo range */
    if (!(sign ^ BlotInvert()) && i>1) {
      c += NumberOfBlots()*BlotSize();
    }
    if (c==127) c-= blot_size; /* nasty hack -c=127 happens when blot_size==1 */
    for (i=0;i<BlotSize();i++)
      blot_buffer[i] = c+i;
    blot_buffer[i]=0;
    /* sprintf(blot_buffer,"%3d ",c); */
    return blot_buffer;
  }
}

char *Blot(x)
real x;
{
  strcpy(blot_buffer2,BlotStart());
  strcat(blot_buffer2,RawBlot(x));
  strcat(blot_buffer2,BlotEnd());
  return blot_buffer2;
}

command_blot_range(tokc,tokv)
int tokc;
char *tokv[];
{
  real x;
  if (GiveHelp(tokc)) {
    IUsage("");
    ISynopsis("print out the blots for the range -1 to +1");
    IHelp(IHelpArgs,
	  NULL);
    return;
  }
  
  fprintf(dout,"-1 %s",BlotStart());
  for (x= -1;x<1/2*(real)NumberOfBlots();x+=1/(real)NumberOfBlots())
    fprintf(dout,"%s",RawBlot(x));
  fprintf(dout,"%s 0\n",BlotEnd());
  fprintf(dout," 0 %s",BlotStart());
  for (x=0;x<1-1/2*(real)NumberOfBlots();x+=1/(real)NumberOfBlots())
    fprintf(dout,"%s",RawBlot(x));
  fprintf(dout,"%s 1\n",BlotEnd());
  return;
}

command_blot(tokc,tokv)
int tokc;
char *tokv[];
{
  if (GiveHelp(tokc)) {
    IUsage("<value1> <value2> ...");
    ISynopsis("print out the given values as blots");
    IHelp(IHelpArgs,
	  NULL);
    return;
  }

  fprintf(dout,"%s",BlotStart());
  while (*++tokv)
    fprintf(dout,"%s",RawBlot(atof(*tokv)));
  fprintf(dout,"%s\n",BlotEnd());
  return;
}

static char blot_env_str[10];

command_blot_reverse(tokc,tokv)
int tokc;
char *tokv[];
{
  struct STREAM *stream;
  IDoHelp
    (
     IUsage("[stream-name]");
     ISynopsis("reverse the colors of blots");
     IHelp(IHelpArgs,
	   "If no stream name is supplied, the color of the blots is reversed",
	   "for the standard output.  If a stream name is supplied, the color",
	   "of the blots is reversed for that particular stream.  When a new",
	   "stream is openned, it inherits the blot type of the standard",
	   "output.",
	   "USAGE WITH BBOARD",
	   "When you are sending output to bboard, the type of blot for the",
	   "stream you are sending on is completely ignored, because only",
	   "text and numbers are sent to bboard.  Bboard itself decides what",
	   "type of blot to use by looking at the BLOT environment variable.",
	   "The net-weights command can tell bboard to reverse the blots from",
	   "what it was using if you give it the '+reverse' flag.",
	   NULL);
     );

  if (tokc>2) {
    IExpected("the name of a stream");
    return;
  }
  if (tokc==2) {
    stream = (struct STREAM*)IGetObjectPtr("STREAM",tokv[1]);
    if (stream==NULL) {
      IError("\"%s\" is not a stream",tokv[1]);
      return;
    }
  } else
    stream = current_outstream;
  stream->blot.invert = !stream->blot.invert;
  if (stream == stdout_stream) {
    blot_invert = stream->blot.invert;
    sprintf(blot_env_str,"%s%d",(blot_invert?"-":""),blot_size);

    setenv("BLOT",blot_env_str,1);

    /* printf("environment BLOT=%s\n",getenv("BLOT")); */
  }
  /* fix here to change ENVIRONMENT blot */
  return;
}

command_blot_format(tokc,tokv)
int tokc;
char *tokv[];
{
  struct STREAM *stream;
  if (GiveHelp(tokc)) {
    IUsage("[stream] -number|[stream] -blot|[stream] -text");
    ISynopsis("tell the system the format of the blots");
    IHelp
      (IHelpArgs,
       "The blot format can be set separately for each stream.",
       "The default is the the standard out stream.  When a new stream",
       "is openned, it inherits the same blot format as the standard",
       "output stream currently has.",
       "There are three blot formats:",
       "  number: print as a number - specify precision with \"blot-size\"",
       "  blot:   print as a blot diagram - specify size with \"blot-size\"",
       "  text:   print as \"\\{N}\" can be interpreted later by other programs",
       NULL);
    return;
  }
  
  if (tokc!=3 && tokc!=2 && tokc!=1) {
    IExpected("the name of a stream");
    return;
  }
  if (tokc>=2 && tokv[1][0]!='-') {
    stream = (struct STREAM*)IGetObjectPtr("STREAM",tokv[1]);
    if (stream==NULL) {
      IError("\"%s\" is not a stream",tokv[1]);
      return;
    }
    tokv++;
  } else
    stream = current_outstream;
  if (tokv[1]==NULL) {
    fprintf(dout,"blot-format %s %s\n",stream==NULL?"":stream->name,
	    (stream->blot.format==B_BOLD?"-blot":
	     stream->blot.format==B_NUMBER?"-number":
	     stream->blot.format==B_TEXT?"-text":
	     "-illegal value"));
  } else {
    if (strprefix(tokv[1],"-number")) {
      stream->blot.format = B_NUMBER;
      if (stream->blot.number_size<3) stream->blot.number_size = 3;
    } else if (strprefix(tokv[1],"-blot")) {
      stream->blot.format = B_BOLD;
    } else if (strprefix(tokv[1],"-text")) {
      stream->blot.format = B_TEXT;
    } else {
      IError("illegal blot format");
      return;
    }
    if (stream == stdout_stream) {
      blot_format = stream->blot.format;
      number_size = stream->blot.number_size;
    }
  }
  return;
}

command_blot_size(tokc,tokv)
int tokc;
char *tokv[];
{
  int size;
  struct STREAM *stream;
  IUsage("[stream-name] <blot-size>");
  if (GiveHelp(tokc)) {
    ISynopsis("tell the interface what the size of the blots are");
    IHelp(IHelpArgs,
	  "The stream-name is optional, the default is stdout.",
	  NULL);
    return;
  }

  if (tokc<2 || tokc>3) {
    IError(IPrintUsage(tokv[0],usage));
    return;
  }
  if (tokc==3) {
    size = atoi(tokv[2]);
    stream = (struct STREAM*)IGetObjectPtr("STREAM",tokv[1]);
    if (stream==NULL) {
      IError("\"%s\" is not a stream",tokv[1]);
      return;
    }
  } else {
    size = atoi(tokv[1]);
    stream = current_outstream;
  }
  if (size<=0 || size>24) {
    IExpected("sensible non-zero size of blot");
    return;
  }
  if (stream->blot.format == B_NUMBER) {
    stream->blot.number_size = size;
    if (stream == stdout_stream) {
      number_size = size;
    }
  } else {
    stream->blot.size = size;
    stream->blot.number = stream->blot.range_chars/stream->blot.size;
    if (stream == stdout_stream) {
      blot_size = size;
      number_of_blots = blot_range_chars/blot_size;
    }
  }
  return;
}

static int sign(d)
real d;
{
  if (d<0.0) return -1;
  else return 1;
}
/*
 * sn(d) returns a three character representation of the floating point
 * number 'd'.  It just represents values between -1.0 & +1.0, it was
 * designed so that probability matrices could be printed compactly.
 * If a number is less than 0.1, but greater than or equal to 0.01, it is
 * represented as :n (double decimal point)
 * output       d
 *  ^^          > 1.0
 *  ++          = 1.0      100 %
 *  nn          = 0.nn     nn%
 *  .n          = 0.00n    n/10 of %
 *  :n          = 0.000n   n/100 of %
 */
static char *sn(d)
real d;
{
  char *res,c,p;
  int i;
  static char str[40];
  if (d < -1.0) res = " __";
  else if (d > 1.0) res = " ^^";
  else {
    res = str;
    i = 100 * (d + sign(d)*0.005);
    if ((i != 0) && (fabs(d) >= 0.0095)) {
      sprintf(str,"%3d",i);
      if (i == 100) res = " =1";
      if (i == -100) res = "=-1";
    }
    else {
      i = 1000 * (d + sign(d)*0.0005);
      p = '.';
      if (i==0 || fabs(d) < 0.00095) {p=':'; i = 10000 * (d+sign(d)*0.00005);}
      c = ' ';
      if (i < 0) { i = -i; c = '-'; }
      sprintf(str,"%c%c%1d",c,p,i);
      if (i == 0) res = "  0";
    }
  }
  return res;
}
