/*
 * $Id: itf.c,v 1.1.1.2 91/04/30 16:03:30 xerion Exp $
 * 
 * $Log:	itf.c,v $
 * Revision 1.1.1.2  91/04/30  16:03:30  xerion
 * Modified to interface with display.c (X windows interface).
 * 
 * Revision 1.1.1.1  91/04/29  21:17:36  xerion
 * changed fclose to pclose in read where a pipe is used.
 * 
 * Revision 1.1  90/12/07  11:50:37  drew
 * Initial revision
 * 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include "itf.h"
#include "display.h"

#define DB(X)

#define EVERYTHING	1
#define ONLY_DEFAULTS	2
#define EXCEPT_DEFAULTS	3

#define MAX_COMMAND_LEN 10000
#define ERROR_BUF_SIZE 1000
#define ERROR_BUF_SAFE 200

static int	XMode = 0 ;

PFI pre_command_function = NULL;
PFI post_command_function = NULL;
char *help_string;
extern int n_commands;
static char *prompt;
char *noclobber;
char *plus_minus_options;
static char *redirection;
char *automore;
static char *echo;
static char *read_echo;
static char *options_echo;
static char *use_default_options;
static long status;
static char *print_status;
static char *command_line1;
static char *command_line2;
static char *readpath;
static char *show_function_calls;
char *default_blot;
static int command_line_len;
int auto_more_enabled = 0;
static int command_id;	/* used in IError */
static struct STACK *function_names;
HASH_TABLE *itf_func_table;
int itf_func_context;
int itf_com_context;
HASH_TABLE *lex_table;
int op_context;	/* context for operators in the lex table */
char command_syntax[LEX_TABLE];
char object_syntax[LEX_TABLE];
char simple_syntax[LEX_TABLE];
char itf_error[ERROR_BUF_SIZE];
char itf_tmp_buf[1000];
int itf_errno;
int caught_signal = 0;
static enum {NORMAL, EXIT, QUIT} quit_flag;
static int just_do_defaults=EVERYTHING;
int source_level = 0;
FILE *din=stdin;
FILE *dout=stdout;
int *itf_abort_env;
char *myname;
int mypid;

int command_read();
int command_echo();
int command_quit();
int command_exit();
int command_default();

char *itf_error_str[TOP_ITF_ERRNO+1] = {
  "",
  "Command error: ",
  "Syntax error: ",
  "Lexical error: ",
  "Unknown command: ",
  "Unknown error: ",
  "Unknown variable: ",
  "Unknown structure: ",
  "Unknown field: ",
  "Null variable name",
  "Illegal variable name: ",
  "Illegal object: ",
  "Bad object syntax: ",
  "Array index out of bounds: ",
  "Need array index for ",
  "Need structure field for ",
  "Need structre field or array index for ",
  "Variable is readonly: ",
  "Bad var type: ",
  "Internal null pointer, happenned with: ",
  "Excess in object: ",
  "Attempt to dereference null pointer: ",
  "Expected an object - found nothing ",
  "Out of memory in ",
  "Interrupt",
  "System error signal sent and caught",
  "Variable name is already used: ",
  "Redirection I/O error: ",
  "Redirection I/O error: ",
  "",
  NULL,
};

void IAbort();
void ICatchInterrupt();
void ISegVioRescue();
void IBusErrorRescue();
void IFPErrorRescue();
void IPipeErrorRescue();
void IWindowChange();

IStandardInit(argv)
char *argv[];
{
  int	argc ;
  for (argc = 0 ; argv[argc] != NULL ; ++argc)
    ;
  XMode = initGraphics(&argc, argv) ;

  IInit(2000, MAX_COMMAND_LEN, MAX_COMMAND_LEN+2000, argv);
}

/***
 * IFunctionEnter(name)
 * Tell the interface about entering this function so that the function
 * name can be used in error messages.  Function names are kept on
 * a stack and popped off by ILeaveFunction()
 */
void IFunctionEnter(name)
char *name;
{
  if (function_names==NULL)
    function_names = TempStackNew();
  if (true(show_function_calls))
    fprintf(dout, "%*s%s\n", TempStackHeight(function_names)*2+2, ">>", name);
  TempStackPush(function_names, name);
}

void IFunctionExit(name)
char *name;
{
  char *last_name;
  last_name = TempStackPop(char *, function_names);
  if (strcmp(name, last_name))
    IErrorAbort("IFunctionExit: leaving \"%s\" but found \"%s\" on stack",
		name, last_name);
  if (true(show_function_calls))
    fprintf(dout, "%*s%s\n", TempStackHeight(function_names)*2+2, "<<", name);
}

char *IBaseName(str)
char *str;
{
  char *pos = str;
  if (str==NULL) return str;
  while ((pos=index(str, '/'))!=NULL) str = pos+1;
  return str;
}

static InitReadFile(va_alist)
va_dcl
{
  va_list args;
  char *fmt;
  char buf[200];
  struct stat statbuf;
  va_start(args);
  fmt = va_arg(args, char *);
  strcpy(buf, "read ");
  vsprintf(buf+5, fmt, args);
  if (stat(buf+5, &statbuf)==0) {
    if (!IDoCommandLine(buf)) {
      IReportError(stderr);
      IResetError();
      panic("IInit", "Error reading \"%s\"", buf+5);
    }
  }
}

IInit(max_symbols, max_command_len, lex_space, argv)
int max_symbols, max_command_len, lex_space;
char *argv[];
{
  char *space;
  char buf[200];

  /* some code depends upon all pointers being the same size - check it */
#ifndef lint
  /* this really upsets lint, and it's only for checking anyway */
  POI poi;
  assert(sizeof(poi.longint) == sizeof(poi));
  assert(sizeof(poi.pfpc) == sizeof(poi));
  assert(sizeof(poi.string) == sizeof(poi));
  assert(sizeof(poi.real_ptr) == sizeof(poi));
  assert(sizeof(poi.float_ptr) == sizeof(poi));
  assert(sizeof(poi.double_ptr) == sizeof(poi));
  assert(sizeof(poi.char_ptr) == sizeof(poi));
  assert(sizeof(poi.string_ptr) == sizeof(poi));
  assert(sizeof(poi.long_ptr) == sizeof(poi));
  assert(sizeof(poi.int_ptr) == sizeof(poi));
  assert(sizeof(poi.short_ptr) == sizeof(poi));
  assert(sizeof(poi.boolean_ptr) == sizeof(poi));
  assert(sizeof(poi.ptr) == sizeof(poi));
  assert(sizeof(poi.pptr) == sizeof(poi));
  assert(sizeof(poi.ppptr) == sizeof(poi));
#endif
#ifndef OPERATING_SYSTEM_NAME
#define OPERATING_SYSTEM_NAME "UNIX"
#endif OPERATING_SYSTEM_NAME
  printf("Xerion, V%s, %s %s\n",
	 version, last_change, OPERATING_SYSTEM_NAME);

  itf_func_table =
    INewTable((HASH_TABLE*)0, n_commands, n_commands, ID_INDEX);
  itf_func_context = INewTableContext(itf_func_table);
  itf_com_context = INewTableContext(itf_func_table);
  lex_table = INewTable((HASH_TABLE*)0, 10, 10, 0);
  op_context = INewTableContext(lex_table);
  EnterCommands();	/* call the automatically constructed function */
  space = Malloc(char, lex_space);
  (void) LexInit(lex_space, space, lex_table);
  (void) LexMakeSyntax
    (1,		/* the context value for operators */
     0,			/* don't break words for quotes */
     " 	\n",		/* white space */
     "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz!@$^-+~?/",
     "&|;<>()[].*, =:{}%", /* operator syntax */
     "\"\"''``",	/* quote pairs */
     "0\0b\bn\nt\t",	/* escaped char values */
     "#\n",		/* comment pairs (no two-character comment sequences)*/
     object_syntax);	/* the place where the syntax table is stored */
  (void) LexMakeSyntax
    (1,			/* the context value for operators */
     0,			/* don't break words for quotes */
     " 	\n",		/* white space */
     "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz!@$^-+~?/&|;<>()[].*, =:{}%",
     "", /* operator syntax */
     "\"\"''``",	/* quote pairs */
     "0\0b\bn\nt\t",	/* escaped char values */
     "#\n",		/* comment pairs (no two-character comment sequences)*/
     simple_syntax);	/* the place where the syntax table is stored */
  (void) LexMakeSyntax
    (op_context,	/* the context value for operators */
     0,			/* don't break words for quotes */
     " 	\n",		/* white space */
     "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz!$^*-+~[].?&;()/<>",
     "%=:, {}|<>@",	/* operator syntax */
     "\"\"''``",	/* quote pairs */
     "0\0b\bn\nt\t",	/* escaped char values */
     "#\n",		/* comment pairs (no two-character comment sequences) */
     command_syntax);	/* the place where the syntax table is stored */
  (void) ITableInsert(lex_table, ">>", op_context, NULL);
  (void) ITableInsert(lex_table, "<<", op_context, NULL);
  (void) ITableInsert(lex_table, "%<", op_context, NULL);
  (void) ITableInsert(lex_table, "|<", op_context, NULL);

  /* user generated bindings are entered from the following function */
  IBindInit(max_symbols);
  command_line1 = Malloc(char, max_command_len);
  command_line2 = Malloc(char, max_command_len);
  command_line_len = max_command_len;
  (void) ISetFeatureMask(ITF_SAVE | ITF_VISIBLE | ITF_SYSTEM);
  (void) IBindString("prompt", &prompt, P_OBJECT, 0, "");
  (void) IBindString("readpath", &readpath, P_OBJECT, 0, "");
  (void) IBindString("default-blot", &default_blot, P_OBJECT, 0, "");
  (void) IBindString("noclobber", &noclobber, P_OBJECT, 0, "");
  (void) IBindString("redirection", &redirection, P_OBJECT, 0, "");
  (void) IBindString("automore", &automore, P_OBJECT, 0, "");
  (void) IBindString("plus-minus-options", &plus_minus_options, P_OBJECT, 0, "");
  (void) IBindString("echo", &echo, P_OBJECT, 0, "");
  (void) IBindString("show-function-calls", &show_function_calls, P_OBJECT, 0, "");
  (void) IBindString("read-echo", &read_echo, P_OBJECT, 0, "");
  (void) IBindString("options-echo", &options_echo, P_OBJECT, 0, "");
  (void) IBindString("use-default-options", &use_default_options, P_OBJECT, 0, "");
  (void) IBindString("printstatus", &print_status, P_OBJECT, 0, "");
  (void) IBindLong("status", &status, P_OBJECT, 0, "");
  ISetValue("echo", "false");
  ISetValue("printstatus", "false");
  ISetValue("plus-minus-options", "true");
  /* ISetValue("read-echo", "false"); */
  ISetValue("options-echo", "true");
  ISetValue("use-default-options", "true");
  ISetValue("noclobber", "true");
  ISetValue("redirection", "true");
  auto_more_enabled = (isatty(1) && isatty(0));
  ISetValue("automore", auto_more_enabled ? "false" : "false");
  ISetValue("default-blot", "-2");
  myname = IBaseName(argv[0]);
  sprintf(buf, "%s-> ", myname);
  ISetValue("prompt", buf);
  ISetValue("readpath", ".");
  INewExport("TERM");
  INewExport("PAGER");
  (void) ISetFeatureMask(ITF_SAVE | ITF_VISIBLE /* | ITF_ORPHANABLE */);
  mypid = getpid();
  argv++;
  just_do_defaults = EVERYTHING;
  quit_flag = NORMAL;
  InitReadFile("%s/.xerion", getenv("HOME"));
  InitReadFile(".xerion", getenv("HOME"));
  InitReadFile("%s/.%src", getenv("HOME"), myname);
  InitReadFile(".%src", myname);

  if (true(redirection))
    ISetValue("redirection", "falsie");	/* set to false for reading defaults */
  					/* bad spelling to detect later */
  just_do_defaults = ONLY_DEFAULTS;
  if (*argv)
    printf("Finding defaults in startup files...\n");
  if (!IDoArgvCommands(argv))
    panic("IInit", "Error in command line arguments -- bye bye!", NULL);
  BindObjects();
  DoAutoInit();
  if (strcmp(IGetValue("redirection", NULL), "falsie")==0)
    ISetValue("redirection", "true");
  IInitStream();
  just_do_defaults = EVERYTHING; /* if we use EXCEPT_DEFAULTS here,
				    then the defaults in ul.h override
				    those in a startup file */
  if (*argv)
    printf("Executing commands from startup files...\n");
  if (!IDoArgvCommands(argv))
    panic("IInit", "Error in command line arguments -- bye bye!", NULL);
  just_do_defaults = EVERYTHING;
  (void) signal(SIGINT, ICatchInterrupt);
  (void) signal(SIGPIPE, IPipeErrorRescue);
  (void) signal(SIGSEGV, ISegVioRescue);
  (void) signal(SIGBUS, IBusErrorRescue);
  (void) signal(SIGFPE, IFPErrorRescue);
/*** don't need this - don't need to do anything here for winch to work
#ifdef SIGWINCH
  (void) signal(SIGWINCH, IWindowChange);
#endif
*/
}

int true(s)
char *s;
{
  if (!s || !*s || !strcmp(s, "0")
      || !strcmp(s, "F")
      || !strcmp(s, "f")
      || !strcmp(s, "FALSE")
      || !strcmp(s, "False")
      || !strcmp(s, "-")
      || !strcmp(s, "false")) return 0;
  else return 1;
}
/*
 * functions to assist parsing
 */	

char *IWrongNargs(command, n)
char *command;
char *n;
{
  IError("%s: Wrong number of arguments (expected %s)", command, n);
  return itf_error;
}

char *ITooManyArgs(command, n)
char *command;
char *n;
{
  IError("%s: Too many arguments (expected %s)", command, n);
  return itf_error;
}

char *IArgError(arg, error)
char *arg, *error;
{
  IError("%s argument: %s", error, arg);
  return itf_error;
}

char *IFormatUsage(intro, command, usage)
char *intro, *command, *usage;
{
  char *buf = itf_tmp_buf;
  char *com = command;
  char *p = intro;
  char *start_line = buf;
  while (*p) *(buf++) = *(p++);
  do {
    while (*com) *(buf++) = *(com++);
    *(buf++) = ' ';
    while (*usage && *usage!='|') {
      if (*usage=='%' && (usage[1]=='s' || usage[1]=='p')) {
	if (usage[1]=='p') {
	  if (true(plus_minus_options)) *(buf++) = '+';
	  else *(buf++) = '-';
	} else if (usage[1]=='s') {
	  if (true(plus_minus_options)) buf = strcpyend(buf, "+/-");
	  else buf = strcpyend(buf, "-[no]");
	} else if (usage[1]=='y') {
	  if (true(plus_minus_options)) buf = strcpyend(buf, "+");
	  else buf = strcpyend(buf, "-");
	} else if (usage[1]=='n') {
	  if (true(plus_minus_options)) buf = strcpyend(buf, "-");
	  else buf = strcpyend(buf, "-no");
	} else if (usage[1]=='%') {
	  *(buf++) = '%';
	} else
	  *(buf++) = *(usage--); /* -- to make up for the +2 on next line */
	usage += 2;
      } else
	*(buf++) = *(usage++);
      if (buf[-1]==' ' && (buf - start_line > 60)) {
	strcpy(buf, "\n        ");
	start_line = buf+1;
	buf += strlen(buf);
      }
    }
    if (*usage=='|') {
      com = command;
      p = "\n    ";
      start_line = buf;
      while (*p) *(buf++) = *(p++);
      usage++;
    }
  } while (*usage);
  *buf = 0;
  return itf_tmp_buf;
}

char *IPrintUsage(command, usage)
char *command, *usage;
{
  return IFormatUsage("usage:\n    ", command, usage);
}

char *IInvalidNumber(name)
char *name;
{
  IError("Invalid number: \"%s\"", name);
  return itf_error;
}

void IExpectedToken(what, this, after)
char *what, *this, *after;
{
  char *quote;
  quote = (*what=='<' ? "" : "\"");
  if (this!=NULL && after!=NULL)
    IError("Expected %s%s%s after \"%s\", but found \"%s\"",
	   quote, what, quote, after, this);
  else if (this==NULL && after!=NULL)
    IError("Expected %s%s%s after \"%s\", but found nothing",
	   quote, what, quote, after);
  else if (this!=NULL && after==NULL)
    IError("Expected %s%s%s, but found \"%s\"", quote, what, quote, this);
  else if (this==NULL && after==NULL)
    IError("Expected %s%s%s, but found nothing", quote, what, quote);
}

#ifdef OBSELETE
int IExpectInteger(a)
char *a;
{
  if (a==NULL) a = "(null)";
  else if (IIsInteger(a)) return 1;
  IExpectedToken("<integer>", a, NULL);
  return 0;
}

int IExpectString(a)
char *a;
{
  if (a==NULL) a = "(null)";
  else if (LexQuotedTok(a)) return 1;
  IExpectedToken("<string>", a, NULL);
  return 0;
}

int IExpectVarName(a)
char *a;
{
  if (a==NULL) a = "(null)";
  else if (IIsLegalVarName(a)) return 1;
  IExpectedToken("<varname>", a, NULL);
  return 0;
}

int IExpectNumber(a)
char *a;
{
  if (a==NULL) a = "(null)";
  else if (IIsNumber(a)) return 1;
  IExpectedToken("<real number>", a, NULL);
  return 0;
}
#endif OBSELETE

int IExpectInteger(this, after)
char *this, *after;
{
  if (this!=NULL && IIsInteger(this)) return 1;
  IExpectedToken("<integer>", this, after);
  return 0;
}

int IExpectString(this, after)
char *this, *after;
{
  if (this!=NULL && LexQuotedTok(this)) return 1;
  IExpectedToken("<string>", this, after);
  return 0;
}

int IExpectVarName(this, after)
char *this, *after;
{
  if (this!=NULL && IIsLegalVarName(this)) return 1;
  IExpectedToken("<varname>", this, after);
  return 0;
}

int IExpectNumber(this, after)
char *this, *after;
{
  if (this!=NULL && IIsNumber(this)) return 1;
  IExpectedToken("<real number>", this, after);
  return 0;
}

void IExpected(s)
char *s;
{
  if (!itf_errno) itf_errno = COMMAND_ERROR;
  IError("Expected %s", s);
}

/***
 * should be called like IError("error is %d", x)
 * This checks for overflow in the itf_error buffer - if
 * less than ERROR_BUF_SAFE chars remain, it doesn't put any more in
 */
/*VARARGS0*/
void IError(va_alist)
va_dcl
{
  va_list ap;
  va_start(ap);
  IErrorV(&ap);
  va_end(ap);
}

void Error(va_alist)
va_dcl
{
  va_list ap;
  va_start(ap);
  IErrorV(&ap);
  va_end(ap);
}

ErrorV(ap)
va_list *ap;
{
  IErrorV(ap);
}

void IErrorAbort(va_alist)
va_dcl
{
  va_list ap;
  va_start(ap);
  IErrorV(&ap);
  va_end(ap);
  IAbort();
}

/***
 * ErrorAbort is the same as IErrorAbort
 */
void ErrorAbort(va_alist)
va_dcl
{
  va_list ap;
  va_start(ap);
  IErrorV(&ap);
  va_end(ap);
  IAbort();
}

/***
 * should be called like IErrorV(&ap)
 * This checks for overflow in the itf_error buffer - if
 * less than ERROR_BUF_SAFE chars remain, it doesn't put any more in
 */
void IErrorV(ap)
va_list *ap;
{
  char *fmt, *str;
  TBL_ITEM *item = ITableLookupById(itf_func_table, command_id);
  char *last_command;
  last_command = item==NULL?"unknown":item->name;
  str = itf_error;
  if (itf_errno<=0 || itf_errno>=TOP_ITF_ERRNO)
    itf_errno = COMMAND_ERROR;
  if (*itf_error) {
    str += strlen(itf_error);
    *str++ = '\n';
    /* strcat(str, "\nThis led to error: "); */
    /* str += strlen(str); */
  }
  fmt = va_arg(*ap, char*);
  if (str-itf_error > ERROR_BUF_SIZE-ERROR_BUF_SAFE) {
    fprintf(stderr, "%s\n%s: (in command \"%s\"):",
	    "IError: out of buffer space for saving up error messages!",
	    itf_error_str[itf_errno], last_command);
    vfprintf(stderr, fmt, *ap);
    fputc('\n', stderr);
  } else {
    sprintf(str, "%s", itf_error_str[itf_errno]);
    str += strlen(str);
    if (itf_errno==COMMAND_ERROR) {
      sprintf(str, "%s: ", last_command);
      str += strlen(str);
    }
    if (function_names && function_names->top) {
      sprintf(str, "%s: ", TempStackTop(char *, function_names));
      str += strlen(str);
    }
    vsprintf(str, fmt, *ap);
  }
  if (!itf_errno)
    itf_errno = COMMAND_ERROR;
}

/***
 * should be called as:
 * IOutOfMemory("function-name", NULL)
 * or
 * IOutOfMemory("function-name", "format", arg1, arg2, ...)
 * VARARGS0
 */
char *IOutOfMemory(va_alist)
va_dcl
{
  va_list ap;
  char *fmt, *name, *p;
  va_start(ap);
  name = va_arg(ap, char*);
  fmt = va_arg(ap, char*);
  itf_errno = OUT_OF_MEMORY;
  IError("\"%s\" ", name);
  if (fmt) {
    p = itf_error + strlen(itf_error);
    vsprintf(p, fmt, ap);
  }
  va_end(ap);
  return itf_error;
}

int IHaveCulprit()
{
  return !(itf_error[0]==0);
}

/***
 * Prints out a help line doing the following
 * - indent 4 spaces unless the line is all upper case
 * - change '%p' to '-'
 * - change '%s' to '+/-' or '-[no]' depending on value of 'plus_minus_options'
 */
static void PrintHelpLine(out, line)
FILE *out;
char *line;
{
  int all_caps, i;
  all_caps = 1; /* a "header" line is all capitals, and is not indented */
  if (line[0]!=0) {
    for (i=0;line[i];i++)
      if (!isupper(line[i])&&!isspace(line[i])) {all_caps=0;break;}
    if (!all_caps) fputs("    ", out);
    else if (line[0]) fputc('\n', out); /* extra nl before heading */
    while (*line) {
      if (*line!='%') putc(*(line++), out);
      else {
	if (line[1]=='s') {
	  if (true(plus_minus_options)) fputs("+/-", out);
	  else fputs("-[no]", out);
	} else if (line[1]=='p') {
	  if (true(plus_minus_options)) fputc('+', out);
	  else fputc('-', out);
	} else if (line[1]=='y') {
	  if (true(plus_minus_options)) fputc('+', out);
	  else fputc('-', out);
	} else if (line[1]=='n') {
	  if (true(plus_minus_options)) fputc('-', out);
	  else fputs("-no", out);
	} else if (line[1]=='%') {
	  fputc('%', out);
	} else
	  putc(*(line--), out); /* -- to make up for +2 on next line */
	line+=2;
      }
    }
  }
  fputc('\n', out);
}

void IHelp(va_alist)
va_dcl
{
  va_list ap;
  int help_type;
  char *name, *usage, *synopsis, *long_help;
  va_start(ap);
  help_type	= va_arg(ap, int);
  name		= va_arg(ap, char*);
  usage		= va_arg(ap, char*);
  synopsis	= va_arg(ap, char*);
  long_help	= va_arg(ap, char*);
  if (help_type==SHORT_HELP) {
    help_string = synopsis;
  } else {
    struct TBL_ITEM *item;
    extern HASH_TABLE *arg_table;
    extern int arg_table_context;
    fprintf(dout, "NAME\n    %s - %s\n\n", name, synopsis);
    if (usage!=NULL && *usage!=0) {
      fputs(IFormatUsage("USAGE\n    ", name, usage), dout);
      fputc('\n', dout);
    }
    /* try to find the command in the args stuff, and print out usage */
    /* arg_table may not be set yet, so don't call if not */
    if (arg_table!=NULL)
      item = ITableLookup(arg_table, name, arg_table_context);
    else
      item = NULL;
    if (item!=NULL) {
      fputs("USAGE\n", dout);
      PrintArgUsage((struct ARG_DESC*)item->data.ptr, dout, 0);
      fputc('\n', dout);
    }
    
    if (long_help!=NULL) {
      fprintf(dout, "DESCRIPTION\n");
      while (long_help!=NULL) {
	PrintHelpLine(dout, long_help);
	long_help	= va_arg(ap, char*);
      }
      fputc('\n', dout);
    }
    if (item!=NULL) {
      fputs("OPTIONS AND ARGUMENTS\n", dout);
      PrintArgHelp((struct ARG_DESC*)item->data.ptr, dout, 0);
    }
    fprintf(dout, "\n");
  }
}

IResetError()
{
  itf_errno = 0;
  itf_error[0] = 0;
}

/***
 * Swaps the abort env with env
 *
 * Should be used as:
 * jmp_buf env;
 * int *old_env;
 * ...
 * if (setjmp(env))
 *   handle jump
 * else
 *   old_env = ISetAbortEnv(env);
 * ...
 * ISetAbortEnv(env);
 * return;
 */
int *ISetAbortEnv(env)
int *env;
{
  int *old;
  old = itf_abort_env;
  itf_abort_env = env;
  return old;
}

/***
 * Abort to the last place that set 'itf_abort_env'
 */
void IAbort()
{
  longjmp(itf_abort_env, 1);
}

/***
 * Abort to the last place that set 'itf_abort_env'
 */
void Abort()
{
  longjmp(itf_abort_env, 1);
}

/***
 * Catch an Interrupt
 */
void ICatchInterrupt()
{
#ifdef SYSV_SIGNALS
  (void) signal(SIGINT, ICatchInterrupt);
#endif
  if (!itf_errno) {
    itf_errno = INTERRUPT;
    caught_signal = SIGINT;
    fprintf(stderr, "\n");
  }
  longjmp(itf_abort_env, 1);
}

static int seg_vio_count=0;
static int bus_error_count=0;
static int fp_error_count=0;
/***
 * rescue from a segmentation violation
 */
void ISegVioRescue()
{
#ifdef SYSV_SIGNALS
  (void) signal(SIGSEGV, ISegVioRescue);
#endif
  fprintf(stderr, "Segmentation violation caught - aborted\n");
  seg_vio_count++;
  itf_errno = ERROR_SIGNAL;
  caught_signal = SIGSEGV;
  IErrorAbort("");
}

void IBusErrorRescue()
{
#ifdef SYSV_SIGNALS
  (void) signal(SIGBUS, IBusErrorRescue);
#endif
  fprintf(stderr, "Bus error caught - aborted\n");
  bus_error_count++;
  itf_errno = ERROR_SIGNAL;
  caught_signal = SIGBUS;
  IErrorAbort("");
}

void IFPErrorRescue()
{
#ifdef SYSV_SIGNALS
  (void) signal(SIGFPE, IFPErrorRescue);
#endif
  fprintf(stderr, "Floating point error caught - aborted\n");
  fp_error_count++;
  itf_errno = ERROR_SIGNAL;
  caught_signal = SIGFPE;
  IErrorAbort("");
}

void IPipeErrorRescue()
{
#ifdef SYSV_SIGNALS
  (void) signal(SIGPIPE, IPipeErrorRescue);
#endif
  if (errno!=EPIPE) {
    /* fprintf(stderr, "Pipe error caught but errno not set - continuing\n"); */
    return;
  }
  /* fprintf(stderr, "Pipe error caught - aborted\n"); */
  itf_errno = ERROR_SIGNAL;
  caught_signal = SIGPIPE;
  IErrorAbort("");
}

void IWindowChange()
{
  fprintf(stderr, "Got a window change signal, but don't know what to do with it!\n");
  caught_signal = SIGWINCH;
}

IReportError(s)
FILE *s;
{
  if (itf_errno<0 || itf_errno>=TOP_ITF_ERRNO)
    warn("IReportError", "errno %d out of range", itf_errno);
  if (true(print_status)) {
    fprintf(s, "Status: %d:", itf_errno);
    if (itf_errno==ERROR_REPORTED) printf("\n");
  }
  if (itf_errno!=ERROR_REPORTED && itf_error[0]!='\0')
    fprintf(s, "%s\n", itf_error);
}

static int *line_numbers_ok_p = NULL;
IMuckUpLineNumbers()
{
  if (line_numbers_ok_p!=NULL)
    *line_numbers_ok_p = 0;
}

static char* GetInputLine(line, line_len, in, line_no)
char *line;
int line_len;
FILE *in;
int *line_no;
{
  char *p = line;
  char *end = line+line_len;
  char *ret;
  /* int cont = 1; */
  while (1) {
    line_len = end-p;
    ret = fgets(p, line_len, in);
    if (line_no!=NULL) (*line_no)++;
    if (ret==NULL) return (p==line?NULL:line);
    while (*p) p++;
    if ((p-1)>line && p[-2]=='\\') {
      p[-2] = p[-1];
      p[-1] = '\0';
      p--;
      /* cont = 1; */
    } else
      break;
  }
  return ret;
}

#ifdef OLDWAY
ICommandLoop(this_instream, this_outstream, new_prompt)
FILE *this_instream, *this_outstream;
char *new_prompt;
{
  jmp_buf env;
  int *old_env;
  int tokc;
  char **tokv;
  char *line;
  if (quit_flag==QUIT) return; /* 'quit' means QUIT! */
  if (new_prompt!=NULL) (void) ISetValue("prompt", new_prompt);
  if (setjmp(env))
    IReportError(stderr);
  else
    old_env = ISetAbortEnv(env);
  while (1) {
    IResetError();
    dout = this_outstream;
    din = this_instream;
    if (*prompt)
      fputs(prompt, dout);
    if (GetInputLine(command_line1, command_line_len, din, NULL)==NULL)
      break;
    line = VarSubstitute(command_line1, command_line2, command_line_len);
    if (true(echo)) {
      fputs(command_line1, dout);
      fflush(dout);
    }
    if (line!=NULL)
      tokv = LexAnalyse(command_syntax, line, &tokc);
    if (line==NULL || tokv==NULL) {
      IReportError(dout);
      continue;
    }
    if (pre_command_function) pre_command_function();
    (void) IDoCommand(tokc, tokv);
    if (post_command_function) post_command_function();
    if (itf_errno || true(print_status)) {
      IReportError(dout);
    }
    if (quit_flag==QUIT) break;
    if (quit_flag==EXIT) {quit_flag=NORMAL; break;}
  }
  (void) ISetAbortEnv(old_env);
}

#else	/* NOT OLDWAY */

int	IDoSingleCommand(this_instream, this_outstream)
  FILE	*this_instream, *this_outstream;
{
  int	tokc;
  char	**tokv;
  char	*line;

  IResetError();

  dout = this_outstream;
  din  = this_instream;

  if (GetInputLine(command_line1, command_line_len, din, NULL)==NULL)
    exit(0) ;

  line = VarSubstitute(command_line1, command_line2, command_line_len);
  if (true(echo)) {
    fputs(command_line1, dout);
    fflush(dout);
  }

  if (line != NULL)
    tokv = LexAnalyse(command_syntax, line, &tokc);
  if (line == NULL || tokv == NULL) {
    IReportError(dout);
  }

  if (pre_command_function) 
    pre_command_function();
  IDoCommand(tokc, tokv);
  if (post_command_function)
    post_command_function();

  if (itf_errno || true(print_status)) {
    IReportError(dout);
  }
  if (quit_flag ==  QUIT) 
    return 0 ;
  if (quit_flag == EXIT) {
    quit_flag=NORMAL; 
    return 0 ;
  }
  if (*prompt) {
    fputs(prompt, dout);
    fflush(dout);
  }
  return 1 ;
}

static int	nonGraphicsLoop(this_instream, this_outstream)
  FILE	*this_instream, *this_outstream;
{
  while(IDoSingleCommand(this_instream, this_outstream) != 0)
    ;
  return 0 ;
}

int	ICommandLoop(this_instream, this_outstream, new_prompt)
  FILE	*this_instream, *this_outstream;
  char	*new_prompt;
{
  jmp_buf	env;
  int		*old_env;

  if (quit_flag == QUIT) 	/* 'quit' means QUIT! */
    return;
  if (new_prompt != NULL) 
    (void) ISetValue("prompt", new_prompt);

  if (setjmp(env)) {
    IReportError(stderr);
  } else {
    old_env = ISetAbortEnv(env);
  }  
  if (*prompt) {
    fputs(prompt, dout);
    fflush(dout);
  }

  if (XMode != 0) {
    graphicsLoop(this_instream, this_outstream) ;
  } else {
    nonGraphicsLoop(this_instream, this_outstream) ;
  }

  (void) ISetAbortEnv(old_env);
}
#endif



int IDoArgvCommands(argv)
char *argv[];
{
  int tokc;
  char **tokv;
  char *line;
  char *p=command_line1;
  command_line1[0]=0;
  while (*argv) {
    while (*argv && strcmp(*argv, ";")) {
      *(p++) = ' ';
      strcpy(p, *argv);
      while (*p) p++;
      argv++;
    }
    /* printf("Doing arg command: %s\n", command_line); */
    line = VarSubstitute(command_line1, command_line2, command_line_len);
    if (line!=NULL)
      tokv = LexAnalyse(command_syntax, line, &tokc);
    if (line==NULL || tokv==NULL) {
      IReportError(stderr);
      IResetError();
      return 0;
    }
    if (just_do_defaults==EVERYTHING
	|| !strcmp("default", *tokv) || !strcmp("read", *tokv)
	|| !strcmp("var", *tokv)) {
      (void) IDoCommand(tokc, tokv);
      if (itf_errno) {
	IReportError(stderr);
	IResetError();
	return 0;
      }
    }
    p = command_line1;
    command_line1[0] = 0;
    if (*argv) argv++;
  }
  return 1;
}

int IDoCommandLine(line)
char *line;
{
  int tokc;
  LEX_STACK stack;
  char **tokv;
  int com_status = 1;
  if (line==NULL) return 1;
  line = VarSubstitute(line, command_line2, command_line_len);
  if (line!=NULL)
    tokv = LexAnalysePush(command_syntax, line, &tokc, &stack);
  if (line!=NULL && tokv!=NULL)
    IDoCommand(tokc, tokv);
  else
    com_status = 0;
  if (com_status && itf_errno)
    com_status = 0;
  LexAnalysePop(&stack);
  return com_status;
}

#ifdef OBSELETE
#define MAXARGS 10
int IDoCommandList(va_alist)
va_dcl
{
  char *tokv[MAXARGS];
  int tokc=0;
  va_list pvar;
  va_start(pvar);
  while ((tokv[tokc++]=va_arg(pvar, char*))!=NULL)
    if (tokc>=MAXARGS) panic("IDoCommandList", "too many arguments");
  tokc--;
  va_end(pvar);
  return IDoCommand(tokc, tokv);
}
#endif OBSELETE

char *ISwapWords(p)
char *p;
{
  char buf[100];
  if (!index(p, '-')) return p;
  strcpy(buf, index(p, '-')+1);
  *index(p, '-') = 0;
  strcat(buf, "-");
  strcat(buf, p);
  strcpy(p, buf);
  return buf;
}

/*
 * Finds a command - the command might have "syllables" in
 * a different order to the given name, this will rotate
 * the syllables until it finds the command or will return NULL
 */
TBL_ITEM *IFindCommand(name, id)
char *name;
int *id;
{
  int dash_count = 0;
  char *p;
  TBL_ITEM *function;
  function=ITableLookup(itf_func_table, name, itf_com_context);
  if (function==NULL) {
    for (p=name;*p;p++) if (*p=='-') dash_count++;
    if (dash_count>0)
      for (;dash_count>=0;dash_count--) {
	ISwapWords(name);
	if (function==NULL)
	  function=ITableLookup(itf_func_table, name, itf_com_context);
      }
  }
  if (function!=NULL && id!=NULL) *id = function->id;
  return function;
}

/*
 * Executes a command given a tokv vector
 */
int IDoCommand(tokc, tokv)
int tokc;
char *tokv[];
{
  TBL_ITEM *function;
  char *p, **pp, **optv=NULL, *options=NULL;
  int optc=0;
  jmp_buf env;
  int *old_env;
  SignalHandler old_signal_handler;
  int old_command_id = command_id;
  struct STACK *old_function_names = function_names;
  struct STREAM *old_instream, *old_outstream;
  struct STREAM *old_command_stdin = command_stdin;
  struct STREAM *old_command_stdout = command_stdout;
  struct TEMP_ALLOC temp_alloc_rec;
  /* remember whether to redirect - value might change as result of command */
  int redirect = true(redirection);
  if (tokc<1) return 0;
  errno = 0;
  /* pre-process just the command name converting underscores to dashes */
  for(pp=tokv;pp==tokv;pp++)
    for (p= *pp;*p;p++) if (*p=='_') *p='-';
  function = IFindCommand(tokv[0], &command_id);
  command_stdin = current_instream;	/* don't have to remember these coz */
  command_stdout = current_outstream;	/* they're set before each command */
  function_names = NULL; /* stack is allocated if needed */
  if (function!=NULL) {
    if (true(use_default_options)) {
      strcpy(itf_tmp_buf, function->name);
      strcat(itf_tmp_buf, "-options");
      if (IIsObjectName(itf_tmp_buf)) {
	options = IGetValue(itf_tmp_buf, NULL);
	if (itf_errno || options==NULL) {
	  itf_errno = COMMAND_ERROR;
	  IError("failed to get default options");
	  goto finish;
	}
	/* fprintf(dout, "Options are: \"%s\"\n", options); */
	optv = LexAnalyseAdd(command_syntax, options, &optc);
	/*
	for (i=0;i<tokc;i++)
	  fprintf(dout, "tokv[%d] 0x%X = \"%s\"\n", i, tokv+i, tokv[i]);
	for (i=0;i<optc;i++)
	  fprintf(dout, "optv[%d] 0x%X = \"%s\"\n", i, optv+i, optv[i]);
	  */
      }
    }
    IBeginTempAlloc(&temp_alloc_rec);
    if (optc) { /* insert the extra options after the command */
      int i;
      char **tokv2;
      tokv2 = ITempCalloc(char*, tokc+optc+1);
      tokv2[0] = tokv[0];
      for (i=0;i<optc;i++) tokv2[i+1] = optv[i];
      for (i=1;i<tokc;i++) tokv2[optc+i] = tokv[i];
      tokc += optc;
      tokv = tokv2;
    }
    if (/* true(echo) || */ (optc&&true(options_echo))) {
      char *p = IComposeTokvLine(tokv, 1);
      if (p!=NULL) {
	fputs(p, dout);
	fputc('\n', dout);
	fflush(dout);
      }
    }
    if (!redirect || IDoRedirection(&tokc, tokv, &old_instream, &old_outstream)) {
      if (setjmp(env)==0) {
	old_signal_handler = signal(SIGINT, IAbort);
	old_env = ISetAbortEnv(env);
	(function->data.pfi)(tokc, tokv);	/* call the command */
      }
      (void) ISetAbortEnv(old_env);
      (void) signal(SIGINT, old_signal_handler);
      if (redirect) IEndRedirection(old_instream, old_outstream);
    }
    IFreeTempAllocs(&temp_alloc_rec);
  } else {
    itf_errno = UNKNOWN_COMMAND;
    IError("%s", tokv[0]);
  }
 finish:
  function_names = old_function_names;
  command_stdin = old_command_stdin;
  command_stdout = old_command_stdout;
  command_id = old_command_id;
  return itf_errno;
}

char *IComposeTokvLine(tokv, use_quotes)
char *tokv[];
int use_quotes;
{
  char *p = command_line1;
  char *q;
  int on_first_token=1;
  while (*tokv) {
    if (LexSpBefore(*tokv) && !on_first_token) *p++ = ' ';
    if (use_quotes && LexQuotedTok(*tokv)) *p++ = LexQuotedTok(*tokv);
    for (q = *tokv;*q;q++) *p++ = *q;
    if (use_quotes && LexQuotedTok(*tokv)) *p++ = LexQuotedTok(*tokv);
    if (p-command_line1 > command_line_len) {
      itf_errno = LEX_ERROR;
      IError("Command too long");
      return NULL;
    }
    on_first_token=0;
    tokv++;
  }
  *p = 0;
  return command_line1;
}

int IEnterCommand(name, function)
char *name;
PFI function;
{
  TBL_ITEM *item;
  char *p=name;
  for (;*p;p++) if (*p=='_') *p='-';
  item = ITableInsert(itf_func_table, name, itf_com_context, function);
  if (item == NULL && hash_errno==HT_DUPLICATE) {
    /* fprintf(dout, "IEnterCommand: entering new command \"%s\"", name); */
    item = ITableLookup(itf_func_table, name, itf_com_context);
    item->data.pfi = function;
  }
  if (item == NULL && hash_errno==HT_OVERFLOW)
    panic("IEnterCommand", "table overflow on: %s", name);
  return 0;
}

int IEnterFunction(name, function)
char *name;
PFI function;
{
  TBL_ITEM *item;
  char *p=name;
  for (;*p;p++) if (*p=='_') *p='-';
  item = ITableInsert(itf_func_table, name, itf_func_context, function);
  if (item == NULL && hash_errno==HT_DUPLICATE)
    panic("IEnterFunction", "function is already in table: %s", name);
  if (item == NULL && hash_errno==HT_OVERFLOW)
    panic("IEnterFunction", "table overflow on: %s", name);
  return 0;
}

PFI IGetFunction(name)
char *name;
{
  TBL_ITEM *item;
  char *p=name;
  for (;*p;p++) if (*p=='_') *p='-';
  item = ITableLookup(itf_func_table, name, itf_func_context);
  if (item!=NULL)
    return item->data.pfi;
  else
    return NULL;
}

command_sh(tokc, tokv)
int tokc;
char *tokv[];
{
  int echo_command = 0;
  char *command;
  int r;
  if (GiveHelp(tokc)) {
    IUsage("[-echo] <command...>");
    ISynopsis("run a shell command");
    IHelp
      (IHelpArgs,
       "The shell command is executed as typed, except that any escaped",
       "characters are converted to the character they stand for, e.g. '\\n'",
       "is converted to the newline character.",
       "Quote characters are retained in the line sent to the shell, except",
       "if the interface sees only one token in the shell command, in which",
       "case the quotes are omitted.  This allows the quoting of shell",
       "commands in order to get the shell and not the interface to interpret",
       "redirection and other meta-characters",
       "NOTE ABOUT REDIRECTION",
       "Redirection does not work smoothly with the 'sh' command.",
       "More specifically, any redirection that is interpreted by the",
       "command interface will not have any effect on the output of the",
       "shell command.  Thus a command like \"sh cat foo > tmp\" will",
       "result in the contents of \"foo\" being displayed on the terminal",
       "and in \"tmp\" being an empty file.  To get shell commands redirected, ",
       "enclose the shell command in quotes, e.g., \"sh 'cat foo > tmp'\"",
       "This way the shell will do the redirection.",
       "OPTIONS",
       "-echo : echo the command just before sending it to the shell.  What",
       "  is printed is exactly what is sent to the shell (except for the",
       "  the newline that is added for printing.",
       "SIGNALS",
       "The process is given a different parent process group, if the last",
       "character in the last argument is a '&'.",
       "This will prevent it from receiving keyboard interrupt signals, which",
       "could result in the unintended killing of a backgrounded process.",
       NULL);
    return;
  }
  tokv++;
  tokc--;
  if (*tokv && !strcmp("-echo", *tokv)) {
    echo_command=1;
    tokv++;
    tokc--;
  }
  if (tokc<1) {
    IError("no shell command specified");
    return;
  }
  command = IComposeTokvLine(tokv, tokc>1);
  if (command==NULL) return;
  if (echo_command) fprintf(dout, "%s\n", command);
  if (strlen(tokv[tokc-1])>0 && tokv[tokc-1][strlen(tokv[tokc-1])-1]=='&') {
    /***
     * this fork is here because we want to detach the child from our
     * pgroup so that backgrounded processes do not inherit signals
     * REALLY UGLY!
     */
    r = fork();
    if (r<0) IErrorAbort("fork failed");
    if (r==0) {
      setpgrp(getpid(), getpid());
      system(command);
      exit(1);
    }
  } else
  system(command);
}

command_echo(tokc, tokv)
int tokc;
char *tokv[];
{
  if (GiveHelp(tokc)) {
    IUsage("<arg1> <arg2> ...");
    ISynopsis("print its arguments, all on a line");
    IHelp(IHelpArgs,
	  NULL);
    return;
  }

  if (*++tokv) {
    while (*tokv) {
      fprintf(dout, "%s", tokv[0]);
      if (*++tokv) fputc(' ', dout);
    }
    fputc('\n', dout);
  }
}

command_quit(tokc, tokv)
int tokc;
char *tokv[];
{
  if (GiveHelp(tokc)) {
    IUsage("");
    ISynopsis("quit the simulator");
    IHelp(IHelpArgs,
	  NULL);
    return;
  }
  quit_flag = QUIT;
}

command_exit(tokc, tokv)
int tokc;
char *tokv[];
{
  if (GiveHelp(tokc)) {
    IUsage("");
    ISynopsis("stop reading the current input file");
    IHelp(IHelpArgs,
	  "If the simulator is reading a file, it will stop reading it",
	  "If the simulator is at the top level (i.e. prompting), it will quit.",
	  NULL);
    return;
  }
  quit_flag = EXIT;
}

command_args(tokc, tokv)
int tokc;
char *tokv[];
{
  int i=0;
  if (GiveHelp(tokc)) {
    IUsage("<arg1> <arg2> ...");
    ISynopsis("print its arguments, on separate lines");
    IHelp(IHelpArgs,
	  NULL);
    return;
  }
  
  fprintf(dout, "Command had %d arguments\n", tokc);
  while (*tokv) {
    if (LexQuotedTok(*tokv))
      fprintf(dout, "tokv[%d] = %c%s%c%s\n", i,
	      LexQuotedTok(*tokv), tokv[0], LexQuotedTok(*tokv),
	      (LexIsOperator(*tokv) ? " (operator)" : ""));
    else
      fprintf(dout, "tokv[%d] = %s%s\n", i,
	      tokv[0], (LexIsOperator(*tokv) ? " (operator)" : ""));
    tokv++;
    i++;
  }
}

int strcmp_p(a, b)
char **a, **b;
{
  return strcmp(*a, *b);
}

int max(a, b)
int a, b;
{
  return a>b ? a : b;
}

/*
 * The "synopsis" strings from all commands are kept in this array
 */
static char **help_array;
static int help_array_size=0;
BuildSynopsisArray()
{
  TBL_ITEM *function;
  int id, i;
  char *help_tokv[3];

  help_tokv[0] = NULL;
  help_tokv[1] = "help";
  help_tokv[2] = NULL;
  /***
   * Build and sort the array of help strings, first loop to count them,
   * second to sort them
   */
  id = 0;
  while ((function=ITableNextByContext(itf_func_table, &id, itf_com_context))!=NULL) {
    help_tokv[0] = function->name;
    help_string = NULL;
    (void) (function->data.pfi)(SHORT_HELP, help_tokv);
    if (itf_errno) {
      warn("help", "command \"%s\" gave error code %d!!", function->name, itf_errno);
      warn("help", "All commands MUST have a help section");
      IReportError(stderr);
      IResetError();
    }
    if (help_string!=NULL)
      help_array_size++;
    else {
      warn("help", "No help provided for the command \"%s\"", function->name);
      warn("help", "All commands MUST have a help section");
    }
  }
  help_array = (char**)calloc((unsigned)help_array_size, sizeof(char*));
  id = 0;
  i = 0;
  while ((function=ITableNextByContext(itf_func_table, &id, itf_com_context))!=NULL) {
    help_tokv[0] = function->name;
    help_string = NULL;
    (void) (function->data.pfi)(SHORT_HELP, help_tokv);
    if (help_string!=NULL) {
      help_array[i] = (char*)
	malloc((unsigned)strlen(help_string)+6+max(16, strlen(function->name)));
      sprintf(help_array[i], "%-16s - %s\n", function->name, help_string);
      i++;
    }
  }
  qsort((char*)help_array, help_array_size, sizeof(char*), strcmp_p);
}


command_help(tokc, tokv)
int tokc;
char *tokv[];
{
  TBL_ITEM *function;
  char *keyword=0;
  char *help_tokv[3];
  int i;
  if (GiveHelp(tokc)) {
    IUsage("[[-k] command-name]");
    ISynopsis("provide help for commands or keywords");
    IHelp
      (IHelpArgs,
       "\"Help\" can be used in three ways:",
       "help",
       "  Without any arguments, a listing of all commands, with their synopsis",
       "  lines, is provided.\n",
       "help <command-name>",
       "  With a command name as an argument, detailed help for that command",
       "  is printed.\n",
       "help -k <keyword>",
       "  With the -k option, a list of all commands with <keyword> (can be",
       "  an arbitrary string) in their synopsis is printed.",
       NULL);
    return;
  }

  help_tokv[0] = NULL;
  help_tokv[1] = "help";
  help_tokv[2] = NULL;
  if (tokc>1 && !strcmp("-k", tokv[1])) {
    keyword = tokv[2];
    if (keyword==NULL) {
      IExpectedToken("<keyword>", tokv[2], tokv[1]);
      return;
    }
  }
  if (tokc<2) {
    /*
     * print short help on each command
     */
    if (help_array_size==0) BuildSynopsisArray();
    for (i=0;i<help_array_size;i++)
      fputs(help_array[i], dout);
  } else if (keyword) {
    if (help_array_size==0) BuildSynopsisArray();
    for (i=0;i<help_array_size;i++)
      if (substr(keyword, help_array[i]))
	fputs(help_array[i], dout);
  } else {
    /*
     * print long help on a particular command
     */
    function = IFindCommand(tokv[1], NULL);
    if ((function!=NULL)) {
      help_string = NULL;
      help_tokv[0] = tokv[1];
      (function->data.pfi)(LONG_HELP, help_tokv);
      if (itf_errno) {
	warn("help", "command \"%s\" gave error code %d!!", function->name, itf_errno);
	IReportError(stderr);
	IResetError();
      }
      if (help_string!=NULL)
	fprintf(dout, "%s - %s\n", tokv[1], help_string);
    } else {
      itf_errno = UNKNOWN_COMMAND;
      IError("%s", tokv[1]);
    }
  }
}

/* ARGSUSED */
command_hello(tokc, tokv)
int tokc;
char *tokv[];
{
  if (GiveHelp(tokc)) {
    IUsage("");
    ISynopsis("print out the string \"Hello!\"");
    IHelp(IHelpArgs,
	  NULL);
    return;
  }

  fprintf(dout, "Hello!\n");
}

/***
 * find the first file in the path, path components separated by comma,
 * colon, or space, file could end in `.Z'
 */
char *IFindFileInPath(name, path, Z_suffix)
char *name;
char *path;
int Z_suffix;
{
  static char pathbuf[MAXPATHLEN];
  struct stat statbuf;
  char *path_ptr = path;
  char *cur_comp;
  char *r;
  extern int errno;
  if (name==NULL || *name==0)
    return NULL;
  if (*name=='/')
    return name;
  while (*path_ptr) {
    while (*path_ptr==' ')
      path_ptr++;
    r = pathbuf;
    cur_comp = path_ptr;
    while (*cur_comp && *cur_comp!=' ' && *cur_comp!=',' && *cur_comp!=':')
      cur_comp++;				/* put the initial path */
    while (path_ptr<cur_comp && *path_ptr)
      *r++ = *path_ptr++;
    if (r>pathbuf && *(r-1)!='/')
      *r++ = '/';	/* add / at end if needed */
    *r = 0;
    if (!strcmp(pathbuf, "./"))
      r = pathbuf;
    path_ptr = name;
    while (*path_ptr)
      *r++ = *path_ptr++;	/* and the name */
    *r = 0;
    path_ptr = cur_comp;
    if (*path_ptr)
      path_ptr++;
    /* printf("Checking \"%s\"\n", pathbuf); */
    if (stat(pathbuf, &statbuf)==0)
      return pathbuf;
    if (   Z_suffix && errno==ENOENT
	&& !(r[-1]=='Z' && r[-2]=='.')) {/* try adding .Z */
      r[0] = '.';
      r[1] = 'Z';
      r[2] = '\0';
      if (stat(pathbuf, &statbuf)==0)
	return pathbuf;
    }
    if (errno!=ENOENT) {
      IError("System error: %s \"%s\"\n", SysError(errno), pathbuf);
      break;
    }
    errno = 0;
  }
  return NULL;
}

/***
 * Finds a file in readpath, and opens it.
 * The file can end in `.Z'.
 */
FILE *IOpenFileOrAbort(name, mode, full_namep)
char *name;
char *mode;
char **full_namep;
{
  char *full_name;
  int len;
  FILE *file;
  if (!strcmp(mode, "r"))
    full_name = IFindFileInPath(name, readpath, 1);
  else
    full_name = name;
  if (full_name==NULL || itf_errno)
    ErrorAbort("cannot find file \"%s\"", name);
  if (full_namep!=NULL)
    *full_namep = full_name;

  len = strlen(full_name);
  if (!strcmp(mode, "r") && !strcmp(full_name+len-2, ".Z")) {
    char buf[MAXPATHLEN];
    sprintf(buf, "zcat %s", full_name);
    file = popen(buf, "r");
  } else
    file = fopen(full_name, mode);
  if (file==NULL)
    ErrorAbort("cannot open file \"%s\": %s", full_name, SysError(errno));
  return file;
}

int command_read(tokc, tokv)
int tokc;
char *tokv[];
{
  int pipe=0;
  char linebuf[MAX_COMMAND_LEN];
  char *line;
  int line_no=1;
  int *old_line_numbers_ok_p = line_numbers_ok_p;
  int line_numbers_ok = 1;
  int read_status;
  FILE *old_din = din;
  int tokc2;
  char **tokv2;
  char *file_name, *tmp;
  LEX_STACK stack;
  struct STREAM read_stream;
  struct STREAM *old_instream = current_instream;
  int len;
  if (GiveHelp(tokc)) {
    IUsage("<filename> <filename> ...");
    ISynopsis("read a file of commands");
    IHelp
      (IHelpArgs,
       "The file is searched for in the directories listed in the variable",
       "\"readpath\" (unless the filename begins with a '/').",
       "The current directory will not be searched unless \".\" is in",
       "\"readpath\".  The directories in \"readpath\" should be separated",
       "by a comma, colon, or space.",
       NULL);
    return;
  }

  if (tokc<2) {
    IExpected("<file-name>", tokv[0]);
    return;
  }
  while (*++tokv!=NULL) {
    file_name = NULL;
    tmp = IFindFileInPath(*tokv, readpath, 0);
    if (tmp!=NULL) {
      file_name = ITempCalloc(char, strlen(tmp)+1);
      strcpy(file_name, tmp);
    } else {
      /* look for the equivalent .Z file */
      tmp = ITempCalloc(char, strlen(*tokv)+3);
      strcpy(tmp, *tokv);
      strcat(tmp, ".Z");
      tmp = IFindFileInPath(tmp, readpath, 0);
      if (tmp!=NULL) {
	file_name = ITempCalloc(char, strlen(tmp)+1);
	strcpy(file_name, tmp);
      }
    }
    if (file_name==NULL || (din=fopen(file_name, "r")) == 0) {
      IError("could not open file: \"%s\"", (file_name==NULL?*tokv:file_name));
      din = old_din;
      return;
    }
    read_stream.file = din;
    read_stream.desc = file_name;
    read_stream.name = file_name;
    read_stream.machine = NULL;
    read_stream.type = STREAM_FILE;
    read_stream.status = STREAM_OPEN;
    read_stream.port = read_stream.real_port = read_stream.search = 0;
    read_stream.permanent = 0;
    read_stream.next = NULL;
    read_stream.mode = "r";
    ISetDefaultStreamBlot(&read_stream);
    /***
     * check for a .Z suffix - close the file & pipe through zcat
     * the previous open was useful to check that the file was there
     */
    len = strlen(file_name);
    if (len>1 && file_name[len-1]=='Z' && file_name[len-2]=='.') {
      pipe=1;
      fclose(din);
      strcpy(linebuf, "zcat ");
      strcat(linebuf, file_name);
      din = popen(linebuf, "r");
      read_stream.file = din;
      read_stream.desc = linebuf;
      read_stream.type = STREAM_PORT;
    }
    current_instream = &read_stream;
    fprintf(dout, "Reading file \"%s\".\n", file_name);
    while (!feof(din)) {
      if (GetInputLine(linebuf, MAX_COMMAND_LEN, din, &line_no)==0) break;
      if (just_do_defaults==ONLY_DEFAULTS) {
	char *end;	/* if the line starts with a number, don't do it */
	strtod(linebuf, &end);
	if (end > linebuf)
	  continue;
      }
      read_status = 0;
      line = VarSubstitute(linebuf, command_line2, command_line_len);
      if (line!=NULL)
	tokv2 = LexAnalysePush(command_syntax, line, &tokc2, &stack);
      if (line==NULL || tokv2==NULL) {
	IReportError(dout);
	IResetError();
	read_status = 1;
      }
      if (read_status==0 && tokc2>0) {
	if (just_do_defaults==EVERYTHING
	    || !strcmp("default", *tokv2) || !strcmp("read", *tokv2)
	    || !strcmp("var", *tokv2)) {
	  if (true(echo) && true(read_echo))
	    fputs(line, dout);
	  if (pre_command_function) pre_command_function();
	  read_status = IDoCommand(tokc2, tokv2);
	  if (post_command_function) post_command_function();
	}
      }
      if (read_status) {
	itf_errno = COMMAND_ERROR;
	if (line_numbers_ok)
	  IError("Error reading line %d of file: \"%s\"", line_no, file_name);
	else
	  IError("Error reading file \"%s\" somewhere after line %d", file_name, line_no);
	break;
      }
      LexAnalysePop(&stack);
      if (quit_flag==QUIT) break;
      if (quit_flag==EXIT) {quit_flag = NORMAL; break; };
    }
    if (pipe==0) fclose(din);
    else pclose(din);
    din = old_din;
    line_numbers_ok_p = old_line_numbers_ok_p;
    current_instream = old_instream;
    if (itf_errno) break;
  }
}

/* ARGSUSED */
int command_segv(tokc, tokv)
int tokc;
char *tokv[];
{
  int *p;
  if (GiveHelp(tokc)) {
    IUsage("");
    ISynopsis("cause a segmentation violation");
    IHelp(IHelpArgs,
	  NULL);
    return;
  }
  
  p = NULL;
  *p = 0;
}

/***
 * returns 1 if a is a non-null prefix of b
 */
int strprefix(a, b)
char *a;
char *b;
{
  int i=0;
  if (a==NULL || b==NULL) return 0;
  if (*a=='-' && *b=='-') {a++;b++;}
  if (*a=='+' && *b=='+') {a++;b++;}
  if (*a != *b) return 0;
  while (*a!=NULL && *b!=NULL && *a == *b) {i++;a++;b++;}
  if (*a==NULL) return i;
  else return 0;
}

/***
 * stralloc - makes a copy of the string and returns a pointer to it
 */
char *stralloc(str)
char *str;
{
  char *p = NULL;
  if (str!=NULL) p = malloc((unsigned) strlen(str)+1);
  if (p!=NULL) strcpy(p, str);
  return p;
}

/***
 * return 1 is sub is a substring of str
 */
int substr(sub, str)
char *sub, *str;
{
  char *p; /* current position in sub */
  char *t;
  if (sub==NULL || str==NULL) return 0;
  while (*str) {
    p = sub;
    t = str;
    while (*p && *t && *p==*t) {
      p++;
      t++;
    }
    if (!*p) return 1;
    str++;
  }
  return 0;
}

/***
 * Find the n'th argument after the one that matches arg
 */
char *IFindArg(tokc, tokv, arg, n)
char *tokv[];
char *arg;
int n;
{
  int i;
  if (tokv==NULL || arg==NULL) return NULL;
  assert(n>=0);
  for (i=0;i<tokc;i++)
    if (!strcmp(arg, tokv[i]))
      if (i+n<tokc) return tokv[i+n];
      else return NULL;
  return NULL;
}

/***
 * Find the pos of the n'th argument after the one that matches arg
 * Return tokc if can't find it or if n'th arg is beyond tokc
 */
int IFindArgPos(tokc, tokv, arg, n)
int tokc;
char *tokv[];
char *arg;
int n;
{
  int i;
  if (tokv==NULL || arg==NULL) return tokc;
  assert(n>=0);
  for (i=0;i<tokc;i++)
    if (!strcmp(arg, tokv[i]))
      if (i+n<tokc) return i+n;
      else return tokc;
  return tokc;
}

char *INumberth(n)
int n;
{
  static char buf[20];
  switch (n) {
  case 1: return "1'st";
  case 2: return "2'nd";
  case 3: return "3'rd";
  default:
    sprintf(buf, "%d'th", n);
    return buf;
  }
}

InputMatches(file, str)
FILE *file;
char *str;
{
  int c;
  if (str==NULL) return 1;
  if (file==NULL) return 0;
  while (*str!=NULL) {
    c=fgetc(file);
    if (c!=*str) return 0;
    str++;
  }
  return 1;
}

char *SysError(n)
int n;
{
  static char str[30];
  extern char *sys_errlist[];
  extern int sys_nerr;
  if (n>0 || n<sys_nerr)
    return sys_errlist[n];
  else {
    sprintf(str, "No error message for %d", n);
    return str;
  }
}

