/*
 * $Id: itf.h,v 1.1 90/12/07 11:50:40 drew Exp $
 * 
 * $Log:	itf.h,v $
# Revision 1.1  90/12/07  11:50:40  drew
# Initial revision
# 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#ifndef _ITF_
#define _ITF_

typedef int boolean;

#include <setjmp.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <varargs.h>
#include "config.h"
#include "version.h"
#include "assert.h"
#include "hash-table.h"
#include "lex.h"
#include "bind.h"
#include "warn.h"
#include "blot.h"
#include "enum.h"
#include "trace.h"
#include "stream.h"
#include "mem.h"
#include "range.h"
#include "argf.h"
#include "stack.h"

int caught_signal;	/* the most recently caught signal */

#define COMMAND_ERROR	1
#define SYNTAX_ERROR	3
#define LEX_ERROR	2
#define UNKNOWN_COMMAND	4
#define UNKNOWN_ERROR	5
#define UNKNOWN_VAR	6
#define UNKNOWN_STRUCT	7
#define UNKNOWN_FIELD	8
#define NULL_VARNAME	9
#define ILLEGAL_VARNAME	10
#define ILLEGAL_OBJECT	11
#define OBJECT_SYNTAX	12
#define ARRAY_BOUNDS	13
#define NEED_ARRAY_INDEX	14
#define NEED_STRUCT_FIELD	15
#define NEED_FIELD_OR_INDEX	16
#define READONLY_VAR	17
#define BAD_VARTYPE	18
#define INTERNAL_NULL	19
#define OBJECT_EXCESS	20
#define NULL_POINTER	21
#define EMPTY_OBJECT	22
#define OUT_OF_MEMORY	23
#define INTERRUPT	24
#define ERROR_SIGNAL	25
#define VARNAME_IS_USED	26
#define REDIRECTION_ERROR 27
#define STREAM_NOT_OPEN 28
#define ERROR_REPORTED	29
#define TOP_ITF_ERRNO	30


#ifndef TRUE
#define TRUE		1
#define FALSE		0
#define UNDEFINED	-1
extern char *true_string;
extern char *false_string;
extern char *undefined_string;
extern char *bad_boolean_string;
extern char *BooleanString ARGS((boolean));
#endif

#define truth(x) ((x)==TRUE ? true_string : (x)==FALSE ? false_string : (x)==UNDEF ? undefined_string : bad_boolean_string)

#ifndef _PFI_
#define _PFI_
typedef int (*PFI)(); /* pointer to function returning int */
#endif

extern PFI pre_command_function;
extern PFI post_command_function;
/* argc is set to one of these values when the command should provide info */
#define SHORT_HELP -1
#define LONG_HELP -2
#define IDoHelp(C) if (tokc==SHORT_HELP || tokc==LONG_HELP) { C return; }
#define GiveHelp(tokc) (tokc==SHORT_HELP || tokc==LONG_HELP)
#define IUsage(S) static char *usage = S
#define ISynopsis(S) static char *synopsis = S
#define IHelpArgs tokc, tokv[0], usage, synopsis
/* should be set by a command when it is called with argc == SHORT_HELP */
extern char *help_string;
extern int itf_errno;
extern char itf_error[];
extern char itf_tmp_buf[];
extern char *itf_error_str[];
extern void IExpectedToken();	/* (char*,char*) */
extern void IExpected();	/* (char*) */
extern void IError();		/* (format,arg1,arg2,...) */
extern void IErrorV ARGS((va_list *));
extern void IErrorAbort();	/* (format,arg1,arg2,...) */
extern char *ICommandError();	/* (char*) */
extern char *IArgError();	/* (char* arg,char* error) */
extern char *IPrintUsage();	/* (char* command,char* usage) */
extern char *IComposeTokvLine();/* (tokv) */
extern char *IFindArg();	/* (tokv,tokc,arg,n) */
extern int IDoCommand();		/* (tokc,tokv) */
extern int IDoCommandList();	/* ("command","arg1","arg2",...) */
extern char *VarSubstitute();
extern char command_syntax[];
extern char simple_syntax[];
extern char object_syntax[];
extern int IHaveCulprit();
extern int ICommandLoop();
extern int IStandardInit();
extern void IHelp();
extern char *IOutOfMemory();
extern PFI IGetFunction();
extern HASH_TABLE *itf_func_table;
extern int itf_func_context;
extern int itf_com_context;
extern FILE *dout;
extern FILE *din;
extern char *noclobber;
extern char *default_blot;
extern char *stralloc();
extern int strprefix();
extern char *strcpyend ARGS((char*, char*));
extern char *strcatend ARGS((char*, char*));
extern void strncpy_null ARGS((char*, char*, int));
extern char *SysError();
extern FILE *IOpenFileOrAbort ARGS((char *, char *, char **));
extern char *IFindFileInPath ARGS((char *, char *, int));
#define plural(w) (w==1 ? "" : "s")
#define pluralchoice(w,sing,plur) (w==1 ? sing : plur)
char *INumberth();	/* (n) */

#endif
