/*
 * $Id: load-units.c,v 1.1 90/12/07 11:56:32 drew Exp $
 * 
 * $Log:	load-units.c,v $
# Revision 1.1  90/12/07  11:56:32  drew
# Initial revision
# 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <itf/itf.h>
#include <sim/sim.h>

command_load_input(tokc,tokv)
int tokc;
char *tokv[];
{
  int i;
  if (GiveHelp(tokc)) {
    IUsage("<unit 0 value> <unit 1 value> ...>");
    ISynopsis("load values into the input layer");
    IHelp
      (IHelpArgs,
       "The given values are loaded into the input layer using the function:",
       "  LoadInput(net,unit_n,value)",
       "This function could perform a transformation on the value actually",
       "put into the unit, beware of this.  E.g. it could tranform a value",
       "in the range [0,1] to be in the range [-1,1].",
       "If a '-' is given as the unit value, that unit is not changed",
       NULL);
    return;
  }
  
  if (!NetExists(net)) {
    IError("Current net is NULL");
    return;
  }
  tokv++; tokc--;
  if (tokc != net->p->input_width) {
    IError("the number of values is not equal to the input width");
    return;
  }
  for (i=0;i<tokc;i++) {
    if (!strcmp("-",tokv[i])) continue;
    if (IIsNumber(tokv[i])) LoadInput(net,i,atof(tokv[i]));
    else {
      IError("\"%s\" is not a number!",tokv[i]);
      return;
    }
  }
}

command_load_target(tokc,tokv)
int tokc;
char *tokv[];
{
  int i;
  if (GiveHelp(tokc)) {
    IUsage("<unit 0 value> <unit 1 value> ...>");
    ISynopsis("load values into the target fields of the output layer");
    IHelp
      (IHelpArgs,
       "The given values are loaded into the target fields of the output layer",
       "using the function:",
       "  LoadTarget(net,unit_n,value)",
       "This function could perform a transformation on the value actually",
       "put into the unit, beware of this.  E.g. it could tranform a value",
       "in the range [0,1] to be in the range [-1,1].",
       "If a '-' is given as the unit value, that unit is not changed",
       NULL);
    return;
  }

  if (!NetExists(net)) {
    IError("Current net is NULL");
    return;
  }
  tokv++; tokc--;
  if (tokc != net->p->output_width) {
    IError("the number of values is not equal to the output width");
    return;
  }
  for (i=0;i<tokc;i++) {
    if (!strcmp("-",tokv[i])) continue;
    if (IIsNumber(tokv[i])) LoadTarget(net,i,atof(tokv[i]));
    else {
      IError("\"%s\" is not a number!",tokv[i]);
      return;
    }
  }
}

command_load_layer(tokc,tokv)
int tokc;
char *tokv[];
{
  int i,j;
  IUsage("<layer#> <unit 0 value> <unit 1 value> ...>");
  if (GiveHelp(tokc)) {
    ISynopsis("load values into the output fields of the specified layer");
    IHelp
      (IHelpArgs,
       "The given values are loaded into the output fields of the specified",
       "layer using the function",
       "LoadUnit(net,layer_n,unit_n,value).",
       "If a '-' is given as the unit value, that unit is not changed",
       NULL);
    return;
  }
  
  if (!NetExists(net)) {
    IError("Current net is NULL");
    return;
  }
  if (tokc<3) {
    IError(IPrintUsage(tokv[0],usage));
    return;
  }
  j = atoi(tokv[1]);
  tokv++; tokc--;
  tokv++; tokc--;
  if (j<0 || j>=net->n_layers) {
    IError("layer # out of range");
    return;
  }
  if (tokc != net->layer[j].n_units) {
    IError("the number of values is not equal to the number of units in the layer (%d)",net->layer[j].n_units);
    return;
  }
  for (i=0;i<tokc;i++) {
    if (!strcmp("-",tokv[i])) continue;
    if (IIsNumber(tokv[i])) LoadUnit(net,j,i,atof(tokv[i]));
    else {
      IError("\"%s\" is not a number!",tokv[i]);
      return;
    }
  }
}

static int rand_all_outputs_seed=0;
command_rand_all_outputs(tokc,tokv)
int tokc;
char *tokv[];
{
  int i,j;
  int use_seed;
  if (GiveHelp(tokc)) {
    IUsage("[seed]");
    ISynopsis("randomize the outputs of the network");
    IHelp
      (IHelpArgs,
       "Set all the outputs in the network to random values between 0 and 1.",
       NULL);
    return;
  }
  
  if (!NetExists(net)) {
    IError("Current net is NULL");
    return;
  }
  if (tokv[1]) {
    use_seed = atoi(tokv[1]);
    fprintf(dout,"Randomizing outputs with seed %d\n",use_seed);
    srandom(use_seed);
  } /* else
    use_seed = rand_all_outputs_seed++;
    */
  for (i=0;i<=net->p->output_layer_n;i++)
    for (j=0;j<=net->layer[i].n_units;j++)
      LoadUnit(net,i,j,RandReal(0.5,0.5));
  return;
}
