/*
 * $Id: randreal.c,v 1.1 90/12/07 11:56:55 drew Exp $
 * 
 * $Log:	randreal.c,v $
# Revision 1.1  90/12/07  11:56:55  drew
# Initial revision
# 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <itf/itf.h>
#include <lib/randreal.h>

/***
 * Safe (but mathematically incorrect) division - return zero for divide
 * by zero.  Used for statistics calculation - if no statistics are
 * collected divbyzero can easily result.
 */
real safediv(numerator, denominator)
real numerator, denominator;
{
  if (denominator!=0.0) return numerator/denominator;
  else return 0.0;
}

real Uniform(lower, upper)
real lower, upper;
{
  return (random() / (real) MAX_RAND_INT)*(upper-lower)+lower;
}

/***
 * From "Numerical recipes in C", gasdev.c
 */
real Gaussian(mean, variance)
real mean, variance;
{
  static int iset=0;
  static real gset;
  real fac, r, v1, v2;
  
  if  (iset == 0) {
    do {
      v1 = (random() / (MAX_RAND_INT/2.0))-1.0;
      v2 = (random() / (MAX_RAND_INT/2.0))-1.0;
      r=v1*v1+v2*v2;
    } while (r >= 1.0);
    fac=sqrt(-2.0*log(r)/r);
    gset=v1*fac;
    iset=1;
    return (v2*fac*sqrt(variance))+mean;
  } else {
    iset=0;
    return gset*sqrt(variance)+mean;
  }
}

/***
 * From "Numerical recipes in C", gasdev.c
 */
real GaussianNoise(std_dev)
real std_dev;
{
  static int iset=0;
  static real gset;
  real fac, r, v1, v2;
  
  if  (iset == 0) {
    do {
      v1 = (random() / (MAX_RAND_INT/2.0))-1.0;
      v2 = (random() / (MAX_RAND_INT/2.0))-1.0;
      r=v1*v1+v2*v2;
    } while (r >= 1.0);
    fac=sqrt(-2.0*log(r)/r);
    gset=v1*fac;
    iset=1;
    return (v2*fac*std_dev);
  } else {
    iset=0;
    return gset*std_dev;
  }
}

/***
 * From "Numerical recipes in C", gasdev.c
 */
real GaussianStd()
{
  static int iset=0;
  static real gset;
  real fac, r, v1, v2;
  
  if  (iset == 0) {
    do {
      v1 = (random() / (MAX_RAND_INT/2.0))-1.0;
      v2 = (random() / (MAX_RAND_INT/2.0))-1.0;
      r=v1*v1+v2*v2;
    } while (r >= 1.0);
    fac=sqrt(-2.0*log(r)/r);
    gset=v1*fac;
    iset=1;
    return (v2*fac);
  } else {
    iset=0;
    return gset;
  }
}

real Exponential(mean)
real mean;
{
  return -mean*log(((1+random()) / (real) MAX_RAND_INT));
}
