/*
 * $Id: time.c,v 1.1 90/12/07 11:57:13 drew Exp $
 * 
 * $Log:	time.c,v $
# Revision 1.1  90/12/07  11:57:13  drew
# Initial revision
# 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <itf/itf.h>
#include <sim/sim.h>

/***
 * Include files for time-tracking are different on mips and suns
 */
#if defined(mips) || defined(stellar)
#include <sys/types.h>
#include <sys/times.h>
#endif
#ifdef sun
#include <sys/time.h>
#include <sys/resource.h>
#endif

/***
 * Keeping track of the time - different on mips and suns
 */
real last_time,this_time; /* in seconds */
#ifdef sun
struct rusage rusage;
extern getrusage();
#endif
#if defined(mips) || defined(stellar)
struct tms tms;
#endif


int UpdateTime()
{
  int did_it=0;
#ifdef sun
  (void) getrusage(RUSAGE_SELF,&rusage);
  last_time = this_time;
  this_time = rusage.ru_utime.tv_usec*(1e-6) + rusage.ru_utime.tv_sec;
  did_it = 1;
#endif
#if defined(mips) || defined(stellar)
  times(&tms);
  last_time = this_time;
  this_time = (tms.tms_utime + tms.tms_stime)/100.0;
  did_it = 1;
#endif
  /* fprintf(dout,"Updated time: this=%g  last=%g\n",this_time,last_time); */
  if (!did_it) warn("UpdateTime","don't know how to get time for this system");
}

/* ARGSUSED */
int command_time(tokc,tokv)
int tokc;
char *tokv[];
{
  if (tokc==SHORT_HELP || tokc==LONG_HELP) {
    help_string = "print cpu time usage for last command and total";
    return 0;
  }
  fprintf(dout,"Interval: %g Total: %g\n",this_time-last_time,this_time);
  return 0;
}

PreCommandFunction()
{
}

PostCommandFunction()
{
  UpdateTime();
}

autoinit_time()
{
  last_time = 0.0;
  this_time = 0.0;
  pre_command_function = PreCommandFunction;
  post_command_function = PostCommandFunction;
}

