/*
 * $Id: build-unit.c,v 1.1 90/12/07 12:00:32 drew Exp $
 * 
 * $Log:	build-unit.c,v $
 * Revision 1.1  90/12/07  12:00:32  drew
 * Initial revision
 * 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <itf/itf.h>
#include "sim-dummy.h"
#include <sim.h>

/***
 * Connect a unit to a layer - s(ource)_layer and d(estination)_layer
 * Return the number of links built, or -1 in case of error
 */

int ConnectUnit_(s_net,s_layer,d_net,d_layer,d_unit,
		 link_tag,mode,block_start,n_links,link_set_n)
int s_net;
int s_layer;
int d_net;
int d_layer;
int d_unit;
int link_tag;
int mode;
int block_start;
int n_links;
int link_set_n;
{
  NETWORK *net;
  LAYER *layer;
  int i;
  int n_source_units;	/* # of units in the source layer */
  int same_layer=0;	/* 1 if we are making connections in the same layer */
  int connectivity=mode&(CONNECT_COMPLETE|CONNECT_BLOCK|CONNECT_PARTIAL);
  LINK_SET *new_link_set;
  UNIT_INFO *info;
  n_links_built = 0;	/* global int, records the total # of links built */

  /***
   * Assert the things we can do
   */
  assertm(  (mode&CONNECT_COMPLETE)&&(mode&UNI_DIRECTIONAL)
	  ||(mode&CONNECT_PARTIAL)&&(mode&UNI_DIRECTIONAL)
	  ||(mode&CONNECT_BLOCK)&&(mode&UNI_DIRECTIONAL),
	  "not yet implemented in ConnectUnit_");
  if (NetParametersOk("ConnectUnit_",s_net,s_layer,0,d_net,d_layer,d_unit)) return -1;
  if ((mode & (CONNECT_PARTIAL|CONNECT_BLOCK)) && n_links<=0) {
    warn("ConnectUnit_","requested % of links is <= 0?",n_links);
    return -1;
  }

  n_source_units = neta[s_net]->layer[s_layer].n_units;

  /***
   * Connect the two layers with INWARD links.
   * A link set is created for each unit.
   */
  net = neta[d_net];
  layer = net->layer+d_layer;
  if (mode & CONNECT_COMPLETE) n_links = n_source_units;

  info = layer->info+d_unit;

  /***
   * make another link set and put it at the end of the list
   */
  new_link_set = Calloc(LINK_SET,1);
  if (new_link_set==NULL) {
    warn("ConnectUnit_","out of memory for dest link sets, requested %d",
	 sizeof(LINK_SET));
    goto out_of_memory;
  }
  layer->reg_connect = 0;
  n_links_built += n_links;
  new_link_set->connect_layer = s_layer;
  new_link_set->connect_net = s_net;
  new_link_set->mode = INWARD | connectivity | LINK_VECTOR;
  new_link_set->starting_unit_n = block_start;
  new_link_set->n_links = n_links;
  new_link_set->stride = 1;
  new_link_set->unit_id = NULL;
  new_link_set->link_tag = link_tag;
  AddLinkSet(info,new_link_set,-1);
  /***
   * need a vector of links - allocate it
   */
  new_link_set->l.p =
    (LINK_PTR)calloc(n_links,sizeof_link[link_tag]);
  if (new_link_set->l.p==NULL) {
    warn("ConnectUnit_","out of memory for links (requested %d * %d)",
	 n_links,sizeof_link[link_tag]);
    goto out_of_memory;
  }
  /***
   * allocate the vector of unit-id's, if necessary
   */
  if (mode & CONNECT_PARTIAL) {
    new_link_set->unit_id = Calloc(short,n_links);
    if (new_link_set->unit_id==NULL) {
      warn("ConnectUnit_","out of memory for link id's (requested %d * %d)",
	   n_links,sizeof(short));
      goto out_of_memory;
    }
    for (i=0;i<n_links;i++)
      new_link_set->unit_id[i] = i+new_link_set->starting_unit_n;
  }

  net->n_links += n_links_built;
  if (link_sets!=NULL) free(link_sets);
  return n_links_built;

 out_of_memory:
  if (link_sets!=NULL) free(link_sets);
  net->build_error=1;
  return -1;
}

int NetParametersOk(whofor,s_net,s_layer,s_unit,d_net,d_layer,d_unit)
char *whofor;
{
  if (s_net>n_nets || s_net<0) {
    warn(whofor,"s_net=%d out of range",s_net);
    return 0;
  }
  if (d_net>n_nets || d_net<0) {
    warn(whofor,"d_net=%d out of range",d_net);
    return 0;
  }
  if (!NetExists(neta[d_net])) {
    warn(whofor,"net[%d] is not built",d_net);
    return 0;
  }
  if (!NetExists(neta[s_net])) {
    warn(whofor,"net[%d] is not built",s_net);
    return 0;
  }
  if (s_layer<0 || s_layer>=neta[s_net]->n_layers) {
    warn(whofor,"s_layer(%d) is not in net[%d]",s_layer,s_net);
    return 0;
  }
  if (d_layer<0 || d_layer>=neta[d_net]->n_layers) {
    warn(whofor,"d_layer(%d) is not in net[%d]",d_layer,d_net);
    return 0;
  }
  if (d_unit<0 || d_unit>=neta[d_net]->layer[d_layer].n_units) {
    warn(whofor,"d_unit(%d) is not in layer %d of net[%d]",d_unit,d_layer,d_net);
    return 0;
  }
  if (s_unit<0 || s_unit>=neta[s_net]->layer[s_layer].n_units) {
    warn(whofor,"s_unit(%d) is not in layer %d of net[%d]",s_unit,s_layer,s_net);
    return 0;
  }
  return 1;
}
