/*
 * $Id: build.c,v 1.1 90/12/07 12:00:34 drew Exp Locker: drew $
 * 
 * $Log:	build.c,v $
 * Revision 1.1  90/12/07  12:00:34  drew
 * Initial revision
 * 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <itf/itf.h>
#include "sim-dummy.h"
#include <sim.h>

int n_nets=0;
int verbosity=0;
int abort_on_error=1;
int inhibit_deactivate=0;
NETWORK **neta;
NETWORK *net=NULL; /* the ``current'' network, can be used and changed at will */
struct PARAMS *p=NULL;
struct NET_PARAMS *np=NULL;
struct LAYER_PARAMS **lp=NULL;
int n_lps;

char sim_buf[SIMBUFSIZE];
int n_links_built;

void free_imatrix(m,nrl,nrh,ncl,nch)
int **m;
int nrl,nrh,ncl,nch;
{
	int i;

	for(i=nrh;i>=nrl;i--) free((char*) (m[i]+ncl));
	free((char*) (m+nrl));
}

int **imatrix(nrl,nrh,ncl,nch)
int nrl,nrh,ncl,nch;
{
	int i,**m;

	m=(int **)malloc((unsigned) (nrh-nrl+1)*sizeof(int*));
	if (!m) panic("imatrix","allocation failure 1");
	m -= nrl;

	for(i=nrl;i<=nrh;i++) {
		m[i]=(int *)malloc((unsigned) (nch-ncl+1)*sizeof(int));
		if (!m[i]) panic("imatrix","allocation failure 2");
		m[i] -= ncl;
	}
	return m;
}

autoinit_build()
{
  p = (struct PARAMS*)calloc(sizeof_params,1);
  ISetDefaults("PARAMS",(char*)p);
}

int FindOrBuildNetSkel(name)
string name;
{
  NETWORK *new_net = NULL;
  int net_n;
  char anon_name[10];
  if (name!=NULL) {
    new_net = (NETWORK*)IGetObjectPtr("NETWORK",name);
    if (new_net==NULL) {
      if (itf_errno!=UNKNOWN_VAR) IAbort();
      if (!IIsLegalVarName(name)) {
	itf_errno = ILLEGAL_VARNAME;
	IErrorAbort("%s",name);
      }
    }
  }
   
  if (new_net!=NULL) return new_net->my_n;

  /* otherwise find a new one */
  net_n = FindNullNetA(name);
  if (net_n<0) ABORTBUILD(NULL);
  if (neta[net_n]==NULL) {
    new_net = neta[net_n] = Calloc(NETWORK,1);
    new_net->name = NULL;
  } else
    new_net = neta[net_n];
  if (!new_net) {
    IError("out of memory for nets");
    ABORTBUILD(NULL);
  }
  new_net->my_n = net_n;
  new_net->builder = NULL;
  new_net->p = NULL;
  new_net->n_links = 0;
  new_net->build_error = 0;
  ISetDefaults("NETWORK",new_net);
  if (new_net->name==NULL) {
    if (name==NULL)
      sprintf(name=anon_name,"net%d",net_n);
    if ((new_net->name
	 =IBindStruct(name,"NETWORK",(char*)new_net,P_OBJECT,NULL)) == NULL) {
      IError("the name \"%s\" is in use",name);
      ABORTBUILD(new_net);
    }
  }
  return net_n;
}

int BuildNet(n_layers,name,type,params,builder)
int n_layers;
char *name;
int type;
struct NET_PARAMS *params;
char *builder;
{
  NETWORK *new_net;
  int net_n,i;
  if (n_layers<=0) {
    IError("BuildNet: bad number of layers (%d)",n_layers);
    ABORTBUILD(NULL);
  }
  net_n = FindOrBuildNetSkel(name);

  if (net_n<0) {
    IError("BuildNet: cannot find a null slot for new net");
    ABORTBUILD(NULL);
  }
  new_net = neta[net_n];
  if (NetExists(new_net)) {
    IError("BuildNet: net \"%s\" is already built",name);
    ABORTBUILD(NULL);
  }
  new_net->layer = Calloc(LAYER,n_layers);
  if (!new_net->layer) {
    IOutOfMemory("BuildNet","allocating %d layers",n_layers);
    ABORTBUILD(new_net);
  }
  new_net->builder = malloc(strlen(builder)+1);
  if (!new_net->layer) {
    IOutOfMemory("BuildNet","allocating builder string (%d) chars",strlen(builder)+1);
    ABORTBUILD(new_net);
  }
  strcpy(new_net->builder,builder);
  for (i=0;i<n_layers;i++)
    ISetDefaults("LAYER",(char*)(new_net->layer+i));
  new_net->n_layers = n_layers;
  if (params!=NULL)
    new_net->p = params;
  else {
    new_net->p = (struct NET_PARAMS*)calloc(sizeof_net_params,1);
    ISetDefaults("NET_PARAMS",(char*)new_net->p);
  }
  if (!new_net->p) {
    IOutOfMemory("BuildNet","allocating params");
    ABORTBUILD(new_net);
  }
  net=new_net;
  np = net->p;
  return net_n;
}

int FindNullNetA(name)
char *name;
{
  int net_n;
  for (net_n=0;net_n<n_nets;net_n++)
    if (   neta[net_n]==NULL
	|| (    neta[net_n]->p==NULL
	    && (name==NULL || !strcmp(name,neta[net_n]->name)))) break;
  if (net_n>=n_nets) net_n = LengthenNetA();
  return net_n;
}

int LengthenNetA()
{
  int net_n;
  if (n_nets>0)
    neta = (NETWORK**)realloc(neta,(n_nets+1)*sizeof(NETWORK**));
  else
    neta = Calloc(NETWORK*,1);
  if (neta==NULL) {
    IOutOfMemory("BuildNet","reallocating neta");
    ABORTBUILD(NULL);
  }
  /* zero out the newly allocated memory */
  neta[n_nets] = NULL;
  net_n = n_nets;
  n_nets+=1;
  return net_n;
}

/***
 * Return 1 if OK, -1 in case of error
 */
int BuildLayer_(net,layer_n,n_units,unit_tag,params)
NETWORK *net;
int n_units,unit_tag;
struct LAYER_PARAMS *params;
{
  LAYER *layer;
  if (net==NULL || net->layer==NULL) {
    warn("BuildLayer_","null network or null layers!");
    ABORTBUILD(NULL);
  }
  if (n_units<0) {
    warn("BuildLayer_","bad number of units (%d)",n_units);
    ABORTBUILD(net);
  }
  if (layer_n<0 || layer_n>=net->n_layers) {
    warn("BuildLayer_","layer %d out of range [%d,%d]",layer_n,0,net->n_layers);
    ABORTBUILD(net);
  }
  if (unit_tag<1 || unit_tag>n_unit_tags) {
    warn("BuildLayer_","bad unit tag %d",unit_tag);
    ABORTBUILD(net);
  }
  layer = net->layer+layer_n;
  if (layer->n_units!=0) {
    warn("BuildLayer_","layer already constructed (%d)",layer_n);
    ABORTBUILD(net);
  }
  if (   !(layer->unit = (UNIT_PTR)calloc(sizeof_unit[unit_tag],n_units))
      || !(layer->info = Calloc(UNIT_INFO,n_units))) {
    IOutOfMemory("BuildLayer_","allocating %d units or unit info",n_units);
    ABORTBUILD(net);
  }
  layer->n_units = n_units;
  layer->unit_tag = unit_tag;
  layer->auto_built = 0;
  layer->reg_connect = 1;
  layer->format.rows = 1;
  layer->format.cols = n_units;
  layer->format.code = 'x';
  if (params) layer->p = params;
  else layer->p = (struct LAYER_PARAMS*)calloc(sizeof_layer_params,1L);
  if (!layer->p) {
    IOutOfMemory("BuildLayer_","allocating params");
    net->build_error = 1;
    ABORTBUILD(net);
  }
  ISetDefaults("LAYER_PARAMS",(char*)layer->p);
  ISetVectorDefaults(nameof_unit[unit_tag],layer->unit,n_units);
  return 1;
}

/***
 * Connect two layers - s(ource)_layer and d(estination)_layer
 * Return the number of links built (counting bidirectional links
 * as one link, not two), or -1 in case of error
 */

int Connect_(s_net,s_layer,d_net,d_layer,
	     link_tag,mode,block_start,block_size,group_n)
int s_net,d_net;
int s_layer,d_layer;
int link_tag;
int mode;
int block_start;
int block_size;
int group_n;
{
  NETWORK *net = neta[d_net];
  LAYER *layer;
  int i;
  int unit_n;
  int n_source_units;	/* # of units in the source layer */
  int n_dest_units;	/* # of units we are connecting to in the destination layer */
  int same_layer=0;	/* 1 if we are making connections in the same layer */
  char *link_matrix=NULL; /* used if we get our links as a matrix */
  char *link_vector=NULL; /* used to step through link_matrix */
  LINK_SET **link_sets=NULL;/* remember all the link sets on the dest layer */
  int connectivity=mode&(CONNECT_COMPLETE|CONNECT_BLOCK|CONNECT_PARTIAL);
  n_links_built = 0;	/* global int, records the total # of links built */

  /***
   * Assert the things we can do
   */
  assertm(  (mode&CONNECT_COMPLETE)&&(mode&UNI_DIRECTIONAL)
	  ||(mode&CONNECT_COMPLETE)&&(mode&BI_DIRECTIONAL)
	  ||(mode&CONNECT_PARTIAL)&&(mode&UNI_DIRECTIONAL)
	  ||(mode&CONNECT_BLOCK)&&(mode&UNI_DIRECTIONAL)
	  ||(mode&CONNECT_BLOCK)&&(mode&BI_DIRECTIONAL),
	  "not yet implemented in Connect_");
  /***
   * check parameters
   */
  if (d_net>0 && d_net<n_nets) net = neta[d_net];
  else net = NULL;
  if (   !NetParametersOk("Connect_","source",0,s_net,0,s_layer,1,DFLT)
      || !NetParametersOk("Connect_","dest",0,d_net,0,d_layer,1,DFLT)
      )
    ABORTBUILD(net);
  if ((mode & (CONNECT_PARTIAL|CONNECT_BLOCK)) && block_size<=0) {
    IError("Connect_: requested # of links is <= 0?",block_size);
    ABORTBUILD(net);
  }
  if (link_tag<1 || link_tag>n_link_tags) {
    IError("Connect_: bad link tag %d",link_tag);
    ABORTBUILD(net);
  }

  if (d_layer==s_layer && d_net==s_net) same_layer = 1;
  else same_layer = 0;
  if (mode&CONNECT_COMPLETE)
    n_source_units = neta[s_net]->layer[s_layer].n_units;
  else
    n_source_units = block_size;
  n_dest_units = neta[d_net]->layer[d_layer].n_units;

  /***
   * Get a link matrix, if necessary
   */
  if ((mode & BI_DIRECTIONAL) && (mode & CONNECT_COMPLETE) & !same_layer) {
    /***
     * Try to cope with 16 bit ints & small maximum allocation.
     * If we can't allocate that much space in one go, we'll just use
     * more inefficient link storage methods - we will notice later that
     * link_matrix is null
     */
    long total_link_space =
      (long)n_source_units*(long)n_dest_units*(long)sizeof_link[link_tag];
    if (total_link_space < MAX_MALLOC) {
      link_matrix = calloc(n_source_units*n_dest_units,sizeof_link[link_tag]);
      link_vector = link_matrix;
    }
  }

  /***
   * Connect the two layers with INWARD links.
   * A link set is created for each unit.
   */
  net = neta[d_net];
  layer = net->layer+d_layer;
  if ((mode & CONNECT_BLOCK) || (mode & CONNECT_PARTIAL))
    layer->reg_connect = 0;
  if (mode & CONNECT_COMPLETE) {
    block_size = n_source_units;
    block_start = 0;
  }
  if (mode & BI_DIRECTIONAL) {
    link_sets = (LINK_SET**)calloc(n_dest_units,sizeof(LINK_SET*));
    if (link_sets==NULL) {
      IOutOfMemory("Connect_",
	   "allocating temporary link_sets vector, requested %d x %d",
	   n_dest_units,sizeof(LINK_SET*));
      goto out_of_memory;
    }
  }
  for (unit_n=0;unit_n<n_dest_units;unit_n++) {
    LINK_SET *new_link_set;
    UNIT_INFO *info;
    int n_links_in_this_set;
    info = layer->info+unit_n;
    if (   (mode & UNI_DIRECTIONAL)
	|| ((mode & BI_DIRECTIONAL) && !same_layer))
      n_links_in_this_set = block_size;
    else /* BI_DIRECTIONAL & same_layer */
      if (mode & RECURRENT_LINKS)
	n_links_in_this_set = unit_n + 1;
      else
	n_links_in_this_set = unit_n;
    if (n_links_in_this_set<=0) continue;
    /***
     * make another link set and put it at the end of the list
     */
    new_link_set = Calloc(LINK_SET,1);
    if (new_link_set==NULL) {
      IOutOfMemory("Connect_","allocating dest link sets, requested %d",
	   sizeof(LINK_SET));
      goto out_of_memory;
    }
    n_links_built += n_links_in_this_set;
    new_link_set->connect_layer = s_layer;
    new_link_set->connect_net = s_net;
    new_link_set->mode = INWARD | connectivity | LINK_VECTOR;
    if (mode & BI_DIRECTIONAL) new_link_set->mode |= BI_DIRECTIONAL;
    new_link_set->starting_unit_n = block_start;
    new_link_set->n_links = n_links_in_this_set;
    new_link_set->stride = 1;
    new_link_set->unit_id = NULL;
    new_link_set->link_tag = link_tag;
    if (!AddLinkSet(info,new_link_set,group_n)) goto out_of_memory;
    /***
     * need a vector of links - get it from the matrix, or allocate it
     */
    if (link_matrix!=NULL) {
      new_link_set->l.p = (LINK_PTR)link_vector;
      link_vector += n_links_in_this_set * sizeof_link[link_tag];
    } else { /* have to allocate */
      new_link_set->l.p =
	(LINK_PTR)calloc(n_links_in_this_set,sizeof_link[link_tag]);
      if (new_link_set->l.p==NULL) {
	IOutOfMemory("Connect_","allocating links (requested %d * %d)",
	     n_links_in_this_set,sizeof_link[link_tag]);
	goto out_of_memory;
      }
    }
    ISetVectorDefaults(nameof_link[link_tag],new_link_set->l.p,n_links_in_this_set);
    /***
     * allocate the vector of unit-id's, if necessary
     */
    if (mode & CONNECT_PARTIAL) {
      new_link_set->unit_id = Calloc(short,n_links_in_this_set);
      if (new_link_set->unit_id==NULL) {
	IOutOfMemory("Connect_","allocating link id's (requested %d * %d)",
	     n_links_in_this_set,sizeof(short));
	goto out_of_memory;
      }
      for (i=0;i<n_links_in_this_set;i++)
	new_link_set->unit_id[i] = i+new_link_set->starting_unit_n;
    }
    if (link_sets!=NULL) link_sets[unit_n] = new_link_set;
  }

  /***
   * Connect the two layers with OUTWARD links
   */
  net = neta[s_net];
  layer = net->layer+s_layer;
  if (mode & BI_DIRECTIONAL) {
    if (same_layer) {
      /***
       * Putting the other direction (dest->source) for bidirectinal links
       * in the same layers.
       * These link sets are attached to the source units.
       */
      for (unit_n=block_start;unit_n<(n_source_units+block_start);unit_n++) {
	LINK_SET *new_link_set;
	UNIT_INFO *info;
	int n_links_in_this_set;
	int dest_unit_n,link_n;
	info = layer->info+unit_n;
	n_links_in_this_set = n_dest_units - 1 - unit_n;
	if (n_links_in_this_set<=0) continue;
	/***
	 * make another link set and put it at the end of the list
	 */
	new_link_set = Calloc(LINK_SET,1);
	if (new_link_set==NULL) {
	  IOutOfMemory("Connect_","allocating source link sets, requested %d",
	       sizeof(LINK_SET));
	  goto out_of_memory;
	}
	new_link_set->connect_layer = d_layer;
	new_link_set->connect_net = d_net;
	new_link_set->mode =
	  OUTWARD | BI_DIRECTIONAL | connectivity | LINK_POINTER_VECTOR;
	new_link_set->starting_unit_n = unit_n+1;
	new_link_set->n_links = n_links_in_this_set;
	new_link_set->stride = 1;
	new_link_set->unit_id = NULL;
	new_link_set->link_tag = link_tag;
	if (!AddLinkSet(info,new_link_set,group_n)) goto out_of_memory;
	/***
	 * need a vector of pointers to links
	 */
	new_link_set->l.pp =
	  (LINK_PTR*)calloc(n_links_in_this_set,sizeof(LINK_PTR));
	if (new_link_set->l.pp==NULL) {
	  IOutOfMemory("Connect_", "allocating VPS (requested %d * %d)",
	       n_links_in_this_set,sizeof(LINK_PTR));
	  goto out_of_memory;
	}
	for (link_n=0;link_n<n_links_in_this_set;link_n++) {
	  dest_unit_n = new_link_set->starting_unit_n + link_n;
	  new_link_set->l.pp[link_n] =
	    (LINK_PTR) (link_sets[dest_unit_n]->l.cp
			+ unit_n*sizeof_link[link_tag]);
	}
      }
    } else {
      /***
       * Putting the other direction (dest->source) for bidirectinal links
       * between different layers.
       * We are building link sets on the source units
       */
      for (unit_n=block_start;unit_n<(n_source_units+block_start);unit_n++) {
	LINK_SET *new_link_set;
	UNIT_INFO *info;
	int n_links_in_this_set;
	info = layer->info+unit_n;
	n_links_in_this_set = n_dest_units;
	/***
	 * make another link set and put it at the end of the list
	 */
	new_link_set = Calloc(LINK_SET,1);
	if (new_link_set==NULL) {
	  IOutOfMemory("Connect_","allocating source link sets, requested %d",
	       sizeof(LINK_SET));
	  goto out_of_memory;
	}
	new_link_set->connect_layer = d_layer;
	new_link_set->connect_net = d_net;
	new_link_set->mode = OUTWARD | BI_DIRECTIONAL | connectivity;
	new_link_set->starting_unit_n = 0;
	new_link_set->n_links = n_dest_units;
	new_link_set->unit_id = NULL;
	new_link_set->link_tag = link_tag;
	if (!AddLinkSet(info,new_link_set,group_n)) goto out_of_memory;
	/***
	 * The links can done in two ways:
	 * a vector of links with large stride from the matrix,
	 * or a vector of pointers to links.
	 */
	if (link_matrix!=NULL) {
	  /***
	   * Vector from the matrix
	   */
	  new_link_set->stride = n_dest_units;
	  new_link_set->mode |= LINK_VECTOR;
	  new_link_set->l.p = ((LINK_PTR)link_matrix) + (unit_n-block_start);
	  link_vector += n_links_in_this_set * sizeof_link[link_tag];
	} else {
	  /***
	   * Allocate a vector of pointers to links and set them up
	   */
	  int dest_unit_n;
	  new_link_set->stride = 1;
	  new_link_set->mode |= LINK_POINTER_VECTOR;
	  new_link_set->l.pp =
	    (LINK_PTR*)calloc(n_links_in_this_set,sizeof(LINK_PTR));
	  if (new_link_set->l.pp==NULL) {
	    IOutOfMemory("Connect_", "allocating VPS (requested %d * %d)",
		 n_links_in_this_set,sizeof(LINK_PTR));
	    goto out_of_memory;
	  }
	  for (dest_unit_n=0;dest_unit_n<n_links_in_this_set;dest_unit_n++)
	    new_link_set->l.pp[dest_unit_n] =
	      (LINK_PTR) (link_sets[dest_unit_n]->l.cp
			  + (unit_n-block_start)*sizeof_link[link_tag]);
	}
      }
    }
  }
  net->n_links += n_links_built;
  net->n_unique_links = net->n_links;
  if (link_sets!=NULL) free(link_sets);
  return n_links_built;

 out_of_memory:
  net->build_error=1;
  if (link_sets!=NULL) free(link_sets);
  ABORTBUILD(neta[d_net]);
}

/***
 * Completely connect a network with bidirectional links.
 * Return the number of links (which will be half the number
 * of traversible links, i.e. in a network with 5 units there
 * will be 10 links between different units.
 * To get recurrent links, or the flag RECURRENT_LINKS in with the mode.
 * Return -1 if we cannot do it.
 */
int ConnectBidirNet_(net_n,link_tag,mode,n_links)
int net_n;
{
  int s_layer,d_layer;
  int total_n_links=0;
  assertm(mode==CONNECT_COMPLETE, "ConnectBidirNet: bad mode");
  if (net_n<0 || net_n>=n_nets || neta[net_n]==NULL) {
    warn("ConnectCompleteBidirNet","net is NULL");
    ABORTBUILD(NULL);
  }
  if (neta[net_n]->n_layers<=0) {
    warn("ConnectCompleteBidirNet","net has no layers");
    ABORTBUILD(neta[net_n]);
  }
  for (s_layer=0;s_layer<neta[net_n]->n_layers;s_layer++)
    for (d_layer=s_layer;d_layer<neta[net_n]->n_layers;d_layer++) {
      if (Connect_(net_n,s_layer,net_n,d_layer,link_tag,
		   mode|BI_DIRECTIONAL,n_links,0,0) < 0)
	ABORTBUILD(neta[net_n]);
      total_n_links += n_links_built;
    }
  return total_n_links;
}

#define OkLayer(layer_n) ((layer_n)>=0 && (layer_n)<n_layers)

/***
 * Connect a network with connectivity specified on a layer-by-layer
 * basis.  Links can be bi-directional or uni-directional.
 * Default connectivity is nothing connected.
 * Default connectivity is complete, without units linking to themselves
 * (except if links are uni-directional).
 * When bi-directional links are used, links belong to higher layer.
 *
 * Return the number of links (which will be half the number
 * of traversible links, i.e. in a network with 5 units there
 * will be 15 links - 10 links between different units and
 * 5 recurrent links)
 * Return -1 if we cannot do it.
 *
 * The connect_str can have the following components, prefixed with + or -
 *                                  Source layer         Destination layer
 *     *-*        :                     all                    all
 *     x-y        :                     x                      y
 *     x-*        :                     x                      all
 *     *-y        :                     all                    y
 *     n-n        : for all n           n                      n
 *     n-n-i      : for all n           n                      n-i
 *     n-n+i      : for all n           n                      n+i
 *     n->=n+i    : for all n,j>0       n                      n+i+j
 *     n->=n-i    : for all n,j>0       n                      n-i+j
 *     n-<=n+i    : for all n,j>0       n                      n+i-j
 *     n-<=n-i    : for all n,j>0       n                      n-i-j
 *     +/-b       : the following components should/not apply to the bias layer
 */
int ConnectCompleteNet_(net_n,link_tag,mode,connect_str)
int net_n;
int link_tag;
int mode;
char *connect_str;
{
  int s_layer,d_layer,n_layers,i,j,on;
  int total_n_links = -1;
  int **connect;
  int tokc;
  int top_layer;
  LEX_STACK stack;
  char **tokv,c,d,s,*tok;
  assertm(!(mode&(CONNECT_PARTIAL|CONNECT_BLOCK)), "ConnectCompleteNet_: bad mode");
  assertm((mode&(BI_DIRECTIONAL|UNI_DIRECTIONAL)), "ConnectCompleteNet_: bad mode");
  if (net_n<0 || net_n>=n_nets || neta[net_n]==NULL) {
    warn("ConnectCompleteNet_","net is NULL");
    ABORTBUILD(NULL);
  }
  n_layers = neta[net_n]->n_layers;
  if (n_layers<=0) {
    warn("ConnectCompleteNet_","net has no layers");
    ABORTBUILD(neta[net_n]);
  }
  /***
   * Initialize the connect matrix for complete connection if no connect
   * string was provided
   */
  connect = ITempMatrix(int,0,n_layers-1,0,n_layers-1);
  n_layers = neta[net_n]->n_layers-1; /* ignore bias layer */
  top_layer = n_layers-1;
  for (s_layer=0;s_layer<n_layers;s_layer++)
    for (d_layer=0;d_layer<n_layers;d_layer++)
      connect[s_layer][d_layer] = 
        connect_str==NULL && (d_layer>=s_layer || (mode&UNI_DIRECTIONAL));
  /* and now parse the connect string */
  if (connect_str!=NULL) {
    if (verbosity)
      fprintf(dout,"Connecting net with specifiers \"%s\"\n",connect_str);
    tokv = LexAnalysePush(simple_syntax,connect_str,&tokc,&stack);
    if (tokv==NULL) goto getout;
    for (;*tokv;tokv++) {
      on = **tokv!='-';
      tok = *tokv;
      if (*tok=='-' || *tok=='+') tok++;
      if (sscanf(tok,"%d-%d",&s_layer,&d_layer)==2
	  && OkLayer(s_layer) && OkLayer(d_layer))
	connect[s_layer][d_layer] = on;
      else if (sscanf(tok,"$-%d-%d",&i,&d_layer)==2
	  && OkLayer(s_layer=top_layer-i) && OkLayer(d_layer))
	connect[s_layer][d_layer] = on;
      else if (sscanf(tok,"$-%d",&d_layer)==1
	  && OkLayer(s_layer=top_layer) && OkLayer(d_layer))
	connect[s_layer][d_layer] = on;
      else if (sscanf(tok,"$-%d-$-%d",&i,&j)==2
	  && OkLayer(s_layer=top_layer-i) && OkLayer(d_layer=top_layer-j))
	connect[s_layer][d_layer] = on;
      else if (!strcmp(tok,"$-$")
	  && OkLayer(s_layer=top_layer) && OkLayer(d_layer=top_layer))
	connect[s_layer][d_layer] = on;
      else if (sscanf(tok,"%d-$-%d",&s_layer,&j)==2
	  && OkLayer(s_layer) && OkLayer(d_layer=top_layer-j))
	connect[s_layer][d_layer] = on;
      else if (sscanf(tok,"%d-%c",&s_layer,&c)==2 && c=='$'
	  && OkLayer(s_layer) && OkLayer(d_layer=top_layer))
	connect[s_layer][d_layer] = on;
      else if (!strcmp(*tokv,"-b"))
	n_layers = neta[net_n]->n_layers-1;
      else if (!strcmp(*tokv,"+b"))
	n_layers = neta[net_n]->n_layers;
      else if (sscanf(tok,"%d-%c",&s_layer,&c)==2
	       && c=='*' && OkLayer(s_layer))
	for (d_layer=0;d_layer<n_layers;d_layer++)
	  connect[s_layer][d_layer] = on;
      else if (sscanf(tok,"$-%d-%c",&i,&c)==2
	       && c=='*' && OkLayer(s_layer=top_layer-i))
	for (d_layer=0;d_layer<n_layers;d_layer++)
	  connect[s_layer][d_layer] = on;
      else if (sscanf(tok,"%c-%d",&c,&d_layer)==2
	       && c=='*' && OkLayer(d_layer))
	for (s_layer=0;s_layer<n_layers;s_layer++)
	  connect[s_layer][d_layer] = on;
      else if (sscanf(tok,"%c-$-%d",&c,&j)==2
	       && c=='*' && OkLayer(d_layer=top_layer-j))
	for (s_layer=0;s_layer<n_layers;s_layer++)
	  connect[s_layer][d_layer] = on;
      else if (!strcmp(tok,"$-*"))
	for (d_layer=0;d_layer<n_layers;d_layer++)
	  connect[top_layer][d_layer] = on;
      else if (!strcmp(tok,"*-$"))
	for (s_layer=0;s_layer<n_layers;s_layer++)
	  connect[s_layer][top_layer] = on;
      else if (sscanf(tok,"n-n+%d",&i)==1||sscanf(tok,"n-n%d",&i)==1) {
	for (s_layer=0;s_layer<n_layers;s_layer++)
	  if (OkLayer(s_layer+i))
	    connect[s_layer][s_layer+i] = on;
      } else if (sscanf(tok,"n-%c=n%c%d",&d,&s,&i)==3
	       &&(d=='>'||d=='<') && (s=='+'||s=='-')) {
	if (s=='-') i = -i;
	if (d=='>') {
	  for (s_layer=0;s_layer<n_layers;s_layer++)
	    for (d_layer=s_layer+i;d_layer<n_layers;d_layer++)
	      if (OkLayer(d_layer))
		connect[s_layer][d_layer] = on;
	} else {
	  for (s_layer=0;s_layer<n_layers;s_layer++)
	    for (d_layer=s_layer+i;d_layer>=0;d_layer--)
	      if (OkLayer(d_layer))
		connect[s_layer][d_layer] = on;
	}
      } else if (!strcmp(tok,"*-*"))
	for (s_layer=0;s_layer<n_layers;s_layer++)
	  for (d_layer=0;d_layer<n_layers;d_layer++)
	    if (!on)
	      connect[s_layer][d_layer] = 1;
	    else
	      connect[s_layer][d_layer] = (d_layer>=s_layer || (mode&UNI_DIRECTIONAL));
      else if (!strcmp(tok,"n-n"))
	for (s_layer=0;s_layer<n_layers;s_layer++)
	  connect[s_layer][s_layer] = on;
      else {
	IError("ConnectCompleteNet_: bad connect specifier \"%s\"",*tokv);
	goto getout;
      }
    }
  }

  n_layers = neta[net_n]->n_layers;
  total_n_links = 0;
  for (s_layer=0;s_layer<n_layers;s_layer++)
    for (d_layer=0;d_layer<n_layers;d_layer++) {
      if (connect[s_layer][d_layer]) {
	if (verbosity>0)
	  fprintf(dout,"Connecting slayer %d to dlayer %d\n",s_layer,d_layer);
	if (Connect_(net_n,s_layer,net_n,d_layer,link_tag,mode,0,0,0) < 0) {
	  total_n_links = -1;
	  goto getout;
	}
	total_n_links += n_links_built;
      }
    }
 getout:
  LexAnalysePop(&stack);
  return total_n_links;
}

/***
 * Connect a unit to a layer - s(ource)_layer and d(estination)_layer
 * Return the number of links built, or -1 in case of error
 */

int ConnectUnit_(s_net,s_layer,d_net,d_layer,d_unit,
		 link_tag,mode,block_start,block_size,group_n)
int s_net;
int s_layer;
int d_net;
int d_layer;
int d_unit;
int link_tag;
int mode;
int block_size;
int block_start;
int group_n;
{
  NETWORK *net;
  LAYER *layer;
  int i;
  int n_source_units;	/* # of units in the source layer */
  /* int same_layer=0;*//* 1 if we are making connections in the same layer */
  int connectivity=mode&(CONNECT_COMPLETE|CONNECT_BLOCK|CONNECT_PARTIAL);
  LINK_SET *new_link_set;
  UNIT_INFO *info;
  n_links_built = 0;	/* global int, records the total # of links built */

  /***
   * Assert the things we can do
   */
  assertm(  (mode&CONNECT_COMPLETE)&&(mode&UNI_DIRECTIONAL)
	  ||(mode&CONNECT_PARTIAL)&&(mode&UNI_DIRECTIONAL)
	  ||(mode&CONNECT_BLOCK)&&(mode&UNI_DIRECTIONAL),
	  "not yet implemented in ConnectUnit_");
  if (d_net>0 && d_net<n_nets) net = neta[d_net];
  else net = NULL;
  if (   !NetParametersOk("ConnectUnit_","source",0,s_net,0,s_layer,1,DFLT)
      || !NetParametersOk("ConnectUnit_","dest",0,d_net,0,d_layer,0,d_unit))
    ABORTBUILD(net);
  if ((mode & (CONNECT_PARTIAL|CONNECT_BLOCK)) && block_size<=0) {
    warn("ConnectUnit_","requested % of links is <= 0?",block_size);
    ABORTBUILD(net);
  }
  if (link_tag<1 || link_tag>n_link_tags) {
    warn("ConnectUnit_","bad link tag %d",link_tag);
    ABORTBUILD(net);
  }

  n_source_units = neta[s_net]->layer[s_layer].n_units;

  /***
   * Connect the two layers with INWARD links.
   * A link set is created for each unit.
   */
  net = neta[d_net];
  layer = net->layer+d_layer;
  if (mode & CONNECT_COMPLETE) block_size = n_source_units;

  info = layer->info+d_unit;

  /***
   * make another link set and put it at the end of the list
   */
  new_link_set = Calloc(LINK_SET,1);
  if (new_link_set==NULL) {
    IOutOfMemory("ConnectUnit_","allocating dest link set");
    goto out_of_memory;
  }
  layer->reg_connect = 0;
  n_links_built += block_size;
  new_link_set->connect_layer = s_layer;
  new_link_set->connect_net = s_net;
  new_link_set->mode = INWARD | connectivity | LINK_VECTOR;
  new_link_set->starting_unit_n = block_start;
  new_link_set->n_links = block_size;
  new_link_set->stride = 1;
  new_link_set->unit_id = NULL;
  new_link_set->link_tag = link_tag;
  if (!AddLinkSet(info,new_link_set,group_n)) goto out_of_memory;
  /***
   * need a vector of links - allocate it
   */
  new_link_set->l.p =
    (LINK_PTR)calloc(block_size,sizeof_link[link_tag]);
  if (new_link_set->l.p==NULL) {
    IOutOfMemory("ConnectUnit_","allocating links (requested %d * %d)",
	 block_size,sizeof_link[link_tag]);
    goto out_of_memory;
  }
  ISetVectorDefaults(nameof_link[link_tag],new_link_set->l.p,block_size);
  /***
   * allocate the vector of unit-id's, if necessary
   */
  if (mode & CONNECT_PARTIAL) {
    new_link_set->unit_id = Calloc(short,block_size);
    if (new_link_set->unit_id==NULL) {
      IOutOfMemory("ConnectUnit_","allocating link id's (requested %d * %d)",
	   block_size,sizeof(short));
      goto out_of_memory;
    }
    for (i=0;i<block_size;i++)
      new_link_set->unit_id[i] = i+new_link_set->starting_unit_n;
  }

  net->n_links += n_links_built;
  net->n_unique_links = net->n_links;	/* not correct yet */
  return n_links_built;

 out_of_memory:
  ABORTBUILD(net);
}

/***
 * Checks that the Net Parameters are ok - that net, layer, and unit
 * numbers are in range.  The *_ad flags tell whether it is ok
 * to allow the default (DFLT) value for that number.
 * whofor=function name, e.g.: "Connect"
 * what=purpose name,    e.g.: "destination"
 */
int NetParametersOk(whofor,what,net_ad,net_n,layer_ad,layer_n,unit_ad,unit_n)
char *whofor,*what;
{
  if (net_n>n_nets || net_n<(net_ad?DFLT:0)) {
    IError("%s: %s net=%d out of range",whofor,what,net_n);
    return 0;
  }
  if (net_n==DFLT) return 1;

  
  if (!NetExists(neta[net_n])) {
    IError("%s: %s net[%d] is not built or had a build error",
	   whofor,what,net_n);
    return 0;
  }

  if (layer_n<(layer_ad?DFLT:0) || layer_n>=neta[net_n]->n_layers) {
    IError("%s: %s layer(%d) is not in \"%s\"",
	   whofor,what,layer_n,neta[net_n]->name);
    return 0;
  }

  if (layer_n==DFLT) return 1;

  if (unit_n<(unit_ad?DFLT:0) || unit_n>=neta[net_n]->layer[layer_n].n_units) {
    IError("%s: %s unit(%d) is not in layer %d of \"%s\"",
	   whofor,what,unit_n,layer_n,neta[net_n]->name);
    return 0;
  }
  return 1;
}

/***
 * Add a link set to a unit_info record - add it at the tail
 * Three possibilities for group_n:
 * group_n==-1 : use group_n of previous link_set + 1
 * group_n==0  : use group_n of previous link_set
 * group_n>0   : use this value as group_n
 */
AddLinkSet(info,new_link_set,group_n)
UNIT_INFO *info;
LINK_SET *new_link_set;
int group_n;
{
  int i,j;
  assertm(group_n>=-1,"AddLinkSet: group_n out of range");
  if (info->ls==NULL) {
    info->ls = Calloc(LINK_SET*,1);
    if (info->ls==NULL) goto out_of_memory;
    info->n_lss = 1;
    info->last_group_n = new_link_set->group_n = (group_n>0 ? group_n : 1);
    info->ls[0] = new_link_set;
  } else {
    info->n_lss++;
    info->ls = Realloc(LINK_SET*,info->ls,info->n_lss);
    if (info->ls==NULL) goto out_of_memory;
    group_n =(group_n>0  ? group_n:
	      group_n==0 ? info->last_group_n:
	                   info->last_group_n+1);
    new_link_set->group_n = info->last_group_n = group_n;
    /***
     * insert the new link set in the vector - in order of group_n, and
     * after old ones with the same group_n
     */
    for (i=0;i<info->n_lss-1;i++) if (info->ls[i]->group_n>group_n) break;
    /* i is the position where we want to insert, so move ones above up */
    for (j=info->n_lss-1;j>i;j--) info->ls[j] = info->ls[j-1];
    info->ls[i] = new_link_set;
  }
  info->fan_in += new_link_set->n_links;
  return 1;

  out_of_memory:
  IOutOfMemory("AddLinkSet","(re)allocating linkset vector %d",info->n_lss);
  return 0;
}

/***
 * Find a link
 * 1 is returned if the link is found, 0 if not
 * If the link is found, the ls pointer is set to point to the link set,
 * and link_n is set to the number of the link in that linkset
 * Assume that everything is in range - checked by asserts
 */
int FindLink(s_net,d_net,s_layer,d_layer,s_unit,d_unit,out_ls,link_n)
int s_net,d_net,s_layer,d_layer,s_unit,d_unit;
LINK_SET **out_ls;
int *link_n;
{
  LINK_SET *ls;
  int i,j,ls_n; /* i is the # of the link */
  assert(s_net>=0 && s_net<n_nets && d_net>=0 && d_net<n_nets);
  assert(NetExists(neta[s_net]) && NetExists(neta[d_net]));
  assert(s_layer>=0 && s_layer<neta[s_net]->n_layers);
  assert(d_layer>=0 && d_layer<neta[d_net]->n_layers);
  assert(s_unit>=0 && s_unit<neta[s_net]->layer[s_layer].n_units);
  assert(d_unit>=0 && d_unit<neta[d_net]->layer[d_layer].n_units);
  /***
   * First look for a link in the destination layer
   */
  for (ls_n=0;ls_n<neta[d_net]->layer[d_layer].info[d_unit].n_lss;ls_n++) {
    ls = neta[d_net]->layer[d_layer].info[d_unit].ls[ls_n];
    /* -- check how to find the unit# -- */
    for (i=0;i<0;i++) j=ls->n_links; /* just a junk line to use things &
					keep lint happy! */
  }
  i = j;
  /***
   * No inward link in the destination layer - look for a
   * link in the source layer
   */
  return 0;
}

int LinkTag(type_name)
char *type_name;
{
  int tag;
  for (tag=1;tag<=n_link_tags;tag++)
    if (!strcmp(type_name,nameof_link[tag])) return tag;
  return 0;
}

int UnitTag(type_name)
char *type_name;
{
  int tag;
  for (tag=1;tag<=n_unit_tags;tag++)
    if (!strcmp(type_name,nameof_unit[tag])) return tag;
  return 0;
}

int SetLayerFormat(net,layer_n,rows,cols,code,join,lowzero)
NETWORK *net;
int layer_n;
int rows;
int cols;
char code;
int join;
int lowzero;
{
  if (!NetExists(net)) {
    IError("SetLayerFormat: net is null");
    return 0;
  }
  if (layer_n<0 || layer_n>=net->n_layers) {
    IError("SetLayerFormat: layer_n (%d) is out of range",layer_n);
    return 0;
  }
  if (code!='x') {
    IError("SetLayerFormat: bad code (%d) '%c'",code);
    return 0;
  }
  if (rows<=0 || cols<=0 || rows*cols != net->layer[layer_n].n_units) {
    IError("SetLayerFormat: rows*cols != n_units or rows or cols <= 0 (%dx%d!=%d)",
	   rows,cols,net->layer[layer_n].n_units);
    return 0;
  }
  net->layer[layer_n].format.rows = rows;
  net->layer[layer_n].format.cols = cols;
  net->layer[layer_n].format.code = code;
  net->layer[layer_n].format.join = join;
  net->layer[layer_n].format.lowzero = lowzero;
  return 1;
}

int SetLayerFormatStr(net,layer_n,str)
NETWORK *net;
int layer_n;
char *str;
{
  struct AREA_FORMAT format;
  if (!ParseAreaFormat(str,&format)) return 0;
  return SetLayerFormat(net,layer_n,format.rows,format.cols,format.code,
			format.join,format.lowzero);
}

int NetExists(net)
NETWORK *net;
{
  if (net==NULL || net->p==NULL || net->build_error) return 0;
  else return 1;
}

int CheckNetExists(net)
NETWORK *net;
{
  if (!NetExists(net)) {
    IError("net is not built or had a build error");
    return 0;
  } else
    return 1;
}

/***
 * UseNet(net) should be called by commands or functions that
 * change the net, np, or lp pointers.
 * UseNet sets up the np and lp pointers to point to the parameter
 * structures in the given net.
 */
int UseNet(this)
struct NETWORK *this;
{
  int i;
  if (net==NULL)
    IErrorAbort("Net is not built (net==NULL)");
  if (net->p==NULL)
    IErrorAbort("Net is not fully built (net->np==NULL)");
  if (net->build_error)
    IErrorAbort("Net had a build error and is an unusable state");

  net = this;
  np = net->p;
  if (net->n_layers<1 && lp!=NULL) {
    free(lp);
    lp = NULL;
    n_lps = 0;
  }
  if (net->n_layers>0) {
    if (lp==NULL)
      lp = CallocOrAbort(struct LAYER_PARAMS*, net->n_layers);
    else if (net->n_layers!=n_lps)
      lp = ReallocOrAbort(struct LAYER_PARAMS*, lp, net->n_layers);
    n_lps = net->n_layers;
    for (i=0; i<n_lps; i++)
      lp[i] = net->layer[i].p;
  }
}
