/*
 * $Id: network.h,v 1.1 90/12/07 12:00:54 drew Exp $
 * 
 * $Log:	network.h,v $
 * Revision 1.1  90/12/07  12:00:54  drew
 * Initial revision
 * 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

/***
 * network.h
 *
 * Structure declarations for networks.
 *
 * Needs:
 * UNIT_PTR
 * LINK_PTR
 * These can be pointer types or unions of pointer types.
 * The union shouldn't have any tag in it, because it will
 * be used many times.  There are tags in the appropriate
 * places already, they are integers. NOTE: These are created
 * automatically and put in "ul-tags.h".
 *
 * The user needs to declare one or more structures with
 * the keyword "UNIT" in a comment on the first line.  Similarily
 * with "LINK".  The shell program create-ul-tags will generate
 * code which declares UNIT_PTR and LINK_PTR.  The user does need to
 * supply the following three structures,
 *
 * LAYER_PARAMS
 * NET_PARAMS
 * PARAMS
 */

/***
 * The following are used as bits in the `mode' parameter of Connect_()
 * and in the method field in LINK_SET.
 * Exactly one of each of the pairs should be set.
 * If CONNECT_PARTIAL is on, LINK_SET.unit_id should be present.
 */
#define INWARD			0x0001
#define OUTWARD			0x0002

#define CONNECT_COMPLETE	0x0004
#define CONNECT_PARTIAL		0x0008
#define CONNECT_BLOCK		0x4000

#define BI_DIRECTIONAL		0x0010
#define UNI_DIRECTIONAL		0x0020

#define	LINK_VECTOR		0x0040
#define LINK_POINTER_VECTOR	0x0080

#define PARASITIC		0x0100	/* links shared with other linksets */
#define PARASITE_VICTIM		0x0200	/* links shared with other linksets */

#define RECURRENT_LINKS		0x0400	/* units link to themselves */

#define LINK_MATRIX_00		0x0800	/* beginning of a link matrix */
#define	LINK_MATRIX_ROW		0x1000	/* vector is a row from a matrix */
#define LINK_MATRIX_COL		0x2000	/* vector is a col from a matrix */

typedef struct LINK_SET {	/* BIND */
  short		connect_layer;	/* connects with this layer */
  short		connect_net;	/* in this network */
  short		mode;		/* see above for possible values */
  short		n_links;
  short		n_unique_links;
  short		starting_unit_n;
  short		stride;		/* in # of links, not bytes */
  short		link_tag;	/* for link_ptr in case it's a union */
  short		group_n;
  short		deactivate;	/* set to 1 to have ForEachLinkSet skip this */
  union {
    LINK_PTR	p;		/* to vector of links */
    LINK_PTR	*pp;		/* or vector of ptrs to links */
    char	*cp;		/* so that we can get at the value */
    char	**cpp;		/* so that we can get at the value */
  } 		l;
  short		*unit_id;	/* array: for randomly addressed links */
} LINK_SET;

typedef struct UNIT_INFO {	/* BIND */
  short		fan_in;
  short		n_lss;		/* the number of link sets */
  struct LINK_SET	**ls;	/* counter: n_lss array of link sets */
  short		last_group_n;
} UNIT_INFO;

typedef struct LAYER {		/* BIND */
  short		n_units;	/* nosave: */
  short		unit_tag;	/* nosave: */
  short		reg_connect;	/* nosave: true if all units in this */
				/* layer are connected the same */
  short		auto_built;	/* nosave: set to one if this layer is built */
  				/* automatically, and thus should not be */
				/* in a network description */
  struct LAYER_PARAMS	*p;
  struct AREA_FORMAT	format;
  UNIT_TYPE	unit;		/* nosave: counter: n_units */
  struct UNIT_INFO *info;	/* nosave: counter: n_units */
} LAYER;

typedef struct NETWORK {	/* BIND */
  string	name;		/* nosave: arbitrary */
  string	builder;	/* nosave: recreate desc, set by the builder */
  int		net_type;
  int		n_layers;	/* nosave: */
  int		my_n;		/* nosave: */
  int		n_links;	/* nosave: total # of links */
  int		n_unique_links;	/* nosave: */
  int		build_error;	/* nosave: some error occurred - untrustworthy */
  struct LAYER	*layer;		/* counter: n_layers */
  struct NET_PARAMS	*p;
} NETWORK;

extern struct NETWORK **neta;	/* BIND counter: n_nets */
extern struct NETWORK *net;	/* BIND */
extern int n_nets;		/* BIND */
extern struct PARAMS *p;	/* BIND */
extern struct NET_PARAMS *np;	/* BIND */
extern struct LAYER_PARAMS **lp;/* BIND counter: n_lps */
extern int n_lps;		/* BIND */
extern int verbosity;		/* BIND */
extern int inhibit_deactivate;	/* BIND */

#define LEARNING_PARAMETER 0x10L
#define BUILDING_PARAMETER 0x20L
#define METHOD_PARAMETER 0x40L
#define PIPE_PARAMETER 0x80L
#define RESULT_PARAMETER 0x100L

