/*
 * $Id: parasite.c,v 1.2 90/12/17 10:54:13 drew Exp $
 * 
 * $Log:	parasite.c,v $
 * Revision 1.2  90/12/17  10:54:13  drew
 * Commented and tried to clean it up.
 * 
 * Revision 1.1  90/12/07  12:00:56  drew
 * Initial revision
 * 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

/**********************************************************************
 *	File name:   parasite.c
 *	Description: contains functions that allow units to share links
 *			Good luck.
 *
 *	Exported functions:
 *		int	ConnectParasiteUnit()
 *		int	ConnectParasiteLayer()
 *
 *	Exported variables:
 *
 **********************************************************************/

#include <itf/itf.h>
#include "sim-dummy.h"
#include <sim.h>

/**********************************************************************
 *	Name:         ConnectParasiteUnit
 *	Description:  The connect parasite stuff allows the sharing of
 *			links, so that units actually share link memory.
 *	Parameters:
 *	s_net, s_layer, s_start_unit	: source location for the new links
 *	d_net, d_layer, d_unit	: dest location for the new links
 *	v_net, v_layer, v_unit	: victim location
 *	v_group,v_other_layer	: restrictions on the victim linksets
 *	v_link_tag		: ????
 *	new_link_ids		: 1=create new link_id vector
 *
 *	Return Value: int ConnectParasiteUnit - the number of links built
 **********************************************************************/
ConnectParasiteUnit(s_net, s_layer, s_start_unit, d_net, d_layer, d_unit,
		    v_net, v_layer, v_unit, v_group, v_other_layer, v_link_tag,
		    new_link_ids)
{
  LINK_SET	*new_ls,*v_ls;
  UNIT_INFO	*v_info,*d_info;
  NETWORK	*net;
  int		n_links=0;
  int		i,j;

  if (d_layer==DFLT) 	 d_layer = v_layer;
  else if (d_unit==DFLT) d_unit  = v_unit;

  if (d_net>0 && d_net<n_nets) net = neta[d_net];
  else			       net = NULL;

  if (   !NetParametersOk("ConnectParasiteUnit","victim",
			 0,v_net,0,v_layer,0,v_unit)
      || !NetParametersOk("ConnectParasiteUnit","dest",
			 0,d_net,0,d_layer,0,d_unit)
      || !NetParametersOk("ConnectParasiteUnit","source",
			 1,s_net,1,s_layer,1,s_start_unit))
    ABORTBUILD(net);

  v_info = neta[v_net]->layer[v_layer].info + v_unit;
  d_info = neta[d_net]->layer[d_layer].info + d_unit;

  for (i = 0 ; i < v_info->n_lss ; i++) {
    v_ls = v_info->ls[i];
    /* decide if we should parasite on this link set */
    if (   (v_group==DFLT || v_group==v_ls->group_n)
	&& (v_other_layer==DFLT || v_other_layer==v_ls->connect_layer)
	&& (v_link_tag==DFLT || v_link_tag==v_ls->link_tag)) {
      new_ls = Calloc(LINK_SET,1);
      if (new_ls==NULL) {
	IOutOfMemory("ConnectParasiteUnit","allocating link set");
	goto out_of_memory;
      }

      new_ls->connect_layer	= (s_layer==DFLT?v_ls->connect_layer:s_layer);
      new_ls->connect_net	= (s_net==DFLT?v_ls->connect_net:s_net);
      new_ls->mode		= v_ls->mode | PARASITIC;
      v_ls->mode		|= PARASITE_VICTIM;
      new_ls->n_links		= v_ls->n_links;
      new_ls->starting_unit_n	= (s_start_unit==DFLT ? v_ls->starting_unit_n
						      : s_start_unit);
      if (!NetParametersOk("ConnectParasiteUnit","source start",
			   0,v_ls->connect_net,0,v_ls->connect_layer,
			   0,new_ls->starting_unit_n+new_ls->n_links-1))
	ABORTBUILD(net);
      new_ls->stride		= v_ls->stride;
      new_ls->link_tag		= v_ls->link_tag;
      new_ls->group_n		= v_ls->group_n;
      new_ls->l.p		= v_ls->l.p;	/* gets all pointers - union */
      new_ls->unit_id		= v_ls->unit_id;
      if (v_ls->unit_id && new_link_ids) {
	new_ls->unit_id = Calloc(short,new_ls->n_links);
	if (new_ls->unit_id==NULL) {
	  IOutOfMemory("ConnectParasiteUnit","allocating unit_id %d",
		       new_ls->n_links);
	  goto out_of_memory;
	}
	for (j=0;j<new_ls->n_links;j++)
	  new_ls->unit_id[j] = v_ls->unit_id[j];
      }
      if (!AddLinkSet(d_info,new_ls,new_ls->group_n)) goto out_of_memory;
      n_links += new_ls->n_links;
    }
  }
  if (n_links==0) {
    IError("ConnectParasiteUnit: no victim links found");
    ABORTBUILD(net);
  }
  return n_links;

  out_of_memory:
  ABORTBUILD(net);
}
/**********************************************************************/

/**********************************************************************
 *	Name:         ConnectParasiteLayer
 *	Description:  The connect parasite stuff allows the sharing of
 *			links, so that units actually share link memory.
 *	Parameters:
 *	s_net, s_layer, s_start_unit	: source location for the new links
 *	d_net, d_layer		: dest location for the new links
 *	v_net, v_layer		: victim location
 *	v_group,v_other_layer	: restrictions on the victim linksets
 *	v_link_tag		: ????
 *	new_link_ids		: 1=create new link_id vector
 *
 *	Return Value: int ConnectParasiteLayer - the number of links built
 **********************************************************************/
ConnectParasiteLayer(s_net,s_layer,s_start_unit,
		     d_net,d_layer,
		     v_net,v_layer,v_group,
		     v_other_layer,v_link_tag,new_link_ids)
{
  int		i,ret=0,j;
  NETWORK	*net;
  if (d_net>0 && d_net<n_nets) net = neta[d_net]; 
  else			       net = NULL;

  if (   !NetParametersOk("ConnectParasiteLayer","victim",
			  0,v_net,0,v_layer,1,DFLT)
      || !NetParametersOk("ConnectParasiteLayer","dest",
			  0,d_net,0,d_layer,1,DFLT)
      || !NetParametersOk("ConnectParasiteLayer","source",
			  1,s_net,1,s_layer,1,s_start_unit))
    ABORTBUILD(net);

  if (neta[v_net]->layer[v_layer].n_units != 
      neta[d_net]->layer[d_layer].n_units) {
    IError("ConnectParasite: victim & dest layer have different # of units");
    ABORTBUILD(net);
  }
  for (i = 0 ; i < neta[v_net]->layer[v_layer].n_units ; i++) {
    j = ConnectParasiteUnit(s_net,s_layer,s_start_unit,d_net,d_layer,i,
			    v_net,v_layer,i,v_group,v_other_layer,v_link_tag,
			    new_link_ids);
    if (j==-1) ABORTBUILD(net);
    ret += j;
  }
  return ret;
}
/**********************************************************************/
