/*
 * $Id: traverse.c,v 1.1 90/12/07 12:01:25 drew Exp $
 * 
 * $Log:	traverse.c,v $
 * Revision 1.1  90/12/07  12:01:25  drew
 * Initial revision
 * 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <itf.h>
#include "sim-dummy.h"
#include <sim.h>

/***
 * Find a LinkSet
 * 1 is returned if the LinkSet is found, 0 if not
 * If the link is found, the ls pointer is set to point to the link set,
 * Assume that everything is in range - checked by asserts
 */
int FindLinkSet(s_net,d_net,s_layer,d_layer,s_unit,out_ls,more)
int s_net,d_net,s_layer,d_layer,s_unit,d_unit;
LINK_SET **out_ls;
int *link_n;
{
  LINK_SET *ls;
  int i,ls_n; /* i is the # of the link */
  assert(s_net>=0 && s_net<n_nets && d_net>=0 && d_net<n_nets);
  assert(NetExists(neta[s_net]) && NetExists(neta[d_net]));
  assert(s_layer>=0 && s_layer<neta[s_net]->n_layers);
  assert(d_layer>=0 && d_layer<neta[d_net]->n_layers);
  assert(s_unit>=0 && s_unit<neta[s_net]->layer[s_layer].n_units);
  assert(d_unit>=0 && d_unit<neta[d_net]->layer[d_layer].n_units);
  /***
   * First look for a link in the destination layer
   */
  for (ls_n=0;ls_n<neta[d_net]->layer[d_layer].info[d_unit].n_lss;ls_n++) {
    ls = neta[d_net]->layer[d_layer].info[d_unit].ls[ls_n];
    /* -- check how to find the unit# -- */
  }
  /***
   * No inward link in the destination layer - look for a
   * link in the source layer
   */
  return 0;
}

/***
 * Find a link
 * 1 is returned if the link is found, 0 if not
 * If the link is found, the ls pointer is set to point to the link set,
 * and link_n is set to the number of the link in that linkset
 * Assume that everything is in range - checked by asserts
 */
int FindLink(s_net,d_net,s_layer,d_layer,s_unit,d_unit,out_ls,link_n)
int s_net,d_net,s_layer,d_layer,s_unit,d_unit;
LINK_SET **out_ls;
int *link_n;
{
  LINK_SET *ls;
  int i,ls_n; /* i is the # of the link */
  assert(s_net>=0 && s_net<n_nets && d_net>=0 && d_net<n_nets);
  assert(NetExists(neta[s_net]) && NetExists(neta[d_net]));
  assert(s_layer>=0 && s_layer<neta[s_net]->n_layers);
  assert(d_layer>=0 && d_layer<neta[d_net]->n_layers);
  assert(s_unit>=0 && s_unit<neta[s_net]->layer[s_layer].n_units);
  assert(d_unit>=0 && d_unit<neta[d_net]->layer[d_layer].n_units);
  /***
   * First look for a link in the destination layer
   */
  for (ls_n=0;ls_n<neta[d_net]->layer[d_layer].info[d_unit].n_lss;ls_n++) {
    ls = neta[d_net]->layer[d_layer].info[d_unit].ls[ls_n];
    /* -- check how to find the unit# -- */
  }
  panic("FindLink","Not Yet Implemented");
  /***
   * No inward link in the destination layer - look for a
   * link in the source layer
   */
  return 0;
}
