/*
 * $Id: equal-sign.c,v 1.1 90/12/07 12:13:34 drew Exp $
 * 
 * $Log:	equal-sign.c,v $
# Revision 1.1  90/12/07  12:13:34  drew
# Initial revision
# 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <math.h>
#include <stdio.h>

#ifdef mips
#include <fp_class.h>
#ifdef REAL == float
#define signbit(x) ((fp_class_f(x)&(FP_NEG_NORM|FP_NEG_DENORM))!=0)
#else
#define signbit(x) ((fp_class_d(x)&(FP_NEG_NORM|FP_NEG_DENORM))!=0)
#endif REAL == float
#endif mips
/***
 * program to determine what is the fastest way of checking that
 * the signs of two reals are equal
 */

char *my_name;

#define MAXRAND ((1<<31)-1)
usage()
{
  fprintf(stderr,"usage: %s {signbit|zerocomp|multiplication|overhead} <n-repititions>\n",my_name);
  exit(1);
}

main(argc,argv)
int argc;
char *argv[];
{
  int times,i,check;
  REAL x,y,sum;
  my_name = argv[0];
  if (argc<3) usage();
  times = atoi(argv[2]);
  if (times<1) usage();
  check=0;
  sum=0;
  if (!strncmp(argv[1],"signbit",strlen(argv[1]))) {
    for (i=0;i<times;i++) {
      x = random()/(MAXRAND/2.0) - 1;
      y = random()/(MAXRAND/2.0) - 1;
      if (signbit(x)==signbit(y)) check++;
      sum += x*y;
    }
    printf("Comparisons done with (signbit(x)==signbit(y)), check=%d, sum=%g\n",
	   check,sum);
  } else if (!strncmp(argv[1],"zerocomp",strlen(argv[1]))) {
    for (i=0;i<times;i++) {
      x = random()/(MAXRAND/2.0) - 1;
      y = random()/(MAXRAND/2.0) - 1;
      if ((x>=0)==(y>=0)) check++;
      sum += x*y;
    }
    printf("Comparisons done with ((x>=0)==(y>=0)), check=%d, sum=%g\n",
	   check,sum);
  } else if (!strncmp(argv[1],"multiplication",strlen(argv[1]))) {
    for (i=0;i<times;i++) {
      x = random()/(MAXRAND/2.0) - 1;
      y = random()/(MAXRAND/2.0) - 1;
      if (x*y >= 0) check++;
      sum += x*y;
    }
    printf("Comparisons done with (x*y >= 0), check=%d, sum=%g\n",check,sum);
  } else if (!strncmp(argv[1],"overhead",strlen(argv[1]))) {
    for (i=0;i<times;i++) {
      x = random()/(MAXRAND/2.0) - 1;
      y = random()/(MAXRAND/2.0) - 1;
      sum += x*y;
    }
    printf("Comparisons done with (x*y >= 0), check=%d, sum=%g\n",check,sum);
  }
}
