/*
 * $Id: bboard.c,v 1.1 90/12/07 12:29:52 drew Exp $
 * 
 * $Log:	bboard.c,v $
# Revision 1.1  90/12/07  12:29:52  drew
# Initial revision
# 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/wait.h>
#include <sys/param.h>
#include <curses.h>
#include <math.h> /* for atof() */
#include "hash-table.h"
#include <argf.h>

#define DB(X)

#define DEFAULT_SOCKET 2388+getuid()
#define VERSION "1.1"

/***
 * TODO: Improvements to make more independent of blot-size & window boxing
 *  check that the log stuff works ok
 * (3) have "labels" for a window, bboard works out where to put them
 * (4) add option to "pack" labels - no extra line for them - they go
 *     where they fit
 * (5) add ability to work with multiple pages - option says which page to
 *     display
 * (8) keep a list of all the values in order, so that they can be read
 *     just as a list of numbers.  Have new command "valuelist <nvalues>"
 */

#define INPUTSIZE 1000
int blot_size;
HASH_TABLE *ht=NULL;
char input[INPUTSIZE];
int context;
int eline;
char name[100];
static int pause=0;
WINDOW *curwin;
int have_geometry;
int expected_in_row;
int at_start_of_row;
int max_lines;
int cur_top;	/* the bottom right of the last window allocated */
int cur_col;
int col_space = 1;
int row_space = 1;
int allbox = 0;
int boxmarks = 1;
int numsize = 0;
int precision = 0;
int using_stdscr = 0;
char *my_name;
char hostname[MAXHOSTNAMELEN];
int port_number = 0;
int sock;
char id[100];
char *logfile=NULL;
int cur_width = 0;
int ignore = 0;	/* set this true when we have too much input */
int native_blot_invert; /* use this to remember original settings */

typedef struct VALUE {
  WINDOW	*win;
  int		row;	/* value in lines from top */
  int		col;	/* value in spaces from left */
  float		value;
  struct VALUE	*next;
};

struct VALUE *value_list = NULL;
struct VALUE *value_list_end = NULL;
int n_values = 0;
int debug = 0;

main(argc, argv)
int argc;
char *argv[];
{
  int i, end;
  struct ARG_DESC *argd;
  my_name = argv[0];
  argd = StartArgs(argv[0]);
  Args(argd, "[-Ppause <ticks>%i %?]", &pause,
       "pause for <N> milliseconds after each frame");
  Args(argd, "[-Plog <file>%s]", &logfile);
  Args(argd, "[-Pdebug %P [<level>%i]]", &debug, &debug);
  Args(argd, "[-Psocket %P %? [<port-number>%n]]", &port_number,
       "accept input at the given or default port number", &port_number);
  EndArgs(argd);
  argv = ParseArgs(argd, argc, argv, 0);
  if (port_number==1)
    port_number = DEFAULT_SOCKET;

  name[0]=0;
  
  if (logfile) {
    if (freopen(logfile, "w", stderr)==NULL) {
      perror(my_name);
      exit(1);
    }
    fprintf(stderr, "Logging errors\n");
  }
  if (port_number>0)
    CreateSocket();
  if (port_number>0)
    sprintf(id, "%s version %s, port is %d %s",
	    my_name, VERSION, port_number, hostname);
  else
    sprintf(id, "%s version %s", my_name, VERSION);
  autoinit_blot();
  native_blot_invert = GetBlotInvert();
  ht = INewTable(ht, 1000, 1000, ID_INDEX);
  context = INewTableContext(ht);
  initscr();
  curwin = stdscr;
  using_stdscr = 0;
  eline = LINES-20;
  blot_size = BlotSize();
#if !(defined(mips) || defined(stellar))
  SO = getcap("md");
  SE = getcap("me");
#endif
  PutId();
  refresh();
  ProcessInput();
  end = cur_top+max_lines+2;
  if (end > LINES-2) end = LINES-2;
  mvcur(LINES-1, COLS-1, 0, 0);
  mvcur(0, 0, end, 0);
  endwin();
}

CreateSocket()
{
  int length;
  int tries;
  struct sockaddr_in server;
  if (port_number==DEFAULT_SOCKET) tries=10;
  else tries=0;
  gethostname(hostname, MAXHOSTNAMELEN);
  /* create socket */
  if ((sock = socket(AF_INET, SOCK_STREAM, 0))<0) error("socket");

  while (1) {
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = INADDR_ANY;
    server.sin_port = htons((unsigned short) port_number);
    if (bind(sock, &server, sizeof(server))) {
      if (errno==EADDRINUSE && tries>0) {
	port_number++;
	tries--;
      } else
	error("bind");
    } else
      break;
  }

  /* Find out assigned port number and print it out */
  length = sizeof(server);
  if (getsockname(sock, &server, &length)) error("getsockname");

  printf("Socket has port #%d\n", ntohs(server.sin_port));
}

PutId()
{
  StdScr();
  move(LINES-1, 0);
  wclrtoeol(stdscr);
  addstr(id);
  ShowBlot(LINES-1, COLS-(blot_size+4)*2, 1.0);
  addstr("+ve");
  ShowBlot(LINES-1, COLS-(blot_size+4), -1.0);
  addstr("-ve");
}

FILE *file; /* keep this out here so that SkipToEnd() can get at it */

ProcessInput()
{
  if (port_number==0) { /* use standard input */
    if (debug)
      fprintf(stderr, "Reading from stdin\n");
    while (gets(input)!=NULL) {
      if (debug>1)
	fprintf(stderr, "%d> %s\n", ignore, input);
      ProcessLine(input);
    }
  } else {
    int msgsock;
    listen(sock, 5);
    do {
      sprintf(id, "%s version %s, waiting for connection on %d %s",
	      my_name, VERSION, port_number, hostname);
      PutId();
      refresh();
      if (debug)
	fprintf(stderr, "Waiting for connection on port %d %s\n", port_number, hostname);
      if ((msgsock = accept(sock, 0, 0)) == -1)
	perror("accept");
      else {
	sprintf(id, "%s version %s,    accepted connection on %d %s",
		my_name, VERSION, port_number, hostname);
	if (debug)
	  fprintf(stderr, "Accepted connection\n", port_number);
	PutId();
	refresh();
	if (!(file = fdopen(msgsock, "r")))
	  perror("fdopen");
	else
	  while (1) {
	    if (!fgets(input, INPUTSIZE, file)) break;
	    if (debug>1)
	      fprintf(stderr, "%d> %s", ignore, input);
	    ProcessLine(input);
	  }
	close(msgsock);
	/* Clear(); */
      }
    } while (1);
  }
}

SkipToEnd()
{
  while (1) {
    if (!fgets(input, INPUTSIZE, file)) break;
    if (!strcmp("end", input)) break;
  }
}

ProcessLine(line)
char *line;
{
  char what[10], val[256];	/* dcp: increased from 100 to 256 */
  int arg1, arg2, r, len;
  r = sscanf(line, " %s %d %d %[^\n]s", what, &arg1, &arg2, val);
  /* fprintf(stderr, "what=%s, r=%d\n", what, r); */
  if (r<1) return;
  len = strlen(what);
  if (!strncmp("end", what, len) && r==1)
    EndDisplay();
  else if (!strncmp("clear", what, len) && r==1)
    Clear();
  else if (!strncmp("name", what, len) && r>=1
	   && sscanf(line, " %s %s", what, val)==2)
    Name(val);
  else if (!ignore) {
    if (!strncmp("rowhint", what, len) && r==2)
      RowHint(arg1);
    else if (!strncmp("colspace", what, len) && r==2)
      ColSpace(arg1);
    else if (!strncmp("rowspace", what, len) && r==2)
      RowSpace(arg1);
    else if (!strncmp("width", what, len) && r==2)
      cur_width = arg1;
    else if (!strncmp("invert", what, len) && r==1) {
      SetBlotInvert(!native_blot_invert);
      PutId();
    } else if (!strncmp("redraw", what, len) && r==1)
      wrefresh(stdscr);
    else if (!strncmp("refresh", what, len) && r==1)
      if (curwin==stdscr)
	refresh();
      else
	wrefresh(curwin);
    else if (!strncmp("newrow", what, len) && r==1) {
      NewRow(1);
      expected_in_row = 0;
    } else if (!strncmp("numsize", what, len) && r==2) {
      numsize = arg1;
      blot_size = numsize;
    } else if (!strncmp("window", what, len) && r==4)
      Window(arg1, arg2, val);
    else if (!strncmp("string", what, len) && r==4)
      PutString(arg1, arg2, val, 1);
    else if (!strncmp("x", what, len) && r==3)
      PutString(arg1, arg2, blot_size==1?"X":"><", 1);
    else if (!strncmp("ax", what, len) && r==3)
      PutString(arg1, arg2, blot_size==1?"X":"><", 0);
    else if (!strncmp("astring", what, len) && r==4)
      PutString(arg1, arg2, val, 0);
    else if (!strncmp("blot", what, len) && r==4)
      PutBlot(arg1, arg2, val, 1);
    else if (!strncmp("ablot", what, len) && r==4)
      PutBlot(arg1, arg2, val, 0);
    /* else if (!strncmp("box", what, len) && r==1)
       BoxWindow(); */
    else if (!strncmp("allbox", what, len) && r==1)
      allbox = 1;
    else if (!strncmp("noboxmarks", what, len) && r==1)
      boxmarks = 0;
    else if (!strncmp("stdscr", what, len) && r==1)
      StdScr();
    else {
      move(LINES-1, 0);
      printw("bad input format: %s\n", line);
    }
  }
}

Clear()
{
  int id;
  TBL_ITEM *item;
  struct VALUE *p, *next;
  expected_in_row = 0;
  at_start_of_row = 1;
  have_geometry = 0;
  max_lines = 0;
  boxmarks = 1;
  allbox = 0;
  blot_size = BlotSize();
  cur_top = cur_col = 0;
  for (p=value_list;p!=NULL;p=next) {
    next = p->next;
    free(p);
  }
  value_list = NULL;
  n_values = 0;
  cur_width = 0;
  id = 0;
  ignore = 0;
  while ((item=ITableNextByContext(ht, &id, context))!=NULL)
    delwin(item->data.win);
  ITableDeleteByContext(ht, context);
  SetBlotInvert(native_blot_invert);
  clear();
  PutId();
  refresh();
  name[0]=0;
}

Name(new_name)
char *new_name;
{
  if (strcmp(new_name, name)) {
    Clear();
    strcpy(name, new_name);
  }
}

EndDisplay()
{
  have_geometry = 1;
  ignore = 0;
  if (curwin && curwin!=stdscr)
    wrefresh(curwin);
  else if (using_stdscr)
    refresh();
  if (pause) {
    move(LINES-1, 0);
    clrtoeol();
    addstr("pausing...");
    refresh();
#if defined(mips) || defined(stellar)
    sleep(pause/1000);
#else
    usleep(pause*1000);
#endif
    PutId();
    refresh();
  }
}

BoxWindow(name)
char *name;
{
  int i;
  char c;
  box(curwin, '|', '-');
  wmove(curwin, 0, 2);
  waddstr(curwin, name);
  wmove(curwin, 0, 0);
  waddch(curwin, '+');
  wmove(curwin, curwin->_maxy-1, 0);
  waddch(curwin, '+');
  wmove(curwin, 0, curwin->_maxx-1);
  waddch(curwin, '+');
  wmove(curwin, curwin->_maxy-1, curwin->_maxx-1);
  waddch(curwin, '+');
  c = 'a';
  if (boxmarks)
    for (i=2;i<curwin->_maxx-1;i+=2) {
      wmove(curwin, curwin->_maxy-1, i);
      waddch(curwin, c);
      if (c<'z') c++; else c='a';
    }
}

ColSpace(space)
int space;
{
  col_space = space;
}

RowSpace(space)
int space;
{
  row_space = space;
}

RowHint(wins)
int wins;
{
  expected_in_row = wins;
}

NewRow(space)
int space;
{
  cur_top += max_lines;
  if (space) cur_top += row_space;
  cur_col = 0;
  at_start_of_row = 1;
  /* expected_in_row = 0; */
  if (curwin!=stdscr)
    wrefresh(curwin);
  curwin = stdscr;
}

StdScr()
{
  curwin = stdscr;
  using_stdscr = 1;
}

Window(lines, cols, name)
int lines, cols;
char *name;
{
  TBL_ITEM *item;
  int expand_cols=1;
  if (*name=='$') {
    name++;
    expand_cols=0;
  }
  if (curwin && curwin!=stdscr)
    wrefresh(curwin);
  else if (using_stdscr)
    refresh();
  item = ITableLookup(ht, name, context);
  if (item==NULL) {
    if (expand_cols) cols *= blot_size;
    if (cols<4) cols=4;
    if (allbox) {
      if (cols<strlen(name)+2) cols=strlen(name)+2;
      lines += 2;
      cols += 2;
    }
    if (cur_col+cols+col_space >= COLS && cur_col!=0) NewRow(0);
    if (lines > max_lines) max_lines = lines;
    /* NewError(); */
    DB( fprintf(stderr, "new window: %dx%d at %d, %d\n", lines, cols, cur_top, cur_col); )
    if (expected_in_row>0 && cur_col==0) {
      /* work out spacing to center rows */
      if (expected_in_row * cols + (expected_in_row-1) * col_space < COLS)
	/* can fit all of them */
	cur_col = (COLS-expected_in_row*cols-(expected_in_row-1)*col_space)/2;
      else {
	/* can fit some of them, centre those */
	int can_fit = (COLS+col_space)/(cols+col_space);
	cur_col = (COLS-can_fit*cols-(can_fit-1)*col_space)/2;
      }
    }
    if (lines+cur_top>=LINES) {
      ignore = 1;
      move(LINES-1, 0);
      addstr("Error: out of space - try a bigger window");
      refresh();
      return;
    }
    curwin = newwin(lines, cols, cur_top, cur_col);
    ITableInsert(ht, name, context, curwin);
    cur_col += cols + col_space;
    if (allbox) BoxWindow(name);
  } else {
    curwin = item->data.win;
    /* NewError(); */
    /* printw("found old window at %d, %d", curwin->_begy, curwin->_begx); */
  }
  if (curwin!=stdscr) using_stdscr=0;
  cur_width = 0;
}

PutString(line, col, str, centre)
int line, col, centre;
char *str;
{
  if (curwin==stdscr) {
    if (col<0) col += COLS;
    if (line<0) line += LINES;
  } else
    CorrectLineCol(&line, &col, centre, &str);
  wmove(curwin, line, col);
  waddstr(curwin, str);
  if (curwin==stdscr) using_stdscr = 1;
}

CorrectLineCol(line, col, centre, str)
int *line, *col, centre;
char **str;
{
  int width;
  if (**str!='$') {
    (*col) *= blot_size;
    if (centre) {
      if (cur_width<=0 || cur_width*blot_size>=curwin->_maxx)
	width = curwin->_maxx - (allbox?2:0);
      else
	width = cur_width*blot_size;
      *(col) += (curwin->_maxx - width)/2;
    } else
      if (allbox) (*col)++;
  } else {
    (*str)++;
    if (allbox) (*col)++;
  }
  if (allbox) (*line)++;
}

PutBlot(line, col, str, centre)
int line, col, centre;
char *str;
{
  float x;
  x = atof(str);
  CorrectLineCol(&line, &col, centre, &str);
  if (!have_geometry) {
    struct VALUE *v = (struct VALUE*)malloc(sizeof(struct VALUE));
    v->win = curwin;
    v->row = line;
    v->col = col;
    v->value = x;
    v->next = NULL;
    if (value_list==NULL) {
      value_list = value_list_end = v;
    } else {
      value_list_end->next = v;
      value_list_end = v;
    }
  }
  ShowBlot(line, col, atof(str));
}

ShowBlot(line, col, x)
int line, col;
float x;
{
  wmove(curwin, line, col);

  if (numsize==0) { /* display as a blot */
#if defined(stellar) || defined(sgi) || defined(mips)
    wattron(curwin, A_BOLD);
#else
    wstandout(curwin);
#endif

    waddstr(curwin, RawBlot(x));

#if defined(stellar) || defined(sgi) || defined(mips)
    wattroff(curwin, A_BOLD);
#else
    wstandend(curwin);
#endif
  } else {	/* display as a number */
    char buf[20];
    sprintf(buf, "%*g", numsize, /* precision, */ x);
    waddstr(curwin, buf);
  }

  if (curwin==stdscr) using_stdscr = 1;
}

NewError()
{
  move(eline, 0);
  if (++eline >= LINES-1) eline = LINES-20;
}

/***
 * returns 1 if a is a non-null prefix of b
 * leading '-' characters are ignored
 */
int strprefix(a, b)
char *a;
char *b;
{
  int i=0;
  if (a==NULL || b==NULL)
    return 0;
  if (*a=='-' && *b=='-')
    {a++;b++;}
  if (*a != *b)
    return 0;
  while (*a!=NULL && *b!=NULL && *a == *b)
    {a++;b++;i++;}
  if (*a==NULL)
    return i;
  else
    return 0;
}

/*
 * System error: print message and give up.
 */
error(msg)
char *msg;
{
  fprintf(stderr, "%s: ", my_name);
  perror(msg);
  exit(1);
}
