/*
 * $Id: cboard.c,v 1.1 90/12/07 12:30:09 drew Exp $
 * 
 * $Log:	cboard.c,v $
# Revision 1.1  90/12/07  12:30:09  drew
# Initial revision
# 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/wait.h>
#include <sys/param.h>
#include <curses.h>
#include <math.h> /* for atof() */
#include "hash-table.h"

#define DEFAULT_SOCKET 2392
#define VERSION "1.0"

/***
 * TODO: Improvements to make more independent of blot-size & window boxing
 * (1) have one command at start to control boxing "allbox"
 * (2) make windows larger if boxing is required
 * (3) have "labels" for a window, bboard works out where to put them
 * (4) add option to "pack" labels - no extra line for them - they go
 *     where they fit
 * (5) add ability to work with multiple pages - option says which page to
 *     display
 * (6) let strings be offset from centre, as well as absolute position
 *     the offset should be expressed in half-spaces
 * (7) make Clear() clear more things
 */

#define INPUTSIZE 1000
HASH_TABLE *ht=NULL;
char input[INPUTSIZE];
int context;
int eline;
char name[100];
static int pause=0;
WINDOW *curwin;
int have_geometry;
int expected_in_row;
int at_start_of_row;
int max_lines;
int cur_top;	/* the bottom right of the last window allocated */
int cur_col;
int col_space = 1;
int row_space = 1;
int using_stdscr = 0;
char *my_name;
char hostname[MAXHOSTNAMELEN];
int port_number = 0;
int sock;
char id[100];

main(argc,argv)
int argc;
char *argv[];
{
  int i,end;
  my_name = argv[0];
  name[0]=0;
  
  while (*argv && *++argv) {
    if (strprefix(*argv,"-pause"))
      pause = atoi(*++argv);
    else if (strprefix(*argv,"-help")) {
      printf("%s - BlotBoard - for layout and display of blot diagrams\n");
      printf("Options:\n");
      printf("  -pause <N>    pause for <N> milliseconds after each frame\n");
      printf("  -socket <N>   accept input at this port number\n");
      exit(0);
    } else if (strprefix(*argv,"-socket")) {
      if (*++argv) port_number = atoi(*argv);
      else port_number = DEFAULT_SOCKET;
      if (port_number<1000) {
	fprintf(stderr,"%s: port numbers should be above 1000\n",my_name);
	exit(1);
      }
    } else {
      fprintf(stderr,"%s: bad option: %s\n",my_name,argv[0]);
      exit(1);
    }
  }
  if (port_number>0) CreateSocket();
  if (port_number>0)
    sprintf(id,"%s version %s, port is %d %s",
	    my_name,VERSION,port_number,hostname);
  else
    sprintf(id,"%s version %s",my_name,VERSION);
  autoinit_blot();
  ht = INewTable(ht,1000,1000,ID_INDEX);
  context = INewTableContext(ht);
  initscr();
  curwin = stdscr;
  using_stdscr = 0;
  eline = LINES-20;
#if !(defined(mips) || defined(stellar))
  SO = getcap("md");
  SE = getcap("me");
#endif
  move(LINES-1,0);
  addstr(id);
  refresh();
  ProcessInput();
  end = cur_top+max_lines+2;
  if (end > LINES-2) end = LINES-2;
  mvcur(LINES-1,COLS-1,0,0);
  mvcur(0,0,end,0);
  endwin();
}

CreateSocket()
{
  int length;
  int tries;
  struct sockaddr_in server;
  if (port_number==DEFAULT_SOCKET) tries=10;
  else tries=0;
  gethostname(hostname,MAXHOSTNAMELEN);
  /* create socket */
  if ((sock = socket(AF_INET, SOCK_STREAM, 0))<0) error("socket");

  while (1) {
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = INADDR_ANY;
    server.sin_port = htons((unsigned short) port_number);
    if (bind(sock, &server, sizeof(server))) {
      if (errno==EADDRINUSE && tries>0) {
	port_number++;
	tries--;
      } else
	error("bind");
    } else
      break;
  }

  /* Find out assigned port number and print it out */
  length = sizeof(server);
  if (getsockname(sock,&server,&length)) error("getsockname");

  printf("Socket has port #%d\n",ntohs(server.sin_port));
}

ProcessInput()
{
  if (port_number==0) /* use standard input */
    while (gets(input)!=NULL) {
      ProcessLine(input);
    }
  else {
    int msgsock;
    FILE *file;
    listen(sock,5);
    do {
      sprintf(id,"%s version %s, waiting for connection on %d %s",
	      my_name,VERSION,port_number,hostname);
      move(LINES-1,0);
      addstr(id);
      refresh();
      if ((msgsock = accept(sock, 0, 0)) == -1) perror("accept");
      else {
	sprintf(id,"%s version %s, connection accepted                        ", my_name,VERSION);
	move(LINES-1,0);
	addstr(id);
	refresh();
	if (!(file = fdopen(msgsock,"r"))) perror("fdopen");
	while (1) {
	  if (!fgets(input,INPUTSIZE,file)) break;
	  ProcessLine(input);
	}
	close(msgsock);
	Clear();
      }
    } while (1);
  }
}

ProcessLine(line)
char *line;
{
  char what[10],val[100];
  int lines,cols,r,len;
  r = sscanf(line," %s %d %d %[^\n]s",what,&lines,&cols,val);
  /* fprintf(stderr,"what=%s, r=%d\n",what,r); */
  if (r<1) return;
  len = strlen(what);
  if (!strncmp("rowhint",what,len) && r==2)
    RowHint(lines);
  else if (!strncmp("colspace",what,len) && r==2)
    ColSpace(lines);
  else if (!strncmp("rowspace",what,len) && r==2)
    RowSpace(lines);
  else if (!strncmp("clear",what,len) && r==1)
    Clear();
  else if (!strncmp("end",what,len) && r==1)
    EndDisplay();
  else if (!strncmp("redraw",what,len) && r==1)
    wrefresh(stdscr);
  else if (!strncmp("refresh",what,len) && r==1)
    if (curwin==stdscr) refresh();
    else wrefresh(curwin);
  else if (!strncmp("newrow",what,len) && r==1)
    NewRow(1);
  else if (!strncmp("window",what,len) && r==4)
    Window(lines,cols,val);
  else if (!strncmp("name",what,len) && r>=1 && sscanf(line," %s %s",what,val)==2)
    Name(val);
  else if (!strncmp("string",what,len) && r==4)
    PutString(lines,cols,val);
  else if (!strncmp("blot",what,len) && r==4)
    PutBlot(lines,cols,val);
  else if (!strncmp("box",what,len) && r==1)
    BoxWindow();
  else if (!strncmp("stdscr",what,len) && r==1)
    StdScr();
  else {
    move(LINES-1,0);
    printw("bad input format: %s\n",line);
  }
}

Clear()
{
  expected_in_row = 0;
  at_start_of_row = 1;
  have_geometry = 0;
  max_lines = 0;
  cur_top = cur_col = 0;
  ITableDeleteByContext(ht,context);
  name[0]=0;
}

Name(new_name)
char *new_name;
{
  if (strcmp(new_name,name)) {
    Clear();
    strcpy(name,new_name);
  }
}

EndDisplay()
{
  have_geometry = 1;
  if (curwin && curwin!=stdscr) wrefresh(curwin);
  else if (using_stdscr) refresh();
  if (pause) {
    move(LINES-1,0);
    clrtoeol();
    addstr("pausing...");
    refresh();
#if defined(mips) || defined(stellar)
    sleep(pause/1000);
#else
    usleep(pause*1000);
#endif
    move(LINES-1,0);
    clrtoeol();
    addstr(id);
    refresh();
  }
}

BoxWindow()
{
  int i;
  char c;
  box(curwin,'|','-');
  wmove(curwin,0,0);
  waddch(curwin,'+');
  wmove(curwin,curwin->_maxy-1,0);
  waddch(curwin,'+');
  wmove(curwin,0,curwin->_maxx-1);
  waddch(curwin,'+');
  wmove(curwin,curwin->_maxy-1,curwin->_maxx-1);
  waddch(curwin,'+');
  c = 'a';
  for (i=2;i<curwin->_maxx-1;i+=2) {
    wmove(curwin,curwin->_maxy-1,i);
    waddch(curwin,c);
    if (c<'z') c++; else c='a';
  }
}

ColSpace(space)
int space;
{
  col_space = space;
}

RowSpace(space)
int space;
{
  row_space = space;
}

RowHint(wins)
int wins;
{
  expected_in_row = wins;
}

NewRow(space)
int space;
{
  cur_top += max_lines;
  if (space) cur_top += row_space;
  cur_col = 0;
  at_start_of_row = 1;
  expected_in_row = 0;
  if (curwin!=stdscr) wrefresh(curwin);
  curwin = stdscr;
}

StdScr()
{
  curwin = stdscr;
  using_stdscr = 1;
}

Window(lines,cols,name)
int lines,cols;
char *name;
{
  TBL_ITEM *item;
  if (curwin && curwin!=stdscr) wrefresh(curwin);
  else if (using_stdscr) refresh();
  item = ITableLookup(ht,name,context);
  if (item==NULL) {
    if (cur_col+cols+col_space >= COLS && cur_col!=0) NewRow(0);
    if (lines > max_lines) max_lines = lines;
    /* NewError(); */
    /* printw("new window: %dx%d at %d,%d",lines,cols,cur_top,cur_col); */
    if (expected_in_row>0 && cur_col==0) {
      /* work out spacing to center rows */
      if (expected_in_row * cols + (expected_in_row-1) * col_space < COLS)
	cur_col = (COLS-expected_in_row*cols-(expected_in_row-1)*col_space)/2;
    }
    curwin = newwin(lines,cols,cur_top,cur_col);
    ITableInsert(ht,name,context,curwin);
    cur_col += cols + col_space;
  } else {
    curwin = item->data.win;
    /* NewError(); */
    /* printw("found old window at %d,%d",curwin->_begy,curwin->_begx); */
  }
  if (curwin!=stdscr) using_stdscr=0;
}

PutString(line,col,str)
int line,col;
char *str;
{
  if (col<0) col += COLS;
  if (line<0) line += LINES;
  wmove(curwin,line,col);
  waddstr(curwin,str);
  if (curwin==stdscr) using_stdscr = 1;
}

PutBlot(line,col,str)
int line,col;
char *str;
{
  float x;
  char *RawBlot();
  x = atof(str);
  wmove(curwin,line,col);

#ifdef stellar
  wattron(curwin,A_BOLD);
#else
  wstandout(curwin);
#endif

  waddstr(curwin,RawBlot(x));

#ifdef stellar
  wattroff(curwin,A_BOLD);
#else
  wstandend(curwin);
#endif

  if (curwin==stdscr) using_stdscr = 1;
}

NewError()
{
  move(eline,0);
  if (++eline >= LINES-1) eline = LINES-20;
}

/***
 * returns 1 if a is a non-null prefix of b
 * leading '-' characters are ignored
 */
int strprefix(a,b)
char *a;
char *b;
{
  if (a==NULL || b==NULL) return 0;
  if (*a=='-' && *b=='-') {a++;b++;}
  if (*a != *b) return 0;
  while (*a!=NULL && *b!=NULL && *a == *b) {a++;b++;}
  if (*a==NULL) return 1;
  else return 0;
}

/*
 * System error: print message and give up.
 */
error(msg)
char *msg;
{
  fprintf(stderr, "%s: ", my_name);
  perror(msg);
  exit(1);
}
