/*
 * $Id: xblot.c,v 1.1 90/12/07 12:30:32 drew Exp $
 * 
 * $Log:	xblot.c,v $
# Revision 1.1  90/12/07  12:30:32  drew
# Initial revision
# 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <stdio.h>
#include <signal.h>
#include <ctype.h>
#include <argf.h>

char prefix[] = "\\b{";
char postfix[] = "}";
#define BLOT 1
int output_mode = 0;
int trailing_space = 0;
float scale=1.0;
char *myname;

#ifdef mips
typedef signed char signed_char;
#else
typedef char signed_char;
#endif

void Die();

main(argc, argv)
int argc;
char *argv[];
{
  int delimiters = 1;
  int fold = 1;
  int tonum = 0;
  FILE *file = stdin;
  struct ARG_DESC *argd = StartArgs(argv[0]);
  myname = *argv;
  Args(argd, "[-Sdelimit%S %?]", &delimiters,
    "Require '\\b{' '}' around numbers to be printed as blots (default on)");
  Args(argd, "[-Sfoldspace%S %?]", &fold,
    "Discard spaces after a blot (default on) (not relevant in delimit mode)");
  Args(argd, "[-Ptonum%P %?]", &tonum,
       "Translate from xterm bold character blots to numbers");
  EndArgs(argd);
  (void) ParseArgs(argd, argc, argv, 0);
  autoinit_blot();
  signal(SIGINT, Die);
  signal(SIGHUP, Die);
  if (tonum)
    ToNumbers(file, delimiters);
  else if (delimiters)
    ReadDelimited(file);
  else
    ReadNumbers(file, fold);
}

ToNumbers(file, delimiters)
FILE *file;
int delimiters;
{
  static char prefix[] = "[1m";
  static char postfix[] = "[m";
  char *p, *q;
  int in_blot = 0;
  int in_fix = 0;
  int c;
  while ((c=getc(file))!=EOF) {
    if (c=='\0') continue;
    if (!in_blot) {
      if (!in_fix) {
	if (c!=prefix[0]) putchar(c);		/* ordinary char */
	else {
	  p = prefix+1;				/* first char of prefix */
	  in_fix = 1;
	}
      } else {
	if (c==*p) {				/* char in prefix */
	  if (*++p=='\0') {			/* last char of prefix */
	    in_fix = 0;
	    in_blot = 1;
	    fputs("\\b{ ", stdout);
	  }
	} else {				/* char not in prefix */
	  for (q=prefix;q<p;q++) putchar(*q);	/* print out prefix */
	  putchar(c);				/* and char */
	  in_fix = 0;
	}
      }
    } else {					/* in a blot */
      if (!in_fix) {
	float x;
	if (c!=postfix[0]) {			/* ordinary char - is a blot */
	  (void) getchar();			/* throw next away */
	  if (c==79)
	    x = 0;
	  else if (c<78 && c>=32)
	    x = 1.0 - (c - 32.0)/48.0;
	    /* x = -(c-78)/46.0; */
	  else if (c>79 && c<126)
	    x = -2.0 + (c - 32.0)/48.0;
	    /* x = -(126-c)/46.0; */
	  else {
	    fprintf(stderr, "bad char in blot string: %d (%c)\n", c, c);
	    exit(1);
	  }
	  if (in_blot>1) putchar(' ');
	  printf("%.2g", x);
	  in_blot++;
	} else {
	  p = postfix+1;			/* first char of prefix */
	  in_fix = 1;
	}
      } else {
	if (c==*p) {				/* char in prefix */
	  if (*++p=='\0') {			/* last char of prefix */
	    in_fix = 0;
	    in_blot = 0;
	    fputs(" }", stdout);
	  }
	} else {				/* char not in prefix */
	  for (q=postfix;q<p;q++) putchar(*q);	/* print out prefix */
	  putchar(c);				/* and char */
	  in_fix = 0;
	}
      }
    }
  }
}

ReadDelimited(file)
FILE *file;
{
  int p; /* current position in prefix */
  int i;
  int c;
  float x;
  p = 0;
  while ((c=(signed_char)getc(file))!=EOF) {
    if (c==prefix[p]) {
      if (prefix[++p]==0) { /* recognized the prefix, go into blot mode */
	fputs(BlotStart(), stdout);
	while (fscanf(file, "%f", &x)==1) /* while we get numbers, blot them */
	  fputs(RawBlot(x), stdout);
	scanf(postfix); /* may or may not succeed, doesn't really matter */
	fputs(BlotEnd(), stdout);
	p = 0;
      }
    } else {
      /* put any characters of the prefix that we might have recognized */
      for (i=0;i<p;i++)
	putchar(prefix[i]);
      /* and put this character */
      putchar(c);
      p = 0;
    }
  }
}

ReadNumbers(file, fold)
FILE *file;
{
  int c;
  float x;
  char buf[20];
  char *p;
  int point, e, dash;
  while ((c=getc(file))!=EOF) {
    if (isdigit(c) || c=='.' || c=='-') {
      /* read a number into the buffer */
      p = buf;
      point = 0;
      e = 0;
      dash = 0;
      do {
	if (c=='.') point = 1;
	*(p++) = c;
      } while ((c=getc(file))!=EOF
	       && (isdigit(c) || (c=='.'&&!point)
		   || (c=='e'&&!e) || (c=='-'&&!dash))
	       && (p-buf<19));
      /* push back the last character */
      if (c!=EOF) ungetc(c, file);
      *p = 0;
      /* now the number is in the buffer, print as a blot if in range */
      if (sscanf(buf, "%f", &x)!=1 || x<-scale || x>scale) {
	PutString(buf);
      } else {
	EnterBlotMode();
	fputs(RawBlot(x), stdout);
	/* check if there is a space after this, if there is, remember it */
	trailing_space = 0;
	if (fold) {
	  /* chew up all the trailing space */
	  do {
	    c = getc(file);
	    if (c==' ') trailing_space = 1;
	  } while (c==' ');
	  if (c!=EOF) ungetc(c, file);
	} else {
	  /* just look for one trailing space */
	  c = getc(file);
	  if (c==' ') trailing_space = 1;
	  else if (c!=EOF) ungetc(c, file);
	}
      }
    } else {
      /* not a digit, just print the char */
      PutChar(c);
    }
  }
  LeaveBlotMode();
}

void Die()
{
  LeaveBlotMode();
  exit(0);
}

PutString(s)
char *s;
{
  LeaveBlotMode();
  fputs(s, stdout);
}

PutChar(c)
char c;
{
  LeaveBlotMode();
  putchar(c);
}

LeaveBlotMode()
{
  if (output_mode==BLOT) {
    fputs(BlotEnd(), stdout);
    output_mode = 0;
    if (trailing_space) {
      putchar(' ');
      trailing_space = 0;
    }
  }
}

EnterBlotMode()
{
  if (output_mode!=BLOT) {
    fputs(BlotStart(), stdout);
    output_mode = BLOT;
  }
}
